'''
  $Id$
  
  Copyleft ITEG IT-Engineers GmbH 2021
'''

#import base64
#import configparser
import glob
import ipaddress
import lxc
import os.path
#import socket
import sys

#from clazzes.datetools import import to_datetime_timestamp_millis, 
#from datetime import datetime

#from iteg.lxcutils.helpers import formatFileSize
from iteg.lxcutils import helpers
from iteg.lxcutils import lculogging
#from iteg.lxcutils import networkhelpers

log = lculogging.getLogger()

def getLCUConfiguration():
    if LCUConfiguration.staticLCUConfiguration == None:
        LCUConfiguration.staticLCUConfiguration = LCUConfiguration()
    return LCUConfiguration.staticLCUConfiguration

class LCUConfiguration:
    '''
      holds eventual parsed configuration, i.e. /etc/lxcutils/environment, /etc/lxcutils/conf/*conf
    '''
    
    # static holder for first instance
    staticLCUConfiguration = None

    def __init__(self): 
        self.parseEnvironment()
        LCUConfiguration.staticLCUConfiguration = self

    def parseEnvironment(self, restrictContainerToName:str=""):
        # lcu environment
        self.envFileName = "/etc/lxcutils/environment" 
        conf = helpers.readConf(self.envFileName)
        #log.debug("parseEnvironment(): conf=\n%s" % conf)
        self.defaultDnsList = conf.get("CTDNS",          "")
        self.webProxy     = conf.get("PROXY",          "")
        self.dmzNetwork   = conf.get("DMZNETWORK",     "")
        self.dmzMaskBits  = -1
        if len(conf.get("DMZNETMASKBITS", "")) > 0:
            self.dmzMaskBits = int(conf.get("DMZNETMASKBITS", ""))
        self.dmzGateway   = conf.get("DMZGATEWAY",     "")
        self.checkedDmzGateway = None
        self.checkedDmzNetwork = None
        self.wanNetwork   = conf.get("WANNETWORK",     "")
        self.wanMaskBits  = -1
        if len(conf.get("WANNETMASKBITS", "")) > 0:
            self.wanMaskBits = int(conf.get("WANNETMASKBITS", ""))
        self.wanGateway   = conf.get("WANGATEWAY",     "")
        self.checkedWanGateway = None
        self.checkedWanNetwork = None
        self.wanIP        = conf.get("WANIP",          "")
        self.checkedWanIP = None
        self.wanInterface = conf.get("WANIF",          "")
        self.wan6Network  = conf.get("WAN6NETWORK",    "")
        self.wan6Prefix = 0
        if len(conf.get("WAN6PREFIX", "")) > 0:
            self.wan6Prefix = int(conf.get("WAN6PREFIX", ""))
        self.wan6Gateway  = conf.get("WAN6GATEWAY",    "")
        self.checkedWan6Gateway = None
        self.checkedWan6Network = None
        self.ctMtu = 1500
        if len(conf.get("CTMTU", "")) > 0:
            self.ctMtu = int(conf.get("CTMTU", "1500"))
        #if (self.ctMtu < 1000) or (self.ctMtu > 9999):
        #    log.error("ERROR: Ignoring invalid CTMTU %d (not in allowed range 1000..9999)" % self.ctMtu)
        #    self.ctMtu = 1500
        self.checkedCtMtu = None
        # iSCSI
        self.iscsiBasename = conf.get("ISCSIBASENAME", "")
        # checkresults
        self.checkErrors = []
        self.checkWarnings = [] 
        self.checkNotices = [] # below warnings
        # containers
        self.containers = {} # by NAME, not id
        if len(restrictContainerToName) > 0:
            container = LCUContainer(confFileName="/etc/lxcutils/conf/"+restrictContainerToName, lcuConfiguration=self)
            self.containers[container.name] = container
            return
        for confFileName in glob.glob("/etc/lxcutils/conf/*.conf"):
            if confFileName == "/etc/lxcutils/conf/CTx_clear.conf":
                continue
            container = LCUContainer(confFileName=confFileName, lcuConfiguration=self)
            if len(container.name) == 0:
                self.checkErrors.append("ERROR: Ignoring nameless container from %s" % confFileName)
            self.containers[container.name] = container
        self.fnuConfiguration = None
        if os.path.exists("/usr/lib/python3/dist-packages/iteg/freenasutils"):
            from iteg.freenasutils.dto import FNUConfiguration
            self.fnuConfiguration = FNUConfiguration()
        
    def checkEnvironment(self, checkLcuContainersToo:bool=False):
        self.checkErrors = []
        self.checkWarnings = []
        self.checkNotices = [] # below warnings 
        if len(self.defaultDnsList) == 0:
            self.checkWarnings.append("WARNING: Default CTDNS configured empty in %s" % self.envFileName)
        # incomplete DMZ settings?
        if len(self.dmzNetwork) > 0:
            try:
                ipaddress.IPv4Address(self.dmzNetwork)
            except:
                self.checkErrors.append("ERROR: DMZNETWORK in %s is not a valid IPv4 address: '%s'" % (self.envFileName, self.dmzNetwork))
            if self.dmzMaskBits < 8:
                self.checkErrors.append("ERROR: DMZNETWORK configured empty in %s, but DMZNETMASKBITS empty or invalid" % self.envFileName)
            if len(self.dmzGateway) == 0:
                self.checkWarnings.append("WARNING: DMZNETWORK configured non-empty in %s, but DMZGATEWAY empty" % self.envFileName)
            else:
                try:
                    self.checkedDmzGateway = ipaddress.IPv4Address(self.dmzGateway)
                except:
                    self.checkErrors.append("ERROR: DMZGATEWAY in %s is not a valid IPv4 address: '%s'" % (self.envFileName, self.dmzGateway))
            try:
                self.checkedDmzNetwork = ipaddress.IPv4Network("%s/%d" % (self.dmzNetwork, self.dmzMaskBits))
            except:
                self.checkErrors.append("ERROR: DMZNETWORK and DMZNETMASKBITS in %s do not specify a valid IPv4 network: '%s/%d'" % (self.envFileName, self.dmzNetwork, self.dmzMaskBits))
            if (self.checkedDmzGateway != None) and (self.checkedDmzNetwork != None) and (not self.checkedDmzGateway in self.checkedDmzNetwork):
                self.checkWarnings.append("WARNING: DMZGATEWAY in %s is in configured DMZNETWORK, %s not in %s" % (
                    self.envFileName, self.checkedDmzGateway, self.checkedDmzNetwork.with_prefixlen))
        if len(self.wanNetwork) > 0:
            try:
                ipaddress.IPv4Address(self.wanNetwork)
            except:
                self.checkErrors.append("ERROR: WANNETWORK in %s is not a valid IPv4 address: '%s'" % (self.envFileName, self.wanNetwork))
            if self.wanMaskBits < 24:
                self.checkErrors.append("ERROR: WANNETWORK configured empty in %s, but WANNETMASKBITS empty or invalid" % self.envFileName)
            if len(self.wanGateway) == 0:
                self.checkWarnings.append("WARNING: WANNETWORK configured empty in %s, but WANGATEWAY empty" % self.envFileName)
            else:
                try:
                    self.checkedWanGateway = ipaddress.IPv4Address(self.wanGateway)
                except:
                    self.checkErrors.append("ERROR: WANGATEWAY in %s is not a valid IPv4 address: '%s'" % (self.envFileName, self.wanGateway))
            try:
                self.checkedWanNetwork = ipaddress.IPv4Network("%s/%d" % (self.wanNetwork, self.wanMaskBits))
            except:
                self.checkErrors.append("ERROR: WANNETWORK and WANNETMASKBITS in %s do not specify a valid IPv4 network: '%s/%d'" % (self.envFileName, self.wanNetwork, self.wanMaskBits))
            if (self.checkedWanGateway != None) and (self.checkedWanNetwork != None) and (not self.checkedWanGateway in self.checkedWanNetwork):
                self.checkWarnings.append("WARNING: WANGATEWAY in %s is in configured WANNETWORK, %s not in %s" % (
                    self.envFileName, self.checkedWanGateway, self.checkedWanNetwork.with_prefixlen))
        if len(self.wan6Network) > 0:
            try:
                ipaddress.IPv6Address(self.wan6Network)
            except:
                self.checkErrors.append("ERROR: WAN6NETWORK in %s is not a valid IPv6 address: '%s'" % (self.envFileName, self.wan6Network))
            if self.wan6Prefix < 48:
                self.checkErrors.append("ERROR: WAN6NETWORK configured empty in %s, but WAN6PREFIX empty or invalid" % self.envFileName)
            if len(self.wan6Gateway) == 0:
                self.checkWarnings.append("WARNING: WAN6NETWORK configured empty in %s, but WAN6GATEWAY empty" % self.envFileName)
            else:
                try:
                    self.checkedWan6Gateway = ipaddress.IPv6Address(self.wan6Gateway)
                except:
                    self.checkErrors.append("ERROR: WAN6GATEWAY in %s is not a valid IPv6 address: '%s'" % (self.envFileName, self.wan6Gateway))
            try:
                self.checkedWan6Network = ipaddress.IPv6Network("%s/%d" % (self.wan6Network, self.wan6Prefix))
            except:
                self.checkErrors.append("ERROR: WAN6NETWORK and WAN6PREFIX in %s do not specify a valid IPv6 network: '%s/%d'" % (self.envFileName, self.wanNetwork, self.wanMaskBits))
            if (self.checkedWan6Gateway != None) and (self.checkedWan6Network != None) and (not self.checkedWan6Gateway in self.checkedWan6Network):
                if self.checkedWan6Gateway.compressed == "fe80::1":
                    self.checkNotices.append("NOTICE: WAN6GATEWAY in %s is not in configured WAN6NETWORK, %s not in %s, OK for setups at Hetzner and probably elsewhere" % (
                        self.envFileName, self.checkedWan6Gateway, self.checkedWan6Network.with_prefixlen))
                else:
                    self.checkWarnings.append("WARNING: WAN6GATEWAY in %s is not in configured WAN6NETWORK, %s not in %s, possibly OK in some setups though" % (
                        self.envFileName, self.checkedWan6Gateway, self.checkedWan6Network.with_prefixlen))
        if len(self.wanIP) > 0:
            try:
                self.checkedWanIP = ipaddress.IPv4Address(self.wanIP)
            except:
                self.checkErrors.append("ERROR: WANIP in %s is not a valid IPv4 address: '%s'" % (self.envFileName, self.wanIP))
            if (self.checkedWanNetwork != None) and (not self.checkedWanIP in self.checkedWanNetwork):
                self.checkErrors.append("ERROR: WANIP in %s is in configured WANNETWORK, %s not in %s" % (
                    self.envFileName, self.checkedWanIP, self.checkedWanNetwork.with_prefixlen))
        if self.ctMtu != None:
            if (self.ctMtu < 1000) or (self.ctMtu > 9999):
                self.checkErrors.append("ERROR: CTMTU in %d is not in allowed range 1000..9999" % self.ctMtu)
            else:
                self.checkedCtMtu = self.ctMtu
        #if len(self.wanInterface) > 0:
        
        # network checks have to be called specifically by main program:
        # networkhelpers.checkLCUEnvironmentIPsReachable(lcuConfiguration)
        # WISH check interface existence
        
        # containers
        if (checkLcuContainersToo):
            for containerName in self.containers.keys():
                container = self.containers[containerName]
                container.checkConf(lcuConfiguration=self)
                if len(container.checkErrors) > 0:
                    self.checkErrors.extend(container.checkErrors)
                if len(container.checkWarnings) > 0:
                    self.checkWarnings.extend(container.checkWarnings)
                if len(container.checkNotices) > 0:
                    self.checkNotices.extend(container.checkNotices)
        # done

    def getCheckedContainer(self, containerName:str=""):
        if len(containerName) == 0:
            log.error("getCheckedContainer() called with empty container name.")
            sys.exit(1)
        self.checkEnvironment(checkLcuContainersToo=False)
        if len(self.checkErrors) > 0:
            log.error("getCheckedContainer(%s): Aborting without checking container because LCU environment conf contains the following %d errors:\n%s\n" % (
                containerName, len(self.checkErrors), "\n".join(self.checkErrors)))
            sys.exit(1)
        if len(self.checkWarnings) > 0:
            log.warning("getCheckedContainer(%s): LCU environment conf contains the following %d warnings:\n%s\n" % (
                containerName, len(self.checkWarnings), "\n".join(self.checkWarnings)))
            #sys.exit(1)
        if len(self.checkNotices) > 0:
            log.warning("getCheckedContainer(%s): LCU environment conf contains the following %d notices:\n%s\n" % (
                containerName, len(self.checkNotices), "\n".join(self.checkNotices)))
        lcuContainer = self.containers.get(containerName, None)
        if lcuContainer == None:
            log.error("getCheckedContainer(%s): No LCU container named %s, i.e. no /etc/lxcutils/conf/%s.conf, aborting." % (
               containerName, containerName, containerName))
            sys.exit(1)
        lcuContainer.checkConf(lcuConfiguration=self)
        if len(lcuContainer.checkErrors) > 0:
            log.error("getCheckedContainer(%s): Aborting because container conf has the following %d errors:\n%s\n" % (
                containerName, len(lcuContainer.checkErrors), "\n".join(lcuContainer.checkErrors)))
            sys.exit(1)
        if len(lcuContainer.checkWarnings) > 0:
            log.warning("getChecked    Container(%s): Container conf has has the following %d warnings:\n%s\n" % (
                containerName, len(lcuContainer.checkWarnings), "\n".join(lcuContainer.checkWarnings)))
            self.checkWarnings.extend(lcuContainer.checkWarnings)
        if len(lcuContainer.checkNotices) > 0:
            log.notice("getCheckedContainer(%s): Container conf has has the following %d notices:\n%s\n" % (
                containerName, len(lcuContainer.checkNotices), "\n".join(lcuContainer.checkNotices)))
            self.checkNotices.extend(lcuContainer.checkNotices)
        return lcuContainer

class LCUContainer:
    '''
      holds eventual parsed configuration of one container, i.e. one /etc/lxcutils/conf/*conf
      
      support for stuff from /etc/lxcutils/conf/*lxc might be added in the future
      though that probably will go through the python3 bindings from python3-lxc
    '''

    def __init__(self, confFileName:str = "", lcuConfiguration:LCUConfiguration = None):
        self.confFileName = confFileName
        self.parseConfFile(lcuConfiguration=lcuConfiguration)

    def parseConfFile(self, lcuConfiguration:LCUConfiguration = None):
        conf = helpers.readConf(self.confFileName)
        self.name = conf.get("CTNAME", "")
        self.id = 0
        if len(conf.get("CTID", "")) > 0:
            self.id = int(conf.get("CTID", ""))
        self.natName = self.name
        if len(self.natName) > 24:
            if self.id > 0:
                self.natName = "lxc-ct_%d" % self.id
            else:
                log.error("%s has too long CTNAME name (%d > 24) and no CTID set, NAT rules will not work due to length restriction of names for iptables tables." % (self.confFileName, len(self.natName)))
        # targetUid
        self.targetUid = -1
        if self.id > 999 and self.id < 40000:
            self.targetUid = self.id * 100000
        if len(conf.get("CTTARGETUID", "")) > 0:
            self.targetUid = int(conf.get("CTTARGETUID", ""))
        # drbd garbage
        self.drbdGarbage = conf.get("CTDRBDRES", "") + conf.get("CTDRBDNUM", "") + conf.get("CTDRBDPORT", "")
        # dns
        self.hostname = conf.get("CTHOSTNAME", "")
        self.domain = conf.get("CTDOMAIN", "")
        self.searchdomains = conf.get("CTSEARCHDOMAIN", "")
        self.dnsList       = conf.get("CTDNS",          "")
        if len(self.dnsList) == 0:
            if lcuConfiguration == None:
                log.error("%s has empty CTDNS setting for dnsList, but unable to inherit defaultDnsList because lcuConfiguration==None" % self.confFileName)
            else:
                self.dnsList = lcuConfiguration.defaultDnsList
        # IPs
        self.ip0 = conf.get("CTIP0", "")
        self.checkedIp0 = None
        self.ipSix0 = conf.get("CTIPSIX0", "")
        self.checkedIpSix0 = None
        self.mac0 = conf.get("CTMAC0", "")
        self.link0 = conf.get("CTLINK0", "")
        self.nettype0 = conf.get("CTNETTYPE0", "veth")
        self.mtu0 = int(conf.get("CTMTU0", "0"))
        #
        self.ip1 = conf.get("CTIP1", "")
        self.checkedIp1 = None
        self.ipSix1 = conf.get("CTIPSIX1", conf.get("CTIPSIX", ""))
        self.checkedIpSix1 = None
        self.mac1 = conf.get("CTMAC1", "")
        self.link1 = conf.get("CTLINK1", "")
        self.nettype1 = conf.get("CTNETTYPE1", "veth")
        self.mtu1 = int(conf.get("CTMTU1", "0"))
        #
        self.ip2 = conf.get("CTIP2", "")
        self.checkedIp2 = None
        self.ipSix2 = conf.get("CTIPSIX2", "")
        self.checkedIpSix2 = None
        self.mac2 = conf.get("CTMAC2", "")
        self.link2 = conf.get("CTLINK2", "")
        self.nettype2 = conf.get("CTNETTYPE2", "veth")
        self.mtu2 = int(conf.get("CTMTU2", "0"))
        # NAT
        self.natip = conf.get("CTNATIP", "")
        self.checkedNatIP = None
        self.natportstcp = conf.get("CTNATPORTSTCP", "")
        self.natportsudp = conf.get("CTNATPORTSUDP", "")
        # storage
        self.size = -1
        if len(conf.get("CTSIZE", "")) > 0:
            self.size = int(conf.get("CTSIZE", ""))
        self.lvgroup = conf.get("CTLVGROUP", "")
        self.lvname = conf.get("CTLVNAME", "")
        self.lfdir = conf.get("CTLFDIR", "")
        self.lfname = conf.get("CTLFNAME", "")
        self.isportal = conf.get("CTISPORTAL", "")
        self.checkedRootfsDevPath = "" # checked by mounthelpers.checkRootfs*
        # storage addons
        self.nasname = conf.get("CTNASNAME", "")
        self.nasdataset = conf.get("CTNASDATASET", "")
        self.crypttype = conf.get("CTCRYPTTYPE", "")
        # rest 
        self.ostemplate = conf.get("CTOSTEMPLATE", "")
        self.backupTarget = conf.get("CTBACKUPTARGET", "")
        self.backupExecBefore = conf.get("CTBACKUPEXECBEFORE", "")
        # volatile iscsi stuff, filled by mounthelpers.fillIscsiVolatiles() on demand
        self.iscsiName = ""
        self.iscsiBasename = ""
        self.iscsiPortalListeners = "" # list of ip:port
        self.fnuNasInfo = None
        self.fnuExtent = None
        self.fnuTarget = None
        self.fnuTargetExtentAssignment = None
        self.fnuPortal = None
        self.fnuDataset = None
        # check results
        self.checkErrors = []
        self.checkWarnings = []
        self.checkNotices = []


    def checkConf(self, lcuConfiguration:LCUConfiguration):
        self.checkErrors = []
        self.checkWarnings = [] 
        self.checkNotices = []
        if len(self.name) == 0:
            self.checkErrors.append("ERROR: CTNAME configured empty in %s" % self.confFileName)
        if os.path.basename(self.confFileName) != self.name + ".conf":
            self.checkErrors.append("ERROR: CTNAME '%s' differs from basname of %s" % (self.name, self.confFileName))
        if self.id < 1000:
            nrMissingIdProblems = 0
            if len(self.ip0) > 0 and len(self.mac0) == 0:
                self.checkErrors.append("ERROR: CTID configured empty or invalid in %s, but needed to autocompute CTMAC0" % self.confFileName)
                nrMissingIdProblems += 1
            if len(self.ip1) > 0 and len(self.mac1) == 0:
                self.checkErrors.append("ERROR: CTID configured empty or invalid in %s, but needed to autocompute CTMAC1" % self.confFileName)
                nrMissingIdProblems += 1
            if len(self.ip2) > 0 and len(self.mac2) == 0:
                self.checkErrors.append("ERROR: CTID configured empty or invalid in %s, but needed to autocompute CTMAC2" % self.confFileName)
                nrMissingIdProblems += 1
            #if nrMissingIdProblems == 0:
            #    self.checkWarnings.append("WARNING: CTID configured empty or invalid in %s" % self.confFileName)
        # target uid
        if (self.targetUid < -1) or (self.targetUid > 1000000000):
            self.checkErrors.append("ERROR: CTTARGETUID configured invalid in %s: %d" % (self.confFileName, self.targetUid))
        # IPs
        if len(self.ip0) > 0:
            try:
                self.checkedIp0 = ipaddress.IPv4Address(self.ip0)
            except:
                self.checkErrors.append("ERROR: CTIP0 in %s is not a valid IPv4 address: '%s'" % (self.confFileName, self.ip0))
        # FUTURE: maybe check if ip0 is in lcuConfiguration.dmzNetzwork. But this might be ok sometimes
        if len(self.ipSix0) > 0:
            try:
                self.checkedIpSix0 = ipaddress.IPv6Address(self.ipSix0)
            except:
                self.checkErrors.append("ERROR: CTIPSIX0 in %s is not a valid IPv6 address: '%s'" % (self.confFileName, self.ipSix0))
        if (self.mtu0 > 0) and ((self.mtu0 < 1000) or (self.mtu0 > 9999)):
            self.checkErrors.append("ERROR: CTMTU0 in %d is not in allowed range 1000..9999" % self.mtu0)
        # FUTURE: maybe check if ipSix0 is in lcuConfiguration.ip6Netzwork. But this might be ok sometimes
        if len(self.ip1) > 0:
            try:
                self.checkedIp1 = ipaddress.IPv4Address(self.ip1)
            except:
                self.checkErrors.append("ERROR: CTIP1 in %s is not a valid IPv4 address: '%s'" % (self.confFileName, self.ip1))
        # FUTURE: maybe check if ip1 is in lcuConfiguration.wanNetzwork. But this might be ok sometimes
        if len(self.ipSix1) > 0:
            try:
                self.checkedIpSix1 = ipaddress.IPv6Address(self.ipSix1)
            except:
                self.checkErrors.append("ERROR: CTIPSIX1 in %s is not a valid IPv6 address: '%s'" % (self.confFileName, self.ipSix1))
        if (self.mtu1 > 0) and ((self.mtu1 < 1000) or (self.mtu1 > 9999)):
            self.checkErrors.append("ERROR: CTMTU1 in %d is not in allowed range 1000..9999" % self.mtu1)
        # FUTURE: maybe check if ipSix0 is in lcuConfiguration.ip6Netzwork. But this might be ok sometimes
        if len(self.ip2) > 0:
            try:
                self.checkedIp2 = ipaddress.IPv4Address(self.ip2)
            except:
                self.checkErrors.append("ERROR: CTIP2 in %s is not a valid IPv4 address: '%s'" % (self.confFileName, self.ip2))
        if len(self.ipSix2) > 0:
            try:
                self.checkedIpSix2 = ipaddress.IPv6Address(self.ipSix2)
            except:
                self.checkErrors.append("ERROR: CTIPSIX2 in %s is not a valid IPv6 address: '%s'" % (self.confFileName, self.ipSix2))
        if (self.mtu2 > 0) and ((self.mtu2 < 1000) or (self.mtu2 > 9999)):
            self.checkErrors.append("ERROR: CTMTU2 in %d is not in allowed range 1000..9999" % self.mtu2)
        # NAT*
        if len(self.natip) > 0 and self.natip != "auto":
            try:
                self.checkedNatIP = ipaddress.IPv4Address(self.natip)
            except:
                self.checkErrors.append("ERROR: CTNATIP in %s is neither 'auto' nor a valid IPv4 address: '%s'" % (self.confFileName, self.ip2))
             
        # incomplete storage settings?
        haveLvmOpts = (len(self.lvgroup)+len(self.lvname)) > 0
        if haveLvmOpts and (len(self.lvgroup) == 0):
            self.checkErrors.append("ERROR: CTLVGROUP empty but and CTLVNAME not empty in %s" % self.confFileName)
        if haveLvmOpts and (len(self.lvname) == 0):
            self.checkErrors.append("ERROR: CTLVNAME empty but and CTLVGROUP not empty in %s" % self.confFileName)
        haveLoopfileOpts = (len(self.lfdir)+len(self.lfname)) > 0
        if haveLoopfileOpts and (len(self.lfdir) == 0):
            self.checkErrors.append("ERROR: CTLFDIR empty but and CTLFNAME not empty in %s" % self.confFileName)
        if haveLoopfileOpts and (len(self.lfname) == 0):
            self.checkErrors.append("ERROR: CTLFNAME empty but and CTLFDIR not empty in %s" % self.confFileName)
        haveNasOpts = (len(self.nasname)+len(self.nasdataset)) > 0
        #if haveNasOpts and (len(self.nasname) == 0):
        #    self.checkErrors.append("ERROR: CTNASNAME empty but and CTNASDATASET not empty in %s" % self.confFileName)
        #if haveNasOpts and (len(self.nasdataset) == 0):
        #    self.checkErrors.append("ERROR: CTNASDATASET empty but and CTNASNAME not empty in %s" % self.confFileName)
        self.fnuNasInfos = None
        if len(self.nasname) > 0:
            self.fnuNasInfos = lcuConfiguration.fnuConfiguration.nasInfos.get(self.nasname)
            if self.fnuNasInfos == None:
                self.checkErrors.append("ERROR: CTNASNAME %s in %s does not refer to an NAS known to freenasutils" % (
                    self.nasname, self.confFileName))
        if haveLvmOpts or haveLoopfileOpts or haveNasOpts:
            if self.size < 1:
                self.checkErrors.append("ERROR: CTLVM* or CTLF* or CTNASNAME set but CTSIZE not set > 0 in %s" % self.confFileName)
        # iSCS special variants
        if haveNasOpts and (len(self.nasname) == 0) and (len(self.isportal) == 0):
            self.checkErrors.append("ERROR: CTNASDATASET set but both CTNASNAME and CTISPORTAL empty in %s" % self.confFileName)
        if haveNasOpts and (len(self.nasname) == 0) and len(lcuConfiguration.iscsiBasename) == 0:
            self.checkErrors.append("ERROR: CTNASDATASET set but both CTNASNAME and ISCSIBASENAME empty in %s, /etc/lxcutils/environment" % self.confFileName)
        # mutually exclusive storage backends?
        if haveLvmOpts:
            if haveLoopfileOpts:
                self.checkErrors.append("ERROR: Storage options CTLV* (LVM) and CTLF* (loop-mounted .ext4 file) are mutually exclusive but both set in %s" % self.confFileName)
            if haveNasOpts:
                self.checkErrors.append("ERROR: Storage options CTLV* (LVM) and CTIS* (iSCSI) are mutually exclusive but both set in %s" % self.confFileName)
        elif haveLoopfileOpts:
            if haveNasOpts:
                self.checkErrors.append("ERROR: Storage options CTLF* (loop-mounted .ext4 file) and CTIS* (iSCSI) are mutually exclusive but both set in %s" % self.confFileName)
        if (len(self.crypttype) > 0) and (self.crypttype != "luks"):
            self.checkErrors.append("ERROR: Unsupported CTCRYPTTYPE set in %s" % self.confFileName)
        # rest
        if len(self.drbdGarbage) > 0:
            self.checkErrors.append("ERROR: CTDRBD* configured non-empty in %s, but DRBD support has been dropped" % self.confFileName)
        # network checks have to be called specifically by main program:
        # networkhelpers.checkLCUContainerNameResolvable(lcuContainer)
        # rest
        if len(self.ostemplate) == 0:
            self.checkErrors.append("ERROR: CTOSTEMPLATE configured empty in %s" % self.confFileName)
        elif not os.path.isfile("/var/lib/lxcutils/templates/%s.tar.gz" % self.ostemplate):
            self.checkErrors.append("ERROR: CTOSTEMPLATE from %s not found, no /var/lib/lxcutils/templates/%s.tar.gz" % (
                self.confFileName, self.ostemplate))
        # backup stuff
        if (len(self.backupExecBefore) > 0) and (len(self.backupTarget) == 0):
            if len(self.nasdataset) == 0:
                self.checkNotices.append("NOTICE: CTBACKUPEXECBEFORE non-empty but CTBACKUPTARGET empty in %s. CTBACKUPEXECBEFORE will still be performed by lcu-backup." % self.confFileName)
            else:
                log.debug("Suppressed NOTICE for container with non-empty CTNASDATASET: CTBACKUPEXECBEFORE non-empty but CTBACKUPTARGET empty in %s" % self.confFileName)
        # check more?

    
class LXCContainer:
    '''
      holds the name, detected meta data, and an eventual lxc Container object
    '''

    def __init__(self, name:str = "", lxcObject:lxc.Container = None):
        self.name = name
        self.lxcObject = lxcObject
        # self.lxcFileName = lxcObject.config_file_name -> /lxc/foobar/config, not /etc/lxcutils/conf/*.lxc
        self.lxcFileName = "/etc/lxcutils/conf/%s.lxc" % self.name
        self.lxcFileDict = {}
        if os.path.exists(self.lxcFileName):
            log.debug("LXCContainer.init(): Reading .lxc file '%s'" %
                      self.lxcFileName)
            self.lxcFileDict = helpers.readConf(self.lxcFileName)
        else:
            log.warning("LXCContainer.init(): Not reading non-existent .lxc file '%s'" %
                        self.lxcFileName)
        # runtime resources
        self.cur_pid = 0
        self.cur_cpuUse = 0
        self.cur_memUse = 0
        self.cur_kMemUse = 0
        self.cur_txBytes = 0
        self.cur_rxBytes = 0
        self.cur_blkIoUse = 0

    def describeRuntimeResources(self, indent:str="  "):
        desc = ""
        #log.debug("Container name: %s" % self.name)
        desc += indent + "  PID: %d" % self.pid
        desc += indent + "  CPU use: %d" % self.cpuUse
        desc += indent + "  Memory use: %d" % self.memUse
        desc += indent + "  KMem use: %d" % self.kMemUse
        desc += indent + "  TX Bytes: %d" % self.txBytes
        desc += indent + "  RX Bytes: %d" % self.rxBytes
        desc += indent + "  Total Bytes: %d" % (self.txBytes + self.rxBytes)
        desc += indent + "  BlkIO use: %d" % self.blkIoUse





















