'''
  $Id$
  
  Copyleft ITEG IT-Engineers GmbH 2021
'''

#import fileinput
#import ipaddress
#import json
import os
#import re
import subprocess
import sys
import time

from iteg.lxcutils import lculogging

log = lculogging.getLogger()


def checkIscsiLogin(
        basename:str="",
        portalIPPort:str="",
        targetName:str=""):
    '''
        checks if there already is a session for the specified target.
        tries to login if not.
        on success True is returned
        on failure sys.exit(1) is called
    '''
    if len(basename) == 0:
        raise Exception("ERROR: checkIscsiLogin() called with empty basename")
    if len(portalIPPort) == 0:
        raise Exception("ERROR: checkIscsiLogin() called with empty portalIPPort")
    if len(targetName) == 0:
        raise Exception("ERROR: checkIscsiLogin() called with empty targetName")
    if not os.path.exists("/sbin/iscsiadm"):
        raise Exception("ERROR: checkIscsiLogin(): /sbin/iscsiadm not found, please install and configure openiscsi")
    if detectOpenSession(basename=basename, portalIPPort=portalIPPort, targetName=targetName):
        log.debug("checkIscsiLogin(): Good, found session to portal %s with basename %s for target %s" % (
            basename, portalIPPort, targetName))
        return True
    log.notice("checkIscsiLogin(): Trying login because there is no session to portal %s with basename %s for target %s" % (
        basename, portalIPPort, targetName))
    discoveredLine = discoverIscsiTarget(basename=basename, portalIPPort=portalIPPort, targetName=targetName)
    if len(discoveredLine) == 0:
        log.error("checkIscsiLogin(): Aborting because discovery did not indicate possible session to portal %s with basename %s for target %s" % (
            basename, portalIPPort, targetName))
        sys.exit(1)
    performIscsiLogin(basename=basename, portalIPPort=portalIPPort, targetName=targetName)
    if not detectOpenSession(basename=basename, portalIPPort=portalIPPort, targetName=targetName):
        log.error("checkIscsiLogin(): Aborting because after login attempt still no session to portal %s with basename %s for target %s" % (
            basename, portalIPPort, targetName))
        sys.exit(1)
    return True

def detectOpenSession(
        basename:str="",
        portalIPPort:str="",
        targetName:str=""):
    '''
        detects if there already is a session for the specified target.
        if yes, True is returned
    '''
    if len(basename) == 0:
        raise Exception("ERROR: checkIscsiLogin() called with empty basename")
    if len(portalIPPort) == 0:
        raise Exception("ERROR: checkIscsiLogin() called with empty portalIPPort")
    if len(targetName) == 0:
        raise Exception("ERROR: checkIscsiLogin() called with empty targetName")
    if not os.path.exists("/sbin/iscsiadm"):
        raise Exception("ERROR: checkIscsiLogin(): /sbin/iscsiadm not found, please install and configure openiscsi")
    cmdParts = ["/sbin/iscsiadm", "--mode", "session"]
    cp = subprocess.run(cmdParts, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
    if cp.returncode == 21:
        log.warning("Failed to get openiscsi sessions, but returncode 21 suggests good empty session list")
        return False
    if cp.returncode != 0:
        log.warning("Failed to get openiscsi sessions, assuming empty session list (not an error)")
        #sys.exit(1)
        return False
    lines = cp.stdout.split("\n")
    sessionLine = ""
    for line in lines:
        lineParts = line.split(" ")
        # tcp: [1] 192.168.1.2:3260,1 iqn.2011-05.org.clazzes:testvol (non-flash)
        if len(lineParts) < 5:
            continue
        if not lineParts[2].startswith(portalIPPort+","):
            continue
        if lineParts[3] == basename+":"+targetName:
            sessionLine = line
            break
    return len(sessionLine) > 0

def discoverIscsiTarget(
        basename:str="",
        portalIPPort:str="",
        targetName:str=""):
    '''
        performs a iscsiadm discovery and searches if the expeted target is among the listed targets
        if successful, the line is returned, otherwise empty line is returned
    '''
    if len(basename) == 0:
        raise Exception("ERROR: discoverIscsiTarget() called with empty basename")
    if len(portalIPPort) == 0:
        raise Exception("ERROR: discoverIscsiTarget() called with empty portalIPPort")
    if len(targetName) == 0:
        raise Exception("ERROR: discoverIscsiTarget() called with empty targetName")
    if not os.path.exists("/sbin/iscsiadm"):
        raise Exception("ERROR: discoverIscsiTarget(): /sbin/iscsiadm not found, please install and configure openiscsi")
    cmdParts = ["/sbin/iscsiadm", "--mode","discovery", "--type","sendtargets", "--portal",portalIPPort]
    cp = subprocess.run(cmdParts, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
    if cp.returncode != 0:
        log.warning("Failed to perform openiscsi discovery, retrying once in a few seconds")
        time.sleep(5)
        log.notice("Retrying openiscsi discovery ...")
        cp = subprocess.run(cmdParts, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
        if cp.returncode != 0:
            log.error("Failed to perform openiscsi discovery again")
            sys.exit(1)
        log.notice("Successful openiscsi discovery at retry")
    lines = cp.stdout.split("\n")
    sessionLine = ""
    for line in lines:
        lineParts = line.split(" ")
        # tcp: [1] 192.168.1.2:3260,1 iqn.2011-05.org.clazzes:testvol (non-flash)
        if len(lineParts) < 2:
            continue
        if not lineParts[0].startswith(portalIPPort+","):
            continue
        if lineParts[1] == basename+":"+targetName:
            sessionLine = line
            break
    return sessionLine

def performIscsiLogin(
        basename:str="",
        portalIPPort:str="",
        targetName:str=""):
    '''
        tries a login
        if successful, True is returned
    '''
    if len(basename) == 0:
        raise Exception("ERROR: performIscsiLogin() called with empty basename")
    if len(portalIPPort) == 0:
        raise Exception("ERROR: performIscsiLogin() called with empty portalIPPort")
    if len(targetName) == 0:
        raise Exception("ERROR: performIscsiLogin() called with empty targetName")
    if not os.path.exists("/sbin/iscsiadm"):
        raise Exception("ERROR: performIscsiLogin(): /sbin/iscsiadm not found, please install and configure openiscsi")
    cmdParts = ["/sbin/iscsiadm", "--mode","node", "--targetname",basename+":"+targetName, "--portal",portalIPPort, "--login"]
    cp = subprocess.run(cmdParts, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
    if cp.returncode != 0:
        log.warning("Failed to get openiscsi login for target %s:%s on %s, retrying noce in a few seconds" % (
            basename, targetName, portalIPPort))
        time.sleep(5)
        log.notice("Retrying openiscsi login ...")
        cp = subprocess.run(cmdParts, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
        if cp.returncode != 0:
            log.error("Failed to perform openiscsi login again")
            sys.exit(1)
        log.notice("Successful openiscsi login at retry")
    log.good("performIscsiLogin(): Successfully logged in to portal %s with basename %s for target %s" % (
            portalIPPort, basename, targetName))
    '''
    Logging in to [iface: default, target: iqn.2011-05.org.clazzes:testvol, portal: 192.168.1.2,3260] (multiple)
    Login to [iface: default, target: iqn.2011-05.org.clazzes:testvol, portal: 192.168.1.2,3260] successful.
    '''
    return True


def performIscsiLogouts(
        basename:str="",
        portalListeners:list=(),
        targetName:str=""):
    for portalIPPort in portalListeners:
        _performIscsiLogout(basename=basename, portalIPPort=portalIPPort, targetName=targetName)

def _performIscsiLogout(
        basename:str="",
        portalIPPort:str="",
        targetName:str=""):
    '''
        tries a login
        if successful, True is returned
    '''
    if len(basename) == 0:
        raise Exception("ERROR: performIscsiLogout() called with empty basename")
    if len(portalIPPort) == 0:
        raise Exception("ERROR: performIscsiLogout() called with empty portalIPPort")
    if len(targetName) == 0:
        raise Exception("ERROR: performIscsiLogout() called with empty targetName")
    if not os.path.exists("/sbin/iscsiadm"):
        raise Exception("ERROR: performIscsiLogout(): /sbin/iscsiadm not found, please install and configure openiscsi")
    if not detectOpenSession(basename=basename, portalIPPort=portalIPPort, targetName=targetName):
        log.good("performIscsiLogout(): Good, found no session to portal %s with basename %s for target %s, nothing to logout from" % (
            basename, portalIPPort, targetName))
        return True
    cmdParts = ["/sbin/iscsiadm", "--mode","node", "--targetname",basename+":"+targetName, "--portal",portalIPPort, "--logout"]
    cp = subprocess.run(cmdParts, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
    if cp.returncode != 0:
        log.error("Failed to get openiscsi logout")
        sys.exit(1)
    '''
    Logging out of session [sid: 1, target: iqn.2011-05.org.clazzes:testvol, portal: 192.168.1.2,3260]
    Logout of [sid: 1, target: iqn.2011-05.org.clazzes:testvol, portal: 192.168.1.2,3260] successful.
    '''
    log.good("performIscsiLogout(): Successfully logged out of portal %s with basename %s for target %s" % (
            basename, portalIPPort, targetName))
    return True

def rediscoverIscsiTargets(
        basename:str="",
        portalIPPort:str=""):
    '''
        performs a iscsiadm discovery --rescan
    '''
    if len(basename) == 0:
        raise Exception("ERROR: rediscoverIscsiTargets() called with empty basename")
    if len(portalIPPort) == 0:
        raise Exception("ERROR: rediscoverIscsiTargets() called with empty portalIPPort")
    if not os.path.exists("/sbin/iscsiadm"):
        raise Exception("ERROR: rediscoverIscsiTargets(): /sbin/iscsiadm not found, please install and configure openiscsi")
    cmdParts = ["/sbin/iscsiadm", "--mode","discovery", "--type","sendtargets", "--portal",portalIPPort, "--rescan"]
    cp = subprocess.run(cmdParts, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
    if cp.returncode != 0:
        log.error("Failed to perform openiscsi rediscovery")
        sys.exit(1)
    log.info("Successfully performed openiscsi rediscovery")

def rescanIscsiTargets(
        basename:str="",
        portalListeners:list=(),
        targetName:str=""):
    for portalIPPort in portalListeners:
        _rescanIscsiTarget(basename=basename, portalIPPort=portalIPPort, targetName=targetName)

def _rescanIscsiTarget(
        basename:str="",
        portalIPPort:str="",
        targetName:str=""):
    '''
        rescans a target for after a resize
    '''
    if len(basename) == 0:
        raise Exception("ERROR: rescanIscsiTarget() called with empty basename")
    if len(portalIPPort) == 0:
        raise Exception("ERROR: rescanIscsiTarget() called with empty portalIPPort")
    if len(targetName) == 0:
        raise Exception("ERROR: rescanIscsiTarget() called with empty targetName")
    if not os.path.exists("/sbin/iscsiadm"):
        raise Exception("ERROR: rescanIscsiTarget(): /sbin/iscsiadm not found, please install and configure openiscsi")
    cmdParts = ["/sbin/iscsiadm", "--mode","node", "--targetname",basename+":"+targetName, "--portal",portalIPPort, "--rescan"]
    cp = subprocess.run(cmdParts, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
    if cp.returncode != 0:
        log.error("Failed to rescan target %s:%s on %s" % (
            basename, targetName, portalIPPort))
        sys.exit(1)
    log.info("Successfully rescanned target %s:%s on %s" % (
        basename, targetName, portalIPPort))

