'''
  $Id$
  
  Copyleft ITEG IT-Engineers GmbH 2021
'''

import getopt
import glob
import lxc
import os
import shutil
import subprocess
import sys
#import configparser

from iteg.lxcutils import dto
from iteg.lxcutils import helpers
from iteg.lxcutils import lculogging
from iteg.lxcutils import mounthelpers

log = lculogging.getLogger()


def describeUsage(argv0):
    usage = ""
    usage += "Usage: %s [options...] containerName\n" % argv0
    usage += "\n"
    usage += "Refreshes a backup of specified LVM-based or .ext4-based container if\n"
    usage += "1. the container's .conf file has CTBACKUP configured to a supported target,\n"
    usage += "2. the container is running (see --stopped below).\n"
    usage += "\n"
    usage += "If rootfs is on a iSCSI target a FreeNAS/TrueNAS snapshot is attempted\n"
    usage += "while the filesystem is frozen for a moment,\n"
    usage += "BUT iSCSI-backed containers are NOT SYNCED AWAY by lcu-backup,\n"
    usage += "the snapshot IS the backup!\n"
    usage += "\n"
    usage += "If the rootfs is based on a crypted block device NO BACKUP is performed.\n"
    usage += "\n"
    usage += "If the rootfs is based on a dedicated LVM LV, a snapshot LV is backed up,\n"
    usage += "which is a 'frozen' view of the filesystem.\n"
    usage += "\n"
    usage += "If the rootfs is based on an LF* .ext4 file no freezing is performed,\n"
    usage += "because the rsync-based backup could run way too long for fs freezing.\n"
    usage += "\n"
    usage += "If the configured .ext4 backup target does not exist\n"
    usage += "automatic creation is attempted without asking.\n"
    usage += "\n"
    usage += "On consistency:\n"
    usage += "Even the fs freezing does not guarantee internal consistency of open files like\n"
    usage += "databases, logfiles and the like!\n"
    usage += "It is recommended to run eventual data backup jobs or other consistency helpers\n"
    usage += "right before, so the backup contains fresh database dumps and rotated logfiles.\n"
    usage += lculogging.getCommonUsageOptions()
    usage += "  Specific options are:\n"
    usage += "    -s | --stopped\n"
    usage += "              Backup container even if it's running.\n"
    usage += "              Does NOT stop/backup/start the container.\n"
    usage += "              Idea is to not have LXC hosts interfer over shared containers,\n"
    usage += "              and not to re-backup obsolete or cold-standby containers.\n"
    return usage


def run_lcu_backup(argv0,args):
    
    try:
        opts, containerNames = getopt.getopt(args, "dhqstv", ["help","stopped"])
    except getopt.GetoptError as err:
        print(err, file=sys.stderr) # will print something like "option -a not recognized"
        print(describeUsage(argv0), file=sys.stderr)
        sys.exit(64)

    lcuConfiguration = dto.LCUConfiguration()
    backupStoppedCTToo = False
    
    for o, a in opts:
        if o in ["-d", "-q", "-t", "-v"]:
            continue
        elif o in ["-h" , "--help"]:
            print(describeUsage(argv0))
            sys.exit(0)
        elif o in ["-s" , "--stopped"]:
            backupStoppedCTToo = True
        else:
            log.warning("Unhandled option: %s %s" % (
                o, a))
    
    if len(containerNames) != 1:
        log.error("Expecting exactly 1 container name, not %d." %
              len(containerNames))
        print(describeUsage(argv0), file=sys.stderr, flush=True)
        sys.exit(1)
    containerName = containerNames[0]
    lcuContainer = lcuConfiguration.getCheckedContainer(containerName)
    
    finalRc = 0

    if len(lcuContainer.crypttype) > 0:
        log.error("Container has CTCRYPTTYPE set non-empty")
        if len(lcuContainer.nasdataset) > 0:
            log.warning("NO BACKUP of container %s is performed, eventual preparation command and FreeNAS snapshot will be performed now." % containerName)
            finalRc += performBackupExecBefore(lcuContainer)
            createFNUSnapshot(lcuContainer, fnuConfiguration=lcuConfiguration.fnuConfiguration)
            log.warning("NO BACKUP of container %s was performed, eventual preparation command was performed, FreeNAS snapshot was created, exiting for good." % containerName)
        else:
            log.warning("NO BACKUP of container %s is performed, eventual preparation command will be performed now." % containerName)
            finalRc += performBackupExecBefore(lcuContainer)
            log.warning("NO BACKUP of container %s was performed, eventual preparation command was performed, exiting for good." % containerName)
        if finalRc > 0:
            log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
        sys.exit(finalRc)
        
    lxcContainer = dto.LXCContainer(lcuContainer.name, lxcObject=lxc.Container(name=containerName))
    
    ctVarLogPath = "/var/lib/lxc/%s/rootfs/var/log" % containerName
    
    if not lxcContainer.lxcObject.running:
        if not backupStoppedCTToo:
            log.warning("Container %s is stopped and -s not specified, aborting." %
                        len(containerNames))
            if finalRc > 0:
                log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
            sys.exit(finalRc)
        if not os.path.isdir(ctVarLogPath):
            log.warning("Container %s is stopped and %s not visible, trying to mount rootfs." % (
                containerName, ctVarLogPath))
        mounthelpers.mountRootfs(lcuContainer=lcuContainer, lxcContainer=lxcContainer)

    if not os.path.isdir(ctVarLogPath):
        log.error("Container's /var/log directory %s is not visible, aborting." % ctVarLogPath)
        log.error("NO BACKUP of container %s is performed, eventual preparation command will NOT be performed." % containerName)
        finalRc += 1
        log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
        sys.exit(finalRc) # this also makes sure we don't operate on containers with weird rootfs paths

    if len(lcuContainer.backupTarget) == 0:
        log.warning("Container's CTBACKUPTARGET %s is configured empty." % lcuContainer.backupTarget)
        log.warning("NO BACKUP of container %s is performed, eventual preparation command will be performed now." % containerName)
        finalRc += performBackupExecBefore(lcuContainer)
        log.warning("NO BACKUP of container %s was performed, eventual preparation command was performed, exiting for good." % containerName)
        if finalRc > 0:
            log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
        sys.exit(finalRc)

    if lcuContainer.backupTarget == "-":
        log.warning("Container's CTBACKUPTARGET %s is configured to '-'." % lcuContainer.backupTarget)
        log.warning("NO BACKUP of container %s is performed, eventual preparation command will be performed now." % containerName)
        finalRc += performBackupExecBefore(lcuContainer)
        log.warning("NO BACKUP of container %s was performed, eventual preparation command was performed, exiting for good." % containerName)
        if finalRc > 0:
            log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
        sys.exit(finalRc)

    if ":" in lcuContainer.backupTarget:
        log.error("Container's CTBACKUPTARGET %s is configured for some remote rsync target, not supported any more, aborting." % lcuContainer.backupTarget)
        log.error("NO BACKUP of container %s is performed, eventual preparation command will be performed now." % containerName)
        finalRc += performBackupExecBefore(lcuContainer)
        log.error("NO BACKUP of container %s was performed, eventual preparation command was performed, exiting for good." % containerName)
        finalRc += 1
        log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
        sys.exit(finalRc)
    
    if not lcuContainer.backupTarget.startswith("/"):
        log.error("Container's CTBACKUPTARGET %s is not configured to an absolute path for a non-iSCSI container." % lcuContainer.backupTarget)
        log.error("NO BACKUP of container %s is performed, eventual preparation command will be performed now." % containerName)
        finalRc += performBackupExecBefore(lcuContainer)
        log.error("NO BACKUP of container %s was performed, eventual preparation command was performed, exiting for good." % containerName)
        finalRc += 1
        log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
        sys.exit(finalRc)
        
    if len(lcuContainer.nasdataset) > 0:
        if len(lcuContainer.nasname) > 0:
            log.error("Container is living on an FNU-backed ISCSI target, will be only be making iSCSI snapshot.")
            log.error("NO REAL FULL BACKUP of container %s is performed, eventual preparation command will be performed now." % containerName)
            finalRc += performBackupExecBefore(lcuContainer=lcuContainer)
            createFNUSnapshot(lcuContainer=lcuContainer, fnuConfiguration=lcuConfiguration.fnuConfiguration)
            if finalRc > 0:
                log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
            sys.exit(finalRc)
        else:
            log.error("Container is living on an other (not FNU-based) iSCSI target, no backup supported.")
            log.error("NO BACKUP of container %s is performed, eventual preparation command will be performed now." % containerName)
            finalRc += performBackupExecBefore(lcuContainer)
            log.error("NO BACKUP of container %s was performed, eventual preparation command was performed, exiting for good." % containerName)
            finalRc += 1
            log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
            sys.exit(finalRc)
            
        
    ext4target = ""
    rsyncSource = ""
    rsyncTarget = ""

    # detect and check target and xsyncTarget
    rootfsSizeBytes = mounthelpers.detectRootfsSize(lcuContainer=lcuContainer)
    if lcuContainer.backupTarget.endswith(".ext4"):
        log.debug("CTBACKUPTARGET %s is an .ext4 target" %
                  lcuContainer.backupTarget)
        ext4target = lcuContainer.backupTarget
        if rootfsSizeBytes <= 1048576:
            log.error("Unable to detect size of rootfsDev of container %s, unable to create .ext4 backup target %s." % (
                containerName, ext4target))
            log.error("NO BACKUP of container %s is performed, eventual preparation command will NOT be performed." % containerName)
            finalRc += 1
            log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
            sys.exit(finalRc)
        if not os.path.exists(ext4target):
            log.info("Auto-creating loopfile target %s" % ext4target)
            cp = subprocess.run([ "/usr/bin/truncate", "--size=%d" % rootfsSizeBytes, lcuContainer.backupTarget ])
            if cp.returncode != 0:
                log.error("Unable to create .ext4 backup target %s." % ext4target)
                log.error("NO BACKUP of container %s is performed, eventual preparation command will NOT be performed." % containerName)
                finalRc += 1
                log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
                sys.exit(finalRc)
            cp = subprocess.run([ "/sbin/mkfs.ext4", ext4target])
            if cp.returncode != 0:
                log.error("Unable to mkfs.ext4 backup target %s." % ext4target)
                log.error("NO BACKUP of container %s is performed, eventual preparation command will NOT be performed." % containerName)
                finalRc += 1
                log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
                sys.exit(finalRc)
        else:
            targetSize = -1
            try:
                statResult = os.stat(ext4target)
                if statResult != None:
                    targetSize = int(statResult.st_size)
            except:
                log.error("Unable to detect current size of backup target %s." % ext4target)
                log.error("NO BACKUP of container %s is performed, eventual preparation command will NOT be performed." % containerName)
                finalRc += 1
                log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
                sys.exit(finalRc)
            if targetSize < rootfsSizeBytes:
                log.info("Automatically resizing ext4 backup target %s to %d bytes (%s)." % (
                    ext4target, rootfsSizeBytes, helpers.formatFileSize(rootfsSizeBytes)))
                fsToResize = mounthelpers._resizeLoopFile(ext4target, rootfsSizeBytes)
                cmdParts = [ "/sbin/e2fsck", "-f", "-p", fsToResize]
                cp = subprocess.run(cmdParts)
                if cp.returncode != 0:
                    problem = "ERROR: Failed to e2fsck %s for .ext4 backup of LCU/LXC container %s" % (
                         fsToResize, lcuContainer.name)
                    helpers.backcheckOnProblem(rc=cp.returncode, cmdParts=cmdParts, problem=problem)
                else:
                    log.good("Sucessfully e2fsck-ed ext4 fs on %s for .ext4 backup of LCU/LXC container %s" % (
                         fsToResize, lcuContainer.name))
                cmdParts = [ "/sbin/resize2fs", fsToResize]
                cp = subprocess.run(cmdParts)
                if cp.returncode != 0:
                    problem = "ERROR: Failed to resize2fs %s for .ext4 backup of LCU/LXC container %s" % (
                         fsToResize, lcuContainer.name)
                    helpers.backcheckOnProblem(rc=cp.returncode, cmdParts=cmdParts, problem=problem)
                else:
                    log.good("Sucessfully resized ext4 fs on %s for .ext4 backup of LCU/LXC container %s" % (
                         fsToResize, lcuContainer.name))
                
        rsyncTarget = "/var/lib/lxc/%s/rootfs.bak" % lcuContainer.name
        if not os.path.isdir(rsyncTarget):
            log.info("Auto-creating rsyncTarget directory %s" % rsyncTarget)
            os.makedirs(rsyncTarget, 0o700)
        if not os.path.isdir(rsyncTarget):
            log.error("Unable to auto-create rsyncTarget directory %s" % rsyncTarget)
            log.error("NO BACKUP of container %s is performed, eventual preparation command will NOT be performed." % containerName)
            finalRc += 1
            log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
            sys.exit(finalRc)
        if os.path.ismount(rsyncTarget):
            log.warning("rsyncTarget directory %s is already mounted, unmounting it first" % rsyncTarget)
            cp = subprocess.run(["/bin/umount", "-v", rsyncTarget])
            if cp.returncode != 0 or os.path.ismount(rsyncTarget):
                log.error("Unable to umount old mount on rsyncTarget directory %s" % rsyncTarget)
                log.error("NO BACKUP of container %s is performed, eventual preparation command will NOT be performed." % containerName)
                finalRc += 1
                log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
                sys.exit(finalRc)
        # finally mount
        log.info("Mounting backuptarget %s on rsynctarget %s" % (ext4target, rsyncTarget))
        cp = subprocess.run(["/bin/mount", "-t","ext4", "-o","defaults,errors=remount-ro,acl", "-v", ext4target, rsyncTarget])
        if cp.returncode != 0 or not os.path.ismount(rsyncTarget):
                log.error("Unable to mount rsyncTarget directory %s" % rsyncTarget)
                log.error("NO BACKUP of container %s is performed, eventual preparation command will NOT be performed." % containerName)
                finalRc += 1
                log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
                sys.exit(finalRc)
    else:
        log.debug("CTBACKUPTARGET %s is a directory target" %
                  lcuContainer.backupTarget)
        dirTarget = lcuContainer.backupTarget
        if not os.path.isdir(dirTarget):
            log.info("Auto-creating dirTarget directory %s" % dirTarget)
            os.makedirs(dirTarget, 0o700)
        if not os.path.isdir(dirTarget):
            log.error("Unable to auto-create dirTarget directory %s" % dirTarget)
            log.error("NO BACKUP of container %s is performed, eventual preparation command will NOT be performed." % containerName)
            finalRc += 1
            log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
            sys.exit(finalRc)
        rsyncTarget = dirTarget
         
    if len(rsyncTarget) == 0:
        log.error("Unable to detect rsyncTarget for CTBACKUPTARGET %s." % lcuContainer.backupTarget)
        log.error("NO BACKUP of container %s is performed, eventual preparation command will NOT be performed." % containerName)
        finalRc += 1
        log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
        sys.exit(finalRc)

    # save .conf, .lxc
    ctConfBakDir = "/var/lib/lxc/%s/rootfs/var/backups/lxcutils"
    if not os.path.isdir(ctConfBakDir):
        log.info("Auto-creating ctConfBakDir directory %s" % ctConfBakDir)
        os.makedirs(ctConfBakDir, 0o700)
    if os.path.isdir(ctConfBakDir):
        for confFileName in glob.glob("/etc/lxcutils/conf/%s.*" % containerName):
            shutil.copy2(confFileName, "%s/%s" % (ctConfBakDir, os.path.basename(confFileName)), follow_symlinks=True)

    # prep command
    finalRc += performBackupExecBefore(lcuContainer)

    lvSnapDir = ""
    if len(lcuContainer.lvgroup) > 0:
        if lxcContainer.lxcObject.running:
            lvSnapName = "%s_snap" % lcuContainer.lvname
            lvSnapDir = "/dev/%s/%s" % (lcuContainer.lvgroup, lvSnapName)
            rsyncSource = "/var/lib/lxc/%s/rootfs.snap" % containerName
            if os.path.ismount(rsyncSource):
                log.info("Unmounting old rsyncSource directory %s" % rsyncSource)
                cp = subprocess.run(["/bin/umount", "-v", rsyncSource])
                if cp.returncode != 0:
                    log.error("Unable to umouint old rsyncSource directory %s" % rsyncSource)
                    log.error("NO BACKUP of container %s is performed." % containerName)
                    finalRc += 1
                    log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
                    sys.exit(finalRc)
            if os.path.exists(lvSnapDir):
                log.info("Removing old LVM snapshot %s" % lvSnapDir)
                cp = subprocess.run(["/sbin/lvremove", "-y", lvSnapDir])
                if cp.returncode != 0:
                    log.error("Unable to lvremove old LVM snapshot %s" % lvSnapDir)
                    log.error("NO BACKUP of container %s is performed." % containerName)
                    finalRc += 1
                    log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
                    sys.exit(finalRc)
            log.info("Auto-creating LVM snapshot directory %s" % ctConfBakDir)
            #
            cmdParts = ["/sbin/lvcreate",
                        "-L", "%dm" % int(rootfsSizeBytes/1048576/10),
                        "-n", lvSnapName, 
                        "-s","/dev/%s/%s" % (lcuContainer.lvgroup, lcuContainer.lvname)]
            log.debug("Calling\n%s\n" % " ".join(cmdParts))
            cp = subprocess.run(cmdParts)
            if cp.returncode != 0:
                log.error("Unable to create LVM snapshot %s" % lvSnapDir)
                log.error("NO BACKUP of container %s is performed." % containerName)
                finalRc += 1
                log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
                sys.exit(finalRc)
            if not os.path.isdir(rsyncSource):
                log.info("Auto-creating rsyncSource directory %s" % rsyncSource)
                os.makedirs(rsyncSource, 0o700)
            if not os.path.isdir(rsyncSource):
                log.error("Unable to auto-create rsyncSource directory %s" % rsyncSource)
                log.error("NO BACKUP of container %s is performed." % containerName)
                finalRc += 1
                log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
                sys.exit(finalRc)
            cp = subprocess.run(["/bin/mount", "-v", "-t","ext4", "-o","ro", lvSnapDir, rsyncSource])
            if cp.returncode != 0:
                log.error("Unable to mount LVM snapshot %s on rsyncSource %s" % (lvSnapDir, rsyncSource))
                log.error("NO BACKUP of container %s is performed." % containerName)
                finalRc += 1
                log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
                sys.exit(finalRc)
            log.info("Successfully moutned LVM snapshot %s on rsyncSource %s" % (lvSnapDir, rsyncSource))
        else:
            log.info("Backup up a stopped LVM container, mounting and using the normal rootfs directory")
            mounthelpers.mountRootfs(lcuContainer=lcuContainer, lxcContainer=lxcContainer)
            rsyncSource = "/var/lib/lxc/%s/rootfs" % containerName
    
    elif len(lcuContainer.nasdataset) > 0:
        log.warning("Container %s has ISCSI/FNU-backed rootfs, creating zVol snapshot" % containerName)
        createFNUSnapshot(lcuContainer, fnuConfiguration=lcuConfiguration.fnuConfiguration)
        # we could create a zvol and ISCSI target from that snapshot, but it would be a lot of work
        log.warning("Container %s has ISCSI/FNU-backed rootfs, successfully created zVol snapshot but NOT MAKING FULL BACKUP to %s" % (
            containerName, lcuContainer.backupTarget))
        if finalRc > 0:
            log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
        sys.exit(finalRc)
    else:
        log.warning("Using plain rootfs as rsyncSource to backup container %s" % containerName)
        rsyncSource = "/var/lib/lxc/%s/rootfs" % containerName

    if len(rsyncSource) == 0:
        log.error("Unable to detect rsyncSource for container with CTBACKUPTARGET %s." %
                    lcuContainer.backupTarget)
        log.error("NO BACKUP of container %s is performed." %
                    containerName)
        finalRc += 1
        log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)
        sys.exit(finalRc)
        
    # actually backup
    # not:      -x for one-filesystem-only
    rsyncCmdParts = [
        "/usr/bin/rsync",
        "--delete", "--delete-after", "--force", "-g", "-l", "-o", "-p", "-v", "-S", "-t", "-r", "-W",   # xsync baseparams, recursive, local, 
        "-D", "--numeric-ids","--no-whole-file", "--partial", "--exclude=/tmp/", "--timeout=60",          # lcu demand
        rsyncSource+"/", rsyncTarget+"/" ]
    attemptNr = 0
    rsyncOk = False
    while attemptNr < 3:
        if attemptNr > 0:
            log.warning("Rsync attempt %d FAILED, retrying shortly ..." % attemptNr)
        attemptNr += 1
        log.notice("Rsync attempt %d starting ..." % attemptNr)
        log.debug("Calling\n%s\n" % " ".join(rsyncCmdParts))
        cp = subprocess.run(rsyncCmdParts)
        if cp.returncode == 0:
            log.notice("Rsync attempt %d succeeded" % attemptNr)
            rsyncOk = True
            break
    if not rsyncOk:
        log.error("Rsync FAILED %d, giving up!" % attemptNr)
        finalRc += 1
            
    if len(ext4target) > 0:
        log.notice("Unmounting ext4 target %s" % ext4target)
        cp = subprocess.run(["/bin/umount", "-v", ext4target])
        if cp.returncode != 0:
            log.error("Failed to unmount ext4 target %s, continuing cleanup work anyway" % ext4target)
            finalRc += 1

    if len(lvSnapDir) > 0:
        log.info("Unmounting rsyncSource directory %s" % rsyncSource)
        cp = subprocess.run(["/bin/umount", "-v", rsyncSource])
        if cp.returncode != 0:
            log.error("Unable to umouint rsyncSource directory %s, continuing cleanup work anyway" % rsyncSource)
            finalRc += 1
        log.info("Removing LVM snapshot %s" % lvSnapDir)
        cp = subprocess.run(["/sbin/lvremove", "-y", lvSnapDir])
        if cp.returncode != 0:
            log.error("Unable to lvremove LVM snapshot %s, continuing cleanup work anyway" % lvSnapDir)
            finalRc += 1

    if finalRc > 0:
        log.error("Exiting with bad finalRc %d, look above for reasons" % finalRc)

    sys.exit(finalRc)


def performBackupExecBefore(lcuContainer:dto.LCUContainer=None):
    if lcuContainer == None:
        raise Exception("ERROR: performBackupExecBefore() called with lcuContainer None" % lcuContainer.name)
    if len(lcuContainer.backupExecBefore) > 0:
        log.info("Performing preparation command %s" % lcuContainer.backupExecBefore)
        prepCmdParts = []
        prepCmdParts.extend(lcuContainer.backupExecBefore.split(" "))
        cp = subprocess.run(prepCmdParts)
        if cp.returncode != 0:
            log.error("Backup preparation command failed with rc=%d, cmd was: %s" % (
                cp.returncode, lcuContainer.backupExecBefore))
            log.error("Backup of container %s still will be performed." % lcuContainer.name)
            #sys.exit(1)
            return 1
        else:
            log.info("Successfully performed preparation command %s" % lcuContainer.backupExecBefore)
            return 0
    else:
        log.info("Not performing empty backup preparation command for %s." % lcuContainer.name)
        return 0

def createFNUSnapshot(lcuContainer:dto.LCUContainer=None, fnuConfiguration=None):
    if lcuContainer == None:
        raise Exception("ERROR: createFNUSnapshot() called with lcuContainer None" % lcuContainer.name)
    if fnuConfiguration == None:
        raise Exception("ERROR: createFNUSnapshot() called with fnuConfiguration None" % lcuContainer.name)
    from iteg.freenasutils.restclient import FNURestClient
    fnuRestclient = FNURestClient(fnuConfiguration=fnuConfiguration)
    if lcuContainer.fnuNasInfo == None:
        lcuContainer.fnuNasInfo = fnuConfiguration.nasInfos.get(lcuContainer.nasname)
    #fnuRestclient.getSnapshotNameingSchemas(nasInfo=lcuContainer.fnuNasInfo)
    #if lcuContainer.fnuNasInfo.nas_snapshot_naming_schemas == None:
    #    log.error("Unable to create zVol snapshot, nasInfo for %s has no naming schema" % (
    #        lcuContainer.nasname))
    #    sys.exit(1)
    #if len(lcuContainer.fnuNasInfo.nas_snapshot_naming_schemas) == 0:
    #    log.error("Unable to create zVol snapshot, nasInfo for %s has empty naming schema list" % (
    #        lcuContainer.nasname))
    #    sys.exit(1)
    #namingSchema = lcuContainer.fnuNasInfo.nas_snapshot_naming_schemas[0]
    #if len(namingSchema) == 0:
    #    log.error("Unable to create zVol snapshot, nasInfo for %s has empty first naming schema" % (
    #        lcuContainer.nasname))
    #    sys.exit(1)
    namingSchema = "%Y%m%d_%H%M_lcu-backup"
    rootfsDir = "/var/lib/lxc/%s/rootfs" % lcuContainer.name
    doFreeze = os.path.ismount(rootfsDir)
    dieAfterUnfreeze = False
    if doFreeze:
        log.notice("Freezing %s before taking snapshot." % rootfsDir)
        subprocess.run(["/sbin/fsfreeze", "--freeze", rootfsDir])
    else:
        log.warning("NOT freezing %s before taking snapshot because it's not it's own fs." % rootfsDir)
    try:
        fnuRestclient.createSnapshot(nasInfo=lcuContainer.fnuNasInfo, datasetPath=lcuContainer.nasdataset, namingSchema=namingSchema)
    except Exception as e:
        log.error("Unable to create zVol snapshot, but unfreezing rootfs after swallowing following exception:\n%s\n" % e)
        dieAfterUnfreeze = True
    if doFreeze:
        log.notice("Unfreezing %s after making (attempting) snapshot." % rootfsDir)
        subprocess.run(["/sbin/fsfreeze", "--unfreeze", rootfsDir])
    else:
        log.warning("NOT unfreezing %s before taking snapshot because it's not it's own fs." % rootfsDir)
    if dieAfterUnfreeze:
        log.error("Dying after eventual unfreeze because of previously swallowed error")
        sys.exit(1)
    if log.verbosity <= 1:
        log.good("Container %s has ISCSI/FNU-backed rootfs, successfully created zVol snapshot, not trying to detect snapshot name because of non-verbose verbosity level" %
                 lcuContainer.name)
    else:
        log.good("Container %s has ISCSI/FNU-backed rootfs, successfully created zVol snapshot, trying to detect snapshot name" %
                 lcuContainer.name)
        fnuRestclient.getSnapshots(nasInfo=lcuContainer.fnuNasInfo, datasetPath=lcuContainer.nasdataset)
        createdSnapshotName = ""
        for nasSnapshotId in sorted(lcuContainer.fnuNasInfo.nas_snapshots.keys(), reverse=True):
            nasSnapshot = lcuContainer.fnuNasInfo.nas_snapshots[nasSnapshotId]
            if nasSnapshot.snapshot_name.endswith("_lcu-backup"):
                createdSnapshotName = nasSnapshot.snapshot_name
                break
        if len(createdSnapshotName) > 0:
            log.notice("Successfully created zVol snapshot %s" % createdSnapshotName)
        else:
            log.error("Failed to detect name of created snapshot")
    # we could create a zvol and ISCSI target from that snapshot, but it would be a lot of work
    log.warning("Container %s has ISCSI/FNU-backed rootfs, successfully created zVol snapshot but NOT MAKING FULL BACKUP to %s" % (
        lcuContainer.name, lcuContainer.backupTarget))
