'''
  $Id$
  
  Copyleft ITEG IT-Engineers GmbH 2021
'''

#import datetime
import getopt
import io
#import lxc
import os
#import os.path
import shutil
import subprocess
import sys

from datetime import datetime

from iteg.lxcutils import lculogging

from iteg.lxcutils import dto
from iteg.lxcutils import fshelpers
from iteg.lxcutils import helpers
from iteg.lxcutils import mounthelpers
#from iteg.lxcutils import networkhelpers

log = lculogging.getLogger()


def describeUsage(argv0):
    usage = ""
    usage += "Usage: %s [options...] containerName\n" % argv0
    usage += "\n"
    usage += "Creates the specified LCU container including an eventual rootfs device,\n"
    usage += "if there is a checked_OK file from a previous lcu-check4creation run.\n"
    usage += lculogging.getCommonUsageOptions()
    return usage


def run_lcu_bootstrap(argv0,args):

    try:
        opts, containerNames = getopt.getopt(args, "dhqtv", ["help"])
    except getopt.GetoptError as err:
        print(err, file=sys.stderr) # will print something like "option -a not recognized"
        print(describeUsage(argv0), file=sys.stderr)
        sys.exit(64)

    lcuConfiguration = dto.LCUConfiguration()
    
    for o, a in opts:
        if o in ["-d", "-q", "-t", "-v"]:
            continue
        elif o in ["-h" , "--help"]:
            print(describeUsage(argv0))
            sys.exit(0)
        else:
            log.warning("Unhandled option: %s %s" % (
                o, a))
    
    if len(containerNames) != 1:
        log.error("Expecting exactly 1 container name, not %d." %
              len(containerNames))
        print(describeUsage(argv0), file=sys.stderr, flush=True)
        sys.exit(1)
     
    containerName = containerNames[0]
    
    confFileName = "/etc/lxcutils/conf/%s.conf" % containerName
    if not os.path.exists(confFileName):
        log.error("No container conf file %s" %
              confFileName)
        sys.exit(1)
        
    errorsFileName = "/etc/lxcutils/conf/%s.ERRORS" % containerName
    if os.path.exists(errorsFileName):
        log.error("Found ERROR file from check: %s" %
              errorsFileName)
        sys.exit(1)

    warningsFileName = "/etc/lxcutils/conf/%s.WARNINGS" % containerName
    okFileName = "/etc/lxcutils/conf/%s.checked_OK" % containerName
    if os.path.exists(warningsFileName):
        if os.path.getmtime(warningsFileName) <= os.path.getmtime(confFileName):
            log.error("Warning file %s\nis older than conf file %s, aborting." % (
                warningsFileName, confFileName))
            sys.exit(1)
        choice = input("WARNING: Found WARNING file from check: %s\n"
                       "Do you want to ignore this and continue bootstrapping container %s?\n"
                       "Enter 'yes' to approve, anything else to abort: " % (
                        warningsFileName, containerName))
        if choice != "yes":
            log.error("You chose to abort boostrapping %s because of %s" % (
                containerName, warningsFileName))
            sys.exit(1)

    if not os.path.exists(okFileName):
        choice = input("WARNING: No OK file found from check, no %s\n"
                       "Do you want to ignore this and continue bootstrapping container %s?\n"
                       "Enter 'yes' to approve, anything else to abort: " % (
                        okFileName, containerName))
        if choice != "yes":
            log.error("You chose to abort boostrapping %s because of missing %s" % (
                containerName, okFileName))
            sys.exit(1)
    elif os.path.getmtime(okFileName) <= os.path.getmtime(confFileName):
        log.error("OK file %s\nis older than conf file %s, aborting." % (
            okFileName, confFileName))
        sys.exit(1)

    if os.path.exists("/var/lib/lxc/%s" % containerName):
        log.error("Container directory /var/lib/lxc/%s already exists, aborting." % 
            containerName)
        sys.exit(1)

    lcuContainer = lcuConfiguration.getCheckedContainer(containerName)
    
    if lcuContainer == None:
        log.error("Failed to get checked container conf to bootstrap container %s, aborting." % (
            containerName))
        sys.exit(1)

    osTemplateFile = "/var/lib/lxcutils/templates/%s.tar.gz" % lcuContainer.ostemplate
    if not os.path.exists(osTemplateFile):
        log.error("OS template file missing to bootstrap container %s: %s." % (
            containerName, osTemplateFile))
        sys.exit(1)

    log.notice("Creating container #%d %s (%s.%s) ..." % (
        lcuContainer.id, lcuContainer.name, lcuContainer.hostname, lcuContainer.domain))

    mounthelpers.createRootfs(lcuContainer=lcuContainer)

    rootfsMountDir = "/var/lib/lxc/%s/rootfs" % lcuContainer.name
    
    os.chdir(rootfsMountDir)
    
    log.notice("Extracting %s in %s for LCU/LXC container %s" % (
             osTemplateFile, rootfsMountDir, lcuContainer.name))
    cmdParts = [ "tar", "-xzf", osTemplateFile ]
    cp = subprocess.run(cmdParts)
    if cp.returncode != 0:
        problem = "ERROR: Failed to extract %s in %s for LCU/LXC container %s" % (
             osTemplateFile, rootfsMountDir, lcuContainer.name)
        helpers.backcheckOnProblem(rc=cp.returncode, cmdParts=cmdParts, problem=problem)
    else:
        log.good("Sucessfully extracted %s in %s for LCU/LXC container %s" % (
             osTemplateFile, rootfsMountDir, lcuContainer.name))

    # df -h and ask user?
    
    lxcFileName = "/etc/lxcutils/conf/%s.lxc" % containerName

    ### continue lcu2-bootstrap:410ff
    lxcConf = ""
    lxcConf += "# %s\n" % lxcFileName
    now = datetime.now()
    lxcConf += "# created by lxcutils 3.x lcu-bootstrap on %s\n" % now.strftime("%Y-%m-%d %H:%M:%S")
    lxcConf += "# based on %s\n" % confFileName 
    lxcConf += "\n"
    lxcConf += "# distro common conf\n"
    if lcuContainer.ostemplate.startswith("focal"):
        lxcConf += "lxc.include = /usr/share/lxc/config/ubuntu.common.conf\n" 
    else:
        lxcConf += "lxc.include = /usr/share/lxc/config/debian.common.conf\n" 
    lxcConf += "\n" 
    lxcConf += "# lxcutils common conf\n"
    lxcConf += "lxc.include = /etc/lxcutils/conf/lcu-common.lxc\n" 
    lxcConf += "lxc.rootfs.path = %s\n" % rootfsMountDir
    lxcConf += "\n" 
    lxcConf += "lxc.uts.name = %s.%s\n" % (lcuContainer.hostname, lcuContainer.domain) 
    lxcConf += "\n" 
    lxcConf += "#lxc.log.level = 2\n" 
    lxcConf += "lxc.log.file = /var/log/lxc/%s.log\n" % containerName
    lxcConf += "lxc.console.logfile = /var/log/lxc/%s.console\n" % containerName
    lxcConf += "\n" 
    lxcConf += "# lxc3 with systemd in container\n"
    lxcConf += "# see https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=920643\n"
    lxcConf += "# not sure wether this is automatically pulled in from /etc/lxc/default.conf anyway in up-to-time installations\n" 
    lxcConf += "lxc.apparmor.profile = generated\n" 
    lxcConf += "lxc.apparmor.allow_nesting = 1\n" 
    lxcConf += "\n" 
    lxcConf += "# idmap-ing for unpriviledged containers\n"
    if lcuContainer.targetUid > 0:
        lxcConf += "lxc.idmap = u 0 %d 65536\n" % lcuContainer.targetUid
        lxcConf += "lxc.idmap = g 0 %d 65536\n" % lcuContainer.targetUid
    elif lcuContainer.id > 999 and lcuContainer.id < 40000:
        lxcConf += "#lxc.idmap = u 0 %d00000 65536\n" % lcuContainer.id
        lxcConf += "#lxc.idmap = g 0 %d00000 65536\n" % lcuContainer.id
    else:
        lxcConf += "#lxc.idmap = u 0 nnnn00000 65536\n" 
        lxcConf += "#lxc.idmap = g 0 nnnn00000 65536\n"
    lxcConf += "\n" 
    ctMacIntro = ""
    if lcuContainer.id > 999 and lcuContainer.id < 10000:
        ctMacIntro = "00:0C:29:%02d:%02d" % (int(lcuContainer.id / 100), lcuContainer.id % 100)
        log.debug("Computed MAC intro from container ID %d to %s" % (lcuContainer.id, ctMacIntro))
    else:
        log.debug("Container ID %d not suitable for auto-computing a MAC intro")
    lxcConf += "\n"
    if (lcuContainer.checkedIp0 != None) or (lcuContainer.checkedIpSix0 != None):
        lxcConf += "# eth0\n"
        if lcuContainer.nettype0 == "ipvlan":
            lxcConf += "#lxc.net.0.type = veth\n"
            lxcConf += "lxc.net.0.type = ipvlan\n"
            lxcConf += "lxc.net.0.ipvlan.mode = l2\n"
        else:
            lxcConf += "lxc.net.0.type = veth\n"
            lxcConf += "#lxc.net.0.type = ipvlan\n"
            lxcConf += "#lxc.net.0.ipvlan.mode = l2\n"
        if lcuContainer.mtu0 > 0:
            lxcConf += "lxc.net.0.mtu = %d\n" % lcuContainer.mtu0
        elif lcuConfiguration.checkedCtMtu == None:
            lxcConf += "lxc.net.0.mtu = 1500\n" 
        else:
            lxcConf += "lxc.net.0.mtu = %d\n" % lcuConfiguration.checkedCtMtu 
        lxcConf += "lxc.net.0.name = eth0\n" 
        if len(lcuContainer.link0) > 0:
            lxcConf += "lxc.net.0.link = %s\n" % lcuContainer.link0
        else:  
            lxcConf += "lxc.net.0.link = br0\n"
        if lcuContainer.nettype0 == "ipvlan":
            lxcConf += "#" # disable whatever hwaddr would be set below
        if len(lcuContainer.mac0) > 0:
            lxcConf += "lxc.net.0.hwaddr = %s\n" % lcuContainer.mac0
        elif len(ctMacIntro) > 0:
            lxcConf += "lxc.net.0.hwaddr = %s:00\n" % ctMacIntro
        else:
            lxcConf += "lxc.net.0.hwaddr = 00:0C:29:NN:NN:00\n"
        if lcuContainer.checkedIp0 == None:
            lxcConf += "#lxc.net.0.ipv4.address = 10.2.3.4/24\n#" # that last # is intentional, to disable anything we write as ipv4.gateway
        else:
            lxcConf += "lxc.net.0.ipv4.address = %s/%d\n" % (lcuContainer.checkedIp0.exploded, lcuConfiguration.dmzMaskBits)
        if lcuConfiguration.checkedDmzGateway == None:
            lxcConf += "#lxc.net.0.ipv4.gateway = ??\n" 
        elif lcuContainer.checkedIp1 == None:
            lxcConf += "lxc.net.0.ipv4.gateway = %s\n" % lcuConfiguration.checkedDmzGateway.exploded
        else:
            lxcConf += "#lxc.net.0.ipv4.gateway = %s\n" % lcuConfiguration.checkedDmzGateway.exploded
        lxcConf += "lxc.net.0.flags = up\n"
        if lcuContainer.checkedIpSix0 != None:
            lxcConf += "lxc.net.0.ipv6.address = %s/%d\n" % (lcuContainer.checkedIpSix0.compressed, lcuConfiguration.checkedWan6Network.prefixlen)
            if lcuConfiguration.checkedWan6Gateway != None:
                lxcConf += "lxc.net.0.ipv6.gateway = %s\n" % lcuConfiguration.checkedWan6Gateway.compressed
            else:
                lxcConf += "#lxc.net.0.ipv6.gateway = ??\n"
        elif lcuConfiguration.checkedWan6Network != None:
            lxcConf += "#lxc.net.0.ipv6.address = %sxx/%d\n" % (lcuConfiguration.checkedWan6Network.compressed, lcuConfiguration.checkedWan6Network.prefixlen)
            if lcuConfiguration.checkedWan6Gateway != None:
                lxcConf += "#lxc.net.0.ipv6.gateway = %s\n" % lcuConfiguration.checkedWan6Gateway.compressed
            else:
                lxcConf += "#lxc.net.0.ipv6.gateway = ??\n"
        else:
            lxcConf += "#lxc.net.0.ipv6.address = ??/64\n"
            lxcConf += "#lxc.net.0.ipv6.gateway = ??\n"
        lxcConf += "\n" 
    if (lcuContainer.checkedIp1 != None) or (lcuContainer.checkedIpSix1 != None):
        lxcConf += "# eth1\n"
        if lcuContainer.nettype1 == "ipvlan":
            lxcConf += "#lxc.net.1.type = veth\n"
            lxcConf += "lxc.net.1.type = ipvlan\n"
            lxcConf += "lxc.net.1.ipvlan.mode = l2\n"
        else:
            lxcConf += "lxc.net.1.type = veth\n"
            lxcConf += "#lxc.net.1.type = ipvlan\n"
            lxcConf += "#lxc.net.1.ipvlan.mode = l2\n"
        if lcuContainer.mtu1 > 0:
            lxcConf += "lxc.net.1.mtu = %d\n" % lcuContainer.mtu1
        elif lcuConfiguration.checkedCtMtu == None:
            lxcConf += "lxc.net.1.mtu = 1500\n" 
        else:
            lxcConf += "lxc.net.1.mtu = %d\n" % lcuConfiguration.checkedCtMtu 
        lxcConf += "lxc.net.1.name = eth1\n" 
        if len(lcuContainer.link1) > 0:
            lxcConf += "lxc.net.1.link = %s\n" % lcuContainer.link1
        else:  
            lxcConf += "lxc.net.1.link = br1\n"
        if lcuContainer.nettype1 == "ipvlan":
            lxcConf += "#" # disable whatever hwaddr would be set below
        if len(lcuContainer.mac1) > 0:
            lxcConf += "lxc.net.1.hwaddr = %s\n" % lcuContainer.mac1
        elif len(ctMacIntro) > 0:
            lxcConf += "lxc.net.1.hwaddr = %s:01\n" % ctMacIntro
        else:
            lxcConf += "lxc.net.1.hwaddr = 00:0C:29:NN:NN:01\n"
        if lcuContainer.checkedIp1 == None:
            lxcConf += "#lxc.net.1.ipv4.address = 10.2.3.4/24\n#" # that last # is intentional, to disable anything we write as ipv4.gateway
        else:
            lxcConf += "lxc.net.1.ipv4.address = %s/%d\n" % (lcuContainer.checkedIp1.exploded, lcuConfiguration.wanMaskBits)
        if lcuConfiguration.checkedWanGateway == None:
            lxcConf += "#lxc.net.1.ipv4.gateway = ??\n" 
        else:
            lxcConf += "lxc.net.1.ipv4.gateway = %s\n" % lcuConfiguration.checkedWanGateway.exploded
        lxcConf += "lxc.net.1.flags = up\n"
        if lcuContainer.checkedIpSix1 != None:
            lxcConf += "lxc.net.1.ipv6.address = %s/%d\n" % (lcuContainer.checkedIpSix1.compressed, lcuConfiguration.checkedWan6Network.prefixlen)
            if lcuConfiguration.checkedWan6Gateway != None:
                lxcConf += "lxc.net.1.ipv6.gateway = %s\n" % lcuConfiguration.checkedWan6Gateway.compressed
            else:
                lxcConf += "#lxc.net.1.ipv6.gateway = ??\n"
        elif lcuConfiguration.checkedWan6Network != None:
            lxcConf += "#lxc.net.1.ipv6.address = %sxx/%d\n" % (lcuConfiguration.checkedWan6Network.compressed, lcuConfiguration.checkedWan6Network.prefixlen)
            if lcuConfiguration.checkedWan6Gateway != None:
                lxcConf += "#lxc.net.1.ipv6.gateway = %s\n" % lcuConfiguration.checkedWan6Gateway.compressed
            else:
                lxcConf += "#lxc.net.1.ipv6.gateway = ??\n"
        else:
            lxcConf += "#lxc.net.1.ipv6.address = ??/64\n"
            lxcConf += "#lxc.net.1.ipv6.gateway = ??\n"
        lxcConf += "\n" 
    if (lcuContainer.checkedIp2 != None) or (lcuContainer.checkedIpSix2 != None):
        lxcConf += "# eth2\n"
        if lcuContainer.nettype2 == "ipvlan":
            lxcConf += "#lxc.net.2.type = veth\n"
            lxcConf += "lxc.net.2.type = ipvlan\n"
            lxcConf += "lxc.net.2.ipvlan.mode = l2\n"
        else:
            lxcConf += "lxc.net.2.type = veth\n"
            lxcConf += "#lxc.net.2.type = ipvlan\n"
            lxcConf += "#lxc.net.2.ipvlan.mode = l2\n"
        if lcuContainer.mtu2 > 0:
            lxcConf += "lxc.net.2.mtu = %d\n" % lcuContainer.mtu2
        elif lcuConfiguration.checkedCtMtu == None:
            lxcConf += "lxc.net.2.mtu = 1500\n" 
        else:
            lxcConf += "lxc.net.2.mtu = %d\n" % lcuConfiguration.checkedCtMtu 
        lxcConf += "lxc.net.2.name = eth2\n" 
        if len(lcuContainer.link2) > 0:
            lxcConf += "lxc.net.2.link = %s\n" % lcuContainer.link2
        else:  
            lxcConf += "lxc.net.2.link = br2\n"
        if lcuContainer.nettype2 == "ipvlan":
            lxcConf += "#" # disable whatever hwaddr would be set below
        if len(lcuContainer.mac2) > 0:
            lxcConf += "lxc.net.2.hwaddr = %s\n" % lcuContainer.mac2
        elif len(ctMacIntro) > 0:
            lxcConf += "lxc.net.2.hwaddr = %s:02\n" % ctMacIntro
        else:
            lxcConf += "lxc.net.2.hwaddr = 00:0C:29:NN:NN:02\n"
        if lcuContainer.checkedIp2 == None:
            lxcConf += "#lxc.net.2.ipv4.address = 10.2.3.4/24\n#" # that last # is intentional, to disable anything we write as ipv4.gateway
        else:
            lxcConf += "lxc.net.2.ipv4.address = %s/%d\n" % (lcuContainer.checkedIp2.exploded, lcuConfiguration.dmzMaskBits)
        lxcConf += "#lxc.net.2.ipv4.gateway = ??\n" 
        lxcConf += "lxc.net.2.flags = up\n"             
        if lcuContainer.checkedIpSix2 != None:
            lxcConf += "lxc.net.2.ipv6.address = %s/%d\n" % (lcuContainer.checkedIpSix2.compressed, lcuConfiguration.checkedWan6Network.prefixlen)
            if lcuConfiguration.checkedWan6Gateway != None:
                lxcConf += "lxc.net.2.ipv6.gateway = %s\n" % lcuConfiguration.checkedWan6Gateway.compressed
            else:
                lxcConf += "#lxc.net.2.ipv6.gateway = ??\n"
        elif lcuConfiguration.checkedWan6Network != None:
            lxcConf += "#lxc.net.2.ipv6.address = %sxx/%d\n" % (lcuConfiguration.checkedWan6Network.compressed, lcuConfiguration.checkedWan6Network.prefixlen)
            if lcuConfiguration.checkedWan6Gateway != None:
                lxcConf += "#lxc.net.2.ipv6.gateway = %s\n" % lcuConfiguration.checkedWan6Gateway.compressed
            else:
                lxcConf += "#lxc.net.2.ipv6.gateway = ??\n"
        else:
            lxcConf += "#lxc.net.2.ipv6.address = ??/64\n"
            lxcConf += "#lxc.net.2.ipv6.gateway = ??\n"
        lxcConf += "\n" 
    lxcConf += "\n" 

    print(lxcConf, file=io.open(lxcFileName,"w"))
    log.notice("Created LXC file %s" %
          lxcFileName)
    
    if not os.path.islink("/var/lib/lxc/%s/config" % containerName):
        log.notice("Auto-creating /var/lib/lxc/%s/config" %
              containerName)
        os.symlink("/etc/lxcutils/conf/%s.lxc" % containerName, "/var/lib/lxc/%s/config" % containerName)
        if not os.path.islink("/var/lib/lxc/%s/config" % containerName):
            log.error("Aborting after failure to auto-create /var/lib/lxc/%s/config" %
                  containerName)
            sys.exit(1)

    if os.path.isdir("%s/etc/apt" % rootfsMountDir):
        ctAptProxyFileName = "%s/etc/apt/apt.conf.d/99_proxy" % rootfsMountDir
        envProxy = os.environ.get("PROXY", "")
        if  len(envProxy) > 0:
            print('Acquire { http { Proxy "%s"; } }' %
                  envProxy, file=io.open(ctAptProxyFileName,"w"))
            log.notice("Auto-created %s with $PROXY=%s" % (
                ctAptProxyFileName, envProxy))
        elif os.path.exists("/etc/apt/apt.conf.d/99_proxy"):
            shutil.copy("/etc/apt/apt.conf.d/99_proxy", ctAptProxyFileName)
            log.notice("Copied /etc/apt/apt.conf.d/99_proxy to %s" %
                  ctAptProxyFileName)
        else:
            log.warning("Not created %s" %
                  ctAptProxyFileName)

    resolvConf = ""
    if len(lcuContainer.domain) > 0:
        resolvConf += "domain %s\n" % lcuContainer.domain
    if len(lcuContainer.searchdomains) > 0:
        resolvConf += "search %s\n" % lcuContainer.searchdomains
    if len(lcuContainer.dnsList) > 0:
        for dns in lcuContainer.dnsList.split(" "):
            resolvConf += "nameserver %s\n" % dns
    resolvConfFileName = "%s/etc/resolv.conf" % rootfsMountDir
    if len(resolvConf) > 0:
        print(resolvConf, file=io.open(resolvConfFileName,"w"))
        log.notice("Auto-created %s" %
              resolvConfFileName)
    else:
        log.warning("Not created %s" %
              resolvConfFileName)
    
    hostsFileContent = ""
    if lcuContainer.checkedIp1 != None:
        hostsFileContent += "\n%s %s.%s %s\n" % (
            lcuContainer.checkedIp1.exploded, lcuContainer.hostname, lcuContainer.domain, lcuContainer.hostname)
    if lcuContainer.checkedIp0 != None:
        hostsFileContent += "\n%s %s.%s %s\n" % (
            lcuContainer.checkedIp0.exploded, lcuContainer.hostname, lcuContainer.domain, lcuContainer.hostname)
    if lcuContainer.checkedIpSix1 != None:
        hostsFileContent += "\n%s %s.%s %s\n" % (
            lcuContainer.checkedIpSix1.compressed, lcuContainer.hostname, lcuContainer.domain, lcuContainer.hostname)
    if lcuContainer.checkedIpSix0 != None:
        hostsFileContent += "\n%s %s.%s %s\n" % (
            lcuContainer.checkedIpSix0.compressed, lcuContainer.hostname, lcuContainer.domain, lcuContainer.hostname)
    hostsFileName = "%s/etc/hosts" % rootfsMountDir
    if len(hostsFileContent) > 0:
        print(hostsFileContent, file=io.open(hostsFileName,"a"))

    hostnameContent = "%s.%s" % (lcuContainer.hostname, lcuContainer.domain)
    hostnameFileName = "%s/etc/hostname" % rootfsMountDir
    print(hostnameContent, file=io.open(hostnameFileName,"w"))
    log.notice("Auto-created %s" %
          hostnameFileName)
    
    if lcuContainer.targetUid > 0:
        log.notice("Finally shifting uid/gid of finished container %s to base %d" % (
            containerName, lcuContainer.targetUid))
        
        fshelpers.changeUidGidRecursive(path=rootfsMountDir, shiftby=lcuContainer.targetUid, minid=lcuContainer.targetUid, maxid=lcuContainer.targetUid+65536)
        log.good("Successfully shifted uid/gid of finished container %s to base %d" % (
            containerName, lcuContainer.targetUid))
    else:
        log.notice("Not shifting uid/gid of finished container %s to empty or unusable base %d00000" % (
            containerName, lcuContainer.id))

    now = datetime.now()
    log.good("Finished creating LXC container %s at %s" % (
        containerName, now.strftime("%Y-%m-%d %H:%M:%S")))
    log.info("Next step is to\n  lcu-start %s\nand fill the CT."
          % containerName)
    
    sys.exit(0)
