'''
  $Id$
  
  Copyleft ITEG IT-Engineers GmbH 2021
'''

import getopt
# import os
import sys
# import configparser

from iteg.lxcutils import lculogging

# from clazzes.pbadownload import util
from iteg.lxcutils import dto
from iteg.lxcutils import networkhelpers


log = lculogging.getLogger()


def describeUsage(argv0):
    usage = ""
    usage += "Usage: %s [options...]\n" % argv0
    usage += "\n"
    usage += "Checks the LCU environment file /etc/lxcutils/environment and\n"
    usage += "all LCU container's conf file if everything looks good and consistent.\n"
    usage += "\n"
    usage += "By default only errors are printed, because some installations\n"
    usage += "trigger lots of warnings that are unavoidable for those setups.\n"
    usage += lculogging.getCommonUsageOptions()
    usage += "  Specific options are:\n"
    usage += "    -a        Check containers too.\n"
    usage += "    -n        Print notices too (notices are weaker than warnings).\n"
    usage += "    -w        Print warnings and notices too.\n"
    return usage
    

def run_lcu_check_configs(argv0, args):
    
    try:
        opts, garbage = getopt.getopt(args, "adhnqtvw", ["help"])
    except getopt.GetoptError as err:
        print(err, file=sys.stderr)  # will print something like "option -a not recognized"
        print(describeUsage(argv0), file=sys.stderr)
        sys.exit(64)

    lcuConfiguration = dto.LCUConfiguration()

    checkContainerToo = False
    printNotices = False
    printWarnings = False
    
    for o, a in opts:
        if o in ["-d", "-q", "-t", "-v"]:
            continue
        elif o in ["-h" , "--help"]:
            print(describeUsage(argv0))
            sys.exit(0)
        elif o in ["-a"]:
            checkContainerToo = True
        elif o in ["-n"]:
            printNotices = True
        elif o in ["-w"]:
            printWarnings = True
            printNotices = True
        else:
            log.warning("Unhandled option: %s %s" % (o, a))
    if len(garbage) > 0:
        log.warning("Unhandled options: %s" % " ".join(garbage))

    errors = []
    warnings = []
    notices = []
    
    log.info("Checking LCU environment, i.e. /etc/lxcutils/environment")
    lcuConfiguration.checkEnvironment(checkLcuContainersToo=checkContainerToo)
    networkhelpers.checkLCUEnvironmentIPsReachable(lcuConfiguration)
    errors.extend(lcuConfiguration.checkErrors)
    warnings.extend(lcuConfiguration.checkWarnings)
    notices.extend(lcuConfiguration.checkNotices)
    if len(lcuConfiguration.containers) == 0:
        log.warning("No LCU container confs to check, i.e. no /etc/lxcutils/conf/*.conf")
    else:
        for containerName in sorted(lcuConfiguration.containers.keys()):
            lcuContainer = lcuConfiguration.containers[containerName]
            log.info("Checking LCU container conf from %s" %
                     lcuContainer.confFileName)
            lcuContainer.checkConf(lcuConfiguration=lcuConfiguration)
            networkhelpers.checkLCUContainerNameResolvable(lcuContainer)
            if lcuContainer.dnsList != lcuConfiguration.defaultDnsList:
                networkhelpers.checkLCUContainerEnvironmentIPsReachable(lcuContainer)
            errors.extend(lcuContainer.checkErrors)
            warnings.extend(lcuContainer.checkWarnings)
    
    if (len(errors) + len(warnings) + len(notices)) == 0:
        log.good("No errors or warnings or notices detected.")
        sys.exit(0)

    if len(notices) > 0:
        if printNotices:
            log.notice("Detected %d notices:\n%s\n" % (
                len(notices), "\n".join(notices)))
        else:
            log.notice("Detected %d notices (printing suppressed)." % 
                  len(notices))

    if len(warnings) > 0:
        if printWarnings:
            log.warning("Detected %d warnings:\n%s\n" % (
                len(warnings), "\n".join(warnings)))
        else:
            log.warning("Detected %d warnings (printing suppressed)." % 
                  len(warnings))

    if len(errors) > 0:
        log.error("Detected %d errors:\n%s\n" % (
            len(errors), "\n".join(errors)))
        sys.exit(1)
    
    # test logging
    ''' 
    print("Pre-final log.verbosity=%d" % log.verbosity, file=sys.stdout, flush=True)
    log.debug("Dummy debug message")
    log.info("Dummy info message")
    log.notice("Dummy notice message")
    log.good("Dummy good message")
    log.warning("Dummy warning message")
    log.error("Dummy error message")
    '''

    sys.exit(0)
