'''
  $Id$
  
  Copyleft ITEG IT-Engineers GmbH 2021
'''

import getopt
#import io
import lxc
#import os
#import os.path
#import re
#import subprocess
import sys

from iteg.lxcutils import dto
from iteg.lxcutils import lcuhelpers
from iteg.lxcutils import lculogging

log = lculogging.getLogger()


def describeUsage(argv0):
    usage = ""
    usage += "Usage: %s [options...] command [cmdOptions...]\n" % argv0
    usage += "\n"
    usage += "Preforms the specifed command in all running LCU containers.\n"
    usage += lculogging.getCommonUsageOptions()
    usage += "  Specific options are:\n"
    usage += "    -g grex   Grep printed stuff for 'grex', case insensitive.\n"
    usage += "              Use with great care, grepping is done after command completion\n"
    usage += "              and all the command's stdout/stderr output is collected in RAM.\n"
    usage += "    -l        Use lxcObject.attach_wait(). This is the default now.\n"
    usage += "    -L        Use legacy approach calling lxc-attach.\n"
    usage += "    -p        Pause after each container.\n"
    usage += "    -w secs   Wait max. this many seconds, defaults to 0 (forever).\n"
    return usage


def run_lcu_exec_in_all(argv0,args):
    
    try:
        opts, cmdParts = getopt.getopt(args, "dg:hlLpqtvw:", ["help"])
    except getopt.GetoptError as err:
        print(err, file=sys.stderr) # will print something like "option -a not recognized"
        print(describeUsage(argv0), file=sys.stderr)
        sys.exit(64)

    lcuConfiguration = dto.LCUConfiguration()
    useLxcObject = True
    grepFor = ""
    pauseAfterEach = False
    waitMaxSeconds = -1
    
    for o, a in opts:
        if o in ["-d", "-q", "-t", "-v"]:
            continue
        elif o in ["-h" , "--help"]:
            print(describeUsage(argv0))
            sys.exit(0)
        elif o in ["-g"]:
            grepFor = str(a)
        elif o in ["-l"]:
            useLxcObject = True
        elif o in ["-L"]:
            useLxcObject = False
        elif o in ["-p"]:
            pauseAfterEach = True
        elif o in ["-w"]:
            waitMaxSeconds = int(a)
        else:
            log.warning("Unhandled option: %s %s" %
                  (o, a))
    
    if len(cmdParts) < 1:
        log.error("Got no command to perform.")
        print(describeUsage(argv0), file=sys.stderr, flush=True)
        sys.exit(1)
     
    knownCTs = lxc.list_containers(active=True, as_object=True)
    if len(knownCTs) == 0:
        log.warning("No LXC containers configured.")
        sys.exit(0)
    
    runningCTNames = []
    for knownCT in knownCTs:
        if knownCT.running:
            runningCTNames.append(knownCT.name)

    nrRunningCTs = len(runningCTNames) 
    if nrRunningCTs == 0:
        if log.verbosity >= 3:
            log.warning("No LXC containers seem to be running.")
        sys.exit(0)

    rcSum = 0
    for runningCTName in runningCTNames:
        print("", file=sys.stdout) # empty line
        log.notice("Calling for container %s" % runningCTName)
        lcuContainer = lcuConfiguration.getCheckedContainer(runningCTName)
        lxcContainer = dto.LXCContainer(lcuContainer.name, lxcObject=lxc.Container(name=runningCTName))
        rc = lcuhelpers.execInContainer(lxcContainer=lxcContainer, cmdParts=cmdParts,
                                        useLxcObject=useLxcObject, grepFor=grepFor, waitMaxSeconds=waitMaxSeconds)
        rcSum += rc
        log.notice("Done for container %s\n" % runningCTName)
        if pauseAfterEach:
            try:
                input("Press enter to continue with next container.")
            except:
                print("\n", file=sys.stderr, flush=True)
                log.warning("Aborting after container %s" % runningCTName)
                sys.exit(1)
            log.notice("Continuing ...")

    log.debug("Exiting with rcSum %d" % rcSum)
    sys.exit(rcSum)
