'''
  $Id$
  
  Copyleft ITEG IT-Engineers GmbH 2020

  based on script from 
  https://unix.stackexchange.com/questions/156477/migrate-an-unprivileged-lxc-container-between-users
'''

import getopt
import os
import sys

from iteg.lxcutils import dto
from iteg.lxcutils import fshelpers
from iteg.lxcutils import lculogging

log = lculogging.getLogger()


def describeUsage(argv0):
    usage = ""
    usage += "Usage: %s [options...] rootfsPath targetUid\n" % argv0
    usage += "\n"
    usage += "Shifts the specified rootfs to the specified base uid/gid.\n"
    usage += "\n"
    usage += "The target range is 65536 uids/gids long.\n"
    usage += "\n"
    usage += "Warnings about files not beeing (re)shifted are normal because\n"
    usage += "hardlinks are checked as every directory entry they are present as.\n"
    usage += lculogging.getCommonUsageOptions()
    usage += "  Specific options are:\n"
    usage += "    -r        Re-shift recursive even if rootfsPath has correct uid/gid.\n"
    return usage


def run_lcu_shift_ownership(argv0,args):

    try:
        opts, posParams = getopt.getopt(args, "dhqrtv", ["help"])
    except getopt.GetoptError as err:
        print(err, file=sys.stderr) # will print something like "option -a not recognized"
        print(describeUsage(argv0), file=sys.stderr)
        sys.exit(64)

    dto.LCUConfiguration()
    forceRecursive = False

    for o, a in opts:
        if o in ["-d", "-q", "-t", "-v"]:
            continue
        elif o in ["-h" , "--help"]:
            print(describeUsage(argv0))
            sys.exit(0)
        elif o in ["-r"]:
            forceRecursive = True
        else:
            log.warning("Unhandled option: %s %s" %
                  (o, a))

    if len(posParams) < 1:
        log.error("Parameters missing, need exactly 2 positional parameters but got %d\n" %
              len(posParams))
        print(describeUsage(argv0), file=sys.stderr)
        sys.exit(1)

    base_path = posParams[0]
    if not os.path.exists(base_path):
        log.error("Path '%s' does not exist" %
              base_path)
        sys.exit(1)

    if not posParams[1].isnumeric():
        log.error("Target (U)ID '%s' not numerical" %
              posParams[0])
        sys.exit(1)

    idmap_start_target = int(posParams[1])
    if idmap_start_target < 0:
        log.error("Target (U)ID '%s' negative" %
              idmap_start_target)
        sys.exit(1)

    stat_info = os.lstat(base_path)
    idmap_start_before = stat_info.st_uid
    idmap_shiftby = idmap_start_target - idmap_start_before

    if idmap_shiftby == 0:
        if forceRecursive:
            log.notice("Target '%s' already has (U)ID %d, recursing anyway due to -r" % (
                base_path, idmap_start_target))
        else:
            log.good("Target '%s' already has (U)ID %d, nothing to do" % (
                base_path, idmap_start_target))
            sys.exit(0)

    idmap_size = 65536
    idmap_end_before = idmap_start_before + idmap_size
    idmap_end_target = idmap_start_target + idmap_size

    choice = input("OK to shift '%s' from %d..%d by %d to %d..%d?\nEnter 'y' or 'yes' to approve, anything else to abort: " % (
                   base_path, idmap_start_before, idmap_end_before, idmap_shiftby, idmap_start_target, idmap_end_target))
    if choice != "y" and choice != "yes":
        log.error("User chose to abort")
        sys.exit(1)

    log.notice("Actually shifting '%s' from %d..%d by %d to %d..%d ..." % (
        base_path, idmap_start_before, idmap_end_before, idmap_shiftby, idmap_start_target, idmap_end_target))

    fshelpers.changeUidGidRecursive(path=base_path, shiftby=idmap_shiftby, minid=idmap_start_target, maxid=idmap_end_target)
    sys.exit(0)

