'''
  $Id: lcu_start.py 129 2021-02-09 16:01:11Z cl_yubikey_2018a $
  
  Copyleft ITEG IT-Engineers GmbH 2021
'''

import getopt
#import io
import lxc
import os
#import os.path
import shutil
import subprocess
import sys

from iteg.lxcutils import dto
from iteg.lxcutils import lculogging
from iteg.lxcutils import mounthelpers
from iteg.lxcutils import networkhelpers

log = lculogging.getLogger()


def describeUsage(argv0):
    usage = ""
    usage += "Usage: %s [options...] containerName\n" % argv0
    usage += "\n"
    usage += "Starts the specified LCU container after mounting it's rootfs.\n"
    usage += lculogging.getCommonUsageOptions()
    usage += "  Specific options are:\n"
    usage += "    -l        Use lxcObject.start() instead of lxc-start.\n"
    usage += "              This might become the default without notice.\n"
    return usage


def run_lcu_start(argv0,args):
    
    try:
        opts, containerNames = getopt.getopt(args, "dhlqtv", ["help"])
    except getopt.GetoptError as err:
        print(err, file=sys.stderr) # will print something like "option -a not recognized"
        print(describeUsage(argv0), file=sys.stderr)
        sys.exit(64)

    lcuConfiguration = dto.LCUConfiguration()
    useLxcObject = False
    
    for o, a in opts:
        if o in ["-d", "-q", "-t", "-v"]:
            continue
        elif o in ["-h" , "--help"]:
            print(describeUsage(argv0))
            sys.exit(0)
        elif o in ["-l"]:
            useLxcObject = True
        else:
            log.warning("Unhandled option: %s %s" % (
                o, a))
    
    if len(containerNames) != 1:
        log.error("Expecting exactly 1 container name, not %d." %
              len(containerNames))
        print(describeUsage(argv0), file=sys.stderr, flush=True)
        sys.exit(1)
     
    containerName = containerNames[0]
    
    if not os.path.exists("/etc/lxcutils/conf/%s.conf" % containerName):
        log.error("No container conf file /etc/lxcutils/conf/%s.conf" %
              containerName)
        sys.exit(1)
        
    if not os.path.exists("/etc/lxcutils/conf/%s.lxc" % containerName):
        log.error("No container .lxc file /etc/lxcutils/conf/%s.lxc" %
              containerName)
        sys.exit(1)

    #networkhelpers.checkLCUEnvironmentIPsReachable(lcuConfiguration)
    lcuContainer = lcuConfiguration.getCheckedContainer(containerName)
    #networkhelpers.checkLCUContainerNameResolvable(lcuContainer)
    
    lxcContainer = dto.LXCContainer(lcuContainer.name, lxcObject=lxc.Container(name=containerName))
    if lxcContainer == None:
        log.error("Unable to find LXC container for LCU container %s" %
              containerName)
        sys.exit(1)
    
    if lxcContainer.lxcObject == None:
        log.error("Unable to find LXC object for LCU container %s" %
              containerName)
        sys.exit(1)
    
    if lxcContainer.lxcObject.running:
        log.error("LXC container %s is running already" %
              containerName)
        sys.exit(1)

    lcuContainer.checkErrors = []
    lcuContainer.checkWarnings = []
    lcuContainer.checkNotices = []
    nrReachableIPs = networkhelpers.checkLCUContainerIPsAsFree(lcuContainer)
    if nrReachableIPs > 0:
        log.error("Aborting because %d IPs of %s's IPs are pingable, from another machine or a former run:\n%s" % (
            nrReachableIPs, containerName, "\n".join(lcuContainer.checkErrors)))
        sys.exit(1)

    lxcStartExtraOptions = [] 
    consoleLogFilename = lxcContainer.lxcFileDict.get("lxc.console.logfile", "")
    if len(consoleLogFilename) == 0:
        consoleLogFilename = "/var/log/lxc/%s.console" % containerName
        log.notice("lxc file has no lxc.console.logfile configured, falling back to %s" %
              consoleLogFilename)
        lxcStartExtraOptions.extend(["-L", consoleLogFilename])
    else:
        log.notice("lxc.console.logfile is configured as %s" %
              consoleLogFilename)
    if os.path.exists(consoleLogFilename):
        log.notice("Renaming console logfile %s to %s.former" % (
            consoleLogFilename, consoleLogFilename))
        shutil.move(consoleLogFilename, consoleLogFilename+".former")

    rootfsMountDir = "/var/lib/lxc/%s/rootfs" % containerName;
    if not os.path.exists(rootfsMountDir):
        log.notice("Auto-creating %s" % rootfsMountDir)
        os.makedirs(rootfsMountDir, 0o700);
        if not os.path.exists(rootfsMountDir):
            log.error("Aborting after failure to auto-create %s" % rootfsMountDir)
            sys.exit(1)

    if not os.path.exists("/var/lib/lxc/%s/config" % containerName):
        log.notice("Auto-creating /var/lib/lxc/%s/config" %
              containerName)
        os.symlink("/etc/lxcutils/conf/%s.lxc" % containerName, "/var/lib/lxc/%s/config" % containerName)
        if not os.path.islink("/var/lib/lxc/%s/config" % containerName):
            log.error("Aborting after failure to auto-create /var/lib/lxc/%s/config" %
                  containerName)
            sys.exit(1)

    log.notice("Making sure %s's rootfs is mounted ..." %
          containerName)
    mounthelpers.mountRootfs(lcuContainer, lxcContainer)

    if useLxcObject:
        # this variant works since removing useInit=True
        log.notice("Starting LXC container %s with lxcObject.start() ..." %
              containerName)
        #lxcContainer.lxcObject.start(useInit=True, daemonize=True, close_fds=False)
        lxcContainer.lxcObject.start(daemonize=True, close_fds=False)
        # if not ok:
        #    log.error("Failed to start container %s with lxcObject.start(), take a look above and at %s" % (
        #        containerName, consoleLogFilename))
        #    sys.exit(1)
        log.good("Successfully started LXC container %s with lxcObject.start()" %
              containerName)
    else:
        log.notice("Starting LXC container, lxc-start -d -n %s ..." %
              containerName)
        cp = subprocess.run(["/usr/bin/lxc-start", "-d", "-n", containerName]) # extend(lxcStartExtraOptions))
        if cp.returncode != 0:
            log.error("Failed to lxc-start -d -n %s, take a look above and at %s" % (
                containerName, consoleLogFilename))
            sys.exit(1)
        log.good("Successfully started LXC container %s with lxc-start -d" %
              containerName)

    sys.exit(0)
