'''
  $Id$
  
  Copyleft ITEG IT-Engineers GmbH 2021
'''

#import fileinput
#import json
import re
import subprocess
import sys

from iteg.lxcutils import dto
from iteg.lxcutils import lculogging


log = lculogging.getLogger()


def detectCurrentContainerResources(lxcContainer:dto.LXCContainer=None):
    if lxcContainer == None:
        raise Exception("ERROR: detectCurrentContainerResources() called with lxcContainer None")
    if len(lxcContainer.name) == 0:
        raise Exception("ERROR: detectCurrentContainerResources() called with namelesslxcContainer")
        
    try:
        cp = subprocess.run(["/usr/bin/lxc-info","-H","-n",lxcContainer.name], 
                        stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True)
    except:
        log.error("Failed to execute lxc-info for container [%s]" % lxcContainer.name)
        return;
    if cp.returncode != 0:
        log.error("Failed to execute lxc-info for container [%s], got bad returncode %d" % (lxcContainer.name, cp.returncode) )
        return;
    
    '''
    # lxc-info -H -n vs
    Name:           vs
    State:          RUNNING
    PID:            9960
    IP:             192.168.1.2
    CPU use:        1065729162959
    Memory use:     110596096
    KMem use:       18595840
    Link:           vethUPF72O
    TX bytes:       2147985
    RX bytes:       180064748
    Total bytes:    182212733
    '''
    lxcContainer.pid = 0
    lxcContainer.cpuUse = 0
    lxcContainer.memUse = 0
    lxcContainer.kMemUse = 0
    lxcContainer.txBytes = 0
    lxcContainer.rxBytes = 0
    lxcContainer.blkIoUse = 0
    
    lineSplitter = re.compile("^ *([A-Z][A-Za-z ]+): +(.*)$")
    for line in cp.stdout.splitlines():
        m = lineSplitter.match(line)
        if m is None:
            log.debug("Ignoring line [%s]" % line)
            continue
        resName = m.group(1)
        resValue = m.group(2)
        #log.debug("Analysig container [%s]\'s resource [%s] with value [%s]" % (lxcContainer.name, resName, resValue))
        if resName == "State" and resValue != "RUNNING":
            log.error("Cannot detect resources of container [%s] in non-running state [%s]" % (lxcContainer.name, resValue) )
            return
        if resName == "PID":
            lxcContainer.pid = int(resValue)
            continue
        if resName == "CPU use":
            lxcContainer.cpuUse = int(int(resValue)/1000000000)
            continue
        if resName == "Memory use":
            lxcContainer.memUse = int(resValue)
            continue
        if resName == "KMem use":
            lxcContainer.kMemUse = int(resValue)
            continue
        if resName == "TX bytes":
            lxcContainer.txBytes += int(resValue)
            continue
        if resName == "RX bytes":
            lxcContainer.rxBytes += int(resValue)
            continue
        if resName == "BlkIO use":
            lxcContainer.blkIoUse += int(resValue)
            continue
        #log.debug("Ignoring container [%s]\'s resource [%s] with value [%s]" % (lxcContainer.name, resName, resValue))


def execInContainer(lxcContainer:dto.LXCContainer=None, 
                    cmdParts:list=[], 
                    useLxcObject:bool=False, 
                    grepFor:str="",
                    waitMaxSeconds:int=-1,
                    teeFileName:str=""):
    if lxcContainer == None:
        raise Exception("ERROR: execInContainer() called without lxcContainer")
    if len(cmdParts) == 0:
        raise Exception("ERROR: execInContainer() called without cmdParts")
    rc = 0
    
    if useLxcObject:
        log.debug("Performing %s in LXC container '%s' with lxcObject.attach_wait() and timeout %d ..." % (
            " ".join(cmdParts), lxcContainer.name, waitMaxSeconds))
        
        def runArgs(args=[]):
            rc = 0
            timeout = None
            cp = None
            if waitMaxSeconds > 0:
                timeout = waitMaxSeconds
            try:
                if (len(grepFor) + len(teeFileName)) == 0:
                    #log.debug("execInContainer(useLxcObject): Not collecting output.") 
                    rc = subprocess.call(args, timeout=timeout)
                else:
                    #log.debug("execInContainer(useLxcObject): Collecting output.") 
                    cp = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True, timeout=timeout)
                    rc = cp.returncode
            except subprocess.TimeoutExpired:
                log.debug("TIMEOUT: LXC container %s did not finish '%s' within %d seconds." % (
                    lxcContainer.name, " ".join(cmdParts), waitMaxSeconds))
                if rc == 0:
                    rc = 1
            if ((len(grepFor) + len(teeFileName)) == 0) or cp == None:
                #log.debug("execInContainer(useLxcObject): Done early.") 
                return rc
            teeFile = None
            if len(teeFileName) > 0:
                #log.debug("execInContainer(useLxcObject): Teeing to '%s'." % teeFileName) 
                teeFile = open(teeFileName, "w")
            if len(grepFor) > 0:
                #log.debug("execInContainer(useLxcObject): Grepping for '%s'." % grepFor) 
                pattern = re.compile(pattern=grepFor, flags=re.IGNORECASE)
                for catchedLine in cp.stdout.split("\n"):
                    if pattern.search(catchedLine):
                        print(catchedLine, file=sys.stdout)
                        if teeFile != None:
                            print(catchedLine, file=teeFile)
            else:
                #log.debug("execInContainer(useLxcObject): Not grepping but still teeing") 
                for catchedLine in cp.stdout.split("\n"):
                    print(catchedLine, file=sys.stdout)
                    if teeFile != None:
                        print(catchedLine, file=teeFile)
            sys.stdout.flush()
            if teeFile != None:
                teeFile.close()
            return rc
        
        #rc = lxcContainer.lxcObject.console()
        rc = lxcContainer.lxcObject.attach_wait(runArgs, cmdParts) >> 8
        '''
        # Produces "TypeError: attach: object not callable:"
        rc = lxcContainer.lxcObject.attach_wait(runArgs(cmdParts))
        # is executed in host:
        rc = lxcContainer.lxcObject.attach_wait(lxc.attach_run_command(cmdParts))
        rc = lxcContainer.lxcObject.attach_wait(lxc.attach_run_command(cmdParts[0]), cmdParts)
        #rc = lxcContainer.lxcObject.attach_wait(lxc.attach_run_command(cmdParts[0]), cmdParts[1:])
        # Produces "TypeError: attach: object not callable:"
        rc = lxcContainer.lxcObject.attach_wait(cmdParts[0]) # , cmdParts[1:]
        # Produces TypeError: attach_run_shell() takes 0 positional arguments but 1 was given:
        rc = lxcContainer.lxcObject.attach_wait(lxc.attach_run_shell, cmdParts)
        # WORKS
        rc = lxcContainer.lxcObject.attach_wait(lxc.attach_run_command, cmdParts) >> 8
        # WORKS
        rc = lxcContainer.lxcObject.attach_wait(runArgs, cmdParts) >> 8
        '''
        log.debug("Back from performing %s in LXC container %s with lxcObject.attach_wait() with rc=%d" % (
            " ".join(cmdParts), lxcContainer.name, rc))
    else:
        if waitMaxSeconds > 0:
            log.warning("Timeout parameter -w ignored for legacy lxc-attach approach.")
        log.debug("Performing command in LXC container with\n  lxc-attach -n %s -- %s" % (
            lxcContainer.name, " ".join(cmdParts)))
        fullCmd = []
        fullCmd.extend(["/usr/bin/lxc-attach", "-n", lxcContainer.name, "--"])
        fullCmd.extend(cmdParts)
        log.debug("Performing %s" %
            " ".join(fullCmd))
        if (len(grepFor) + len(teeFileName)) == 0:
            #log.debug("execInContainer(lxc-attach): Not collecting output.") 
            cp = subprocess.run(fullCmd)
        else:
            #log.debug("execInContainer(lxc-attach): Collecting output.") 
            cp = subprocess.run(fullCmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
            teeFile = None
            if len(teeFileName) > 0:
                #log.debug("execInContainer(lxc-attach): Teeing to '%s'." % teeFileName) 
                teeFile = open(teeFileName, "w")
            if len(grepFor) > 0:
                #log.debug("execInContainer(lxc-attach): Grepping for '%s'." % grepFor) 
                pattern = re.compile(grepFor, re.IGNORECASE)
                for catchedLine in cp.stdout.split("\n"):
                    if pattern.search(catchedLine):
                        print(catchedLine, file=sys.stdout)
                        if teeFile != None:
                            print(catchedLine, file=teeFile)
            else:
                #log.debug("execInContainer(lxc-attach): Not grepping but still teeing") 
                for catchedLine in cp.stdout.split("\n"):
                    print(catchedLine, file=sys.stdout)
                    if teeFile != None:
                        print(catchedLine, file=teeFile)
            sys.stdout.flush()
            if teeFile != None:
                teeFile.close()
        rc = cp.returncode
        log.debug("Back from\n  lxc-attach -n %s -- %s\nwith rc %d" % (
            lxcContainer.name, " ".join(cmdParts), rc))
    
    return rc





