'''
  $Id$
  
  Copyleft ITEG IT-Engineers GmbH 2021
'''

#import exceptions
import fileinput
#import io
#import json
import os
#import re
import shutil
import socket
import subprocess
import sys
import time

from iteg.lxcutils import dto
from iteg.lxcutils import helpers
from iteg.lxcutils import lculogging
from iteg.lxcutils import iscsihelpers
#from iteg.lxcutils.dto import LCUConfiguration

if os.path.exists("/usr/lib/python3/dist-packages/iteg/freenasutils"):
    #from iteg.freenasutils.dto import FNUConfiguration 
    #from iteg.freenasutils.dto import NasDataset, NasInfo
    #from iteg.freenasutils.dto import NasISCSIExtent, NasISCSITarget, NasISCSITargetExtentAssignment
    from iteg.freenasutils.restclient import FNURestClient

log = lculogging.getLogger()


'''
  Unmounting rootfs
'''

def umountRootfs(lcuContainer:dto.LCUContainer=None, lxcContainer:dto.LCUContainer=None):
    '''
        checks for and performs eventual umount work for lxc.rootfs.path of a LCU/LXC container
        if nothing is mounted on that path, nothing is done
    ''' 
    if lcuContainer == None:
        raise Exception("ERROR: mountRootfs() called with lcuContainer None")
    if lxcContainer == None:
        raise Exception("ERROR: mountRootfs() called with lxcContainer None")
    # roorfs
    rootfsMountDir = lxcContainer.lxcFileDict.get("lxc.rootfs.path", "")
    if len(rootfsMountDir) == 0:
        rootfsMountDir = "/var/lib/lxc/%s/rootfs" % lxcContainer.name
    if not os.path.ismount(rootfsMountDir):
        log.notice("rootfs mount directory %s not mounted, nothing to umount for container %s" % (
            rootfsMountDir, lcuContainer.name))
        #unlockContainer(lcuContainer)
        #return True
    else:
        log.notice("Unmounting %s for LCU/LXC container %s" %
                 (rootfsMountDir, lcuContainer.name))
        #try:
        cp = subprocess.run(["/bin/umount", "-v", rootfsMountDir])
        #except:
        #    #log.err("ERROR: Failed to umount %s for LCU/LXC container %s" %
        #    #     (rootfsMountDir, lcuContainer.name))
        #    raise Exception("ERROR: Failed to umount %s for LCU/LXC container %s" %
        #         (rootfsMountDir, lcuContainer.name))
        if cp.returncode != 0:
            raise Exception("ERROR: Failed to unmount %s for LCU/LXC container %s" % (
                 rootfsMountDir, lcuContainer.name))
        log.good("Sucessfully unmounted %s for LCU/LXC container %s" % (
                 rootfsMountDir, lcuContainer.name))
    # CRYPTTYPE?
    if len(lcuContainer.crypttype) > 0:
        log.debug("umountRootfs(%s): Calling _undecryptRootfsDev()" % lcuContainer.name)
        _undecryptRootfsDev(lcuContainer)
    else:
        log.debug("umountRootfs(%s): No need to call _undecryptRootfsDev()" % lcuContainer.name)
    # iscsi logout?
    if len(lcuContainer.nasdataset) > 0:
        log.notice("Logging out from ISCSI session for %s for LCU/LXC container %s" % (
            rootfsMountDir, lcuContainer.name))
        # fill iscsi volatiles
        log.debug("umountRootfs(%s): Calling _checkRootfsIscsi(performLogin=False) only to collect iscsi infos" % lcuContainer.name)
        _checkRootfsIscsi(lcuContainer=lcuContainer, performLogin=False)
        # logout
        log.debug("umountRootfs(%s): Calling iscsihelpers.performIscsiLogouts()" % lcuContainer.name)
        iscsihelpers.performIscsiLogouts(basename=lcuContainer.iscsiBasename, portalListeners=lcuContainer.iscsiPortalListeners, targetName=lcuContainer.iscsiName)
    else:
        log.debug("umountRootfs(%s): No need to call performIscsiLogout()" % lcuContainer.name)
    
    #if haveContainerLock(lcuContainer):
    unlockContainer(lcuContainer)

    return True


'''
  Mounting rootfs
'''

def mountRootfs(lcuContainer:dto.LCUContainer=None, lxcContainer:dto.LCUContainer=None):
    '''
        checks for and performs eventual mounts work for lxc.rootfs.path of a LCU/LXC container
        lcuContainer MUST be checked, i.e. semantically valid in regard to mount options in .conf file!
        if something is already mounted on that rootfs path nothing is done.
        on failure sys.exit is called.
    ''' 
    if lcuContainer == None:
        raise Exception("ERROR: mountRootfs() called with lcuContainer None")
    if lxcContainer == None:
        raise Exception("ERROR: mountRootfs() called with lxcContainer None")

    # rootfs
    rootfsMountDir = lxcContainer.lxcFileDict.get("lxc.rootfs.path", "")
    if len(rootfsMountDir) == 0:
        rootfsMountDir = "/var/lib/lxc/%s/rootfs" % lxcContainer.name
    if not os.path.exists(rootfsMountDir):
        log.warning("Auto-creating rootfs mount directory %s for container %s" % (
            rootfsMountDir, lcuContainer.name))
        os.makedirs(rootfsMountDir, 0o700)
    if not os.path.isdir(rootfsMountDir):
        raise Exception("ERROR: mountRootfs(%s): rootfs %s missing or not a directory" % (lcuContainer.name, rootfsMountDir))
    # already mounted?
    if os.path.ismount(rootfsMountDir):
        log.notice("rootfs mount directory %s already mounted for container %s, returning right away" % (
            rootfsMountDir, lcuContainer.name))
        return True
    # check lock
    currentLockContent = readContainerLock(lcuContainer=lcuContainer)
    if len(currentLockContent) > 0:
        if haveContainerLock(lcuContainer=lcuContainer):
            log.warning("mountRootfs(): We already have the lock for container %s, continueing" % lcuContainer.name)
        else:
            log.error("mountRootfs(): Container %s is locked by someone else, lock file contains:\n%s\n" % (
                lcuContainer.name, currentLockContent))
            sys.exit(1)
    else:
        log.info("Locking container %s" % lcuContainer.name)
        lockContainer(lcuContainer)
    # detect mount method
    lcuContainer.checkedRootfsDevPath = ""
    if len(lcuContainer.lvgroup) > 0:
        _checkRootfsLvm(lcuContainer=lcuContainer)
    elif len(lcuContainer.lfdir) > 0:
        _checkRootfsLoopFile(lcuContainer=lcuContainer)
    elif len(lcuContainer.nasdataset) > 0:
        _checkRootfsIscsi(lcuContainer=lcuContainer, performLogin=True)
    else:
        # evtl. check fstab?
        log.warning("mountRootfs(%s): Neither CTLV* nor CTLF* nor CTNASNAME set, nothing to do, or is there?" % 
              lcuContainer.name)
        return
    if len(lcuContainer.checkedRootfsDevPath) == 0:
        log.error("mountRootfs(%s): Failed to detect rootfsDevPath to mount" %
              lcuContainer.name)
        sys.exit(1)

    # CRYPTTYPE?
    finalRootfsMountDevPath = lcuContainer.checkedRootfsDevPath
    if len(lcuContainer.crypttype) > 0:
        finalRootfsMountDevPath = _decryptRootfsDev(lcuContainer)
    # mount ....
    log.notice("Mounting %s on %s for LCU/LXC container %s" %
             (finalRootfsMountDevPath, rootfsMountDir, lcuContainer.name))
    #try:
    cp = subprocess.run(["/bin/mount", "-t","ext4", "-o","defaults,errors=remount-ro,acl", "-v", finalRootfsMountDevPath, rootfsMountDir])
    #except:
    #    #log.err("ERROR: Failed to mount %s on %s for LCU/LXC container %s" %
    #    #     (rootfsDevPath, rootfsMountDir, lcuContainer.name))
    #    raise Exception("ERROR: Failed to mount %s on %s for LCU/LXC container %s" %
    #         (rootfsDevPath, rootfsMountDir, lcuContainer.name))
    if cp.returncode != 0:
        log.error("Failed to mount %s on %s for LCU/LXC container %s" %
             (finalRootfsMountDevPath, rootfsMountDir, lcuContainer.name))
        sys.exit(1)
    log.good("Sucessfully mounted %s on %s for LCU/LXC container %s" %
             (finalRootfsMountDevPath, rootfsMountDir, lcuContainer.name))
    return True


def _checkRootfsLvm(lcuContainer:dto.LCUContainer=None, wantMissing:bool=False): # lxcContainer:dto.LCUContainer=None
    '''
      computes the LVM-based lcuContainer.rootfsDevPath 
      and checks if VG and LVM are present,
      on success lcuContainer.rootfsDevPath is set and returned
      on failure sys.exit is called
    '''
    if lcuContainer == None:
        raise Exception("ERROR: _checkRootfsLvm() called with lcuContainer None")
    #if lxcContainer == None:
    #    raise Exception("ERROR: _checkRootfsLvm() called with lxcContainer None")
    if len(lcuContainer.lvgroup) == 0:
        log.error("_checkRootfsLvm(%s): lcuContainer has no LVM group configured" % 
              lcuContainer.name)
        sys.exit(1)
    if not os.path.isdir("/dev/%s" % lcuContainer.lvgroup):
        log.error("_checkRootfsLvm(%s): LVM group %s does not exist, no /dev/%s" % (
            lcuContainer.name, lcuContainer.lvgroup, lcuContainer.lvgroup))
        sys.exit(1)
    if len(lcuContainer.lvname) == 0:
        log.error("_checkRootfsLvm(%s): lcuContainer has no LVM LV configured" % 
              lcuContainer.name)
        sys.exit(1)
    rootfsDevPath = "/dev/%s/%s" % (lcuContainer.lvgroup, lcuContainer.lvname)
    if wantMissing: 
        if os.path.islink(rootfsDevPath):
            log.error("_checkRootfsLvm(%s): VG %s already has a LV %s, found %s" % (
                lcuContainer.name, lcuContainer.lvgroup, lcuContainer.lvname, rootfsDevPath))
            sys.exit(1)
    elif not os.path.islink(rootfsDevPath):
        log.error("_checkRootfsLvm(%s): VG %s has no LV %s, no %s" % (
            lcuContainer.name, lcuContainer.lvgroup, lcuContainer.lvname, rootfsDevPath))
        sys.exit(1)
    log.debug("_checkRootfsLvm(%s): Found good rootfsDevPath %s" % (
            lcuContainer.name, rootfsDevPath))
    lcuContainer.checkedRootfsDevPath = rootfsDevPath
    return rootfsDevPath

def _checkRootfsLoopFile(lcuContainer:dto.LCUContainer=None, wantMissing:bool=False): # lxcContainer:dto.LCUContainer=None
    '''
      computes the LF-based lcuContainer.rootfsDevPath 
      and checks if the .ext4 file is present.
      on success lcuContainer.rootfsDevPath is set and returned
      on failure sys.exit is called
    '''
    if lcuContainer == None:
        raise Exception("ERROR: _checkRootfsLoopFile() called with lcuContainer None")
    if len(lcuContainer.lfdir) == 0:
        log.error("_checkRootfsLoopFile(%s): lcuContainer has no CTLFDIR configured" % 
              lcuContainer.name)
        sys.exit(1)
    if not os.path.isdir(lcuContainer.lfdir):
        log.error("_checkRootfsLoopFile(%s): CTLFDIR %s does not exist, no %s" % (
            lcuContainer.name, lcuContainer.lfdir, lcuContainer.lfdir))
        sys.exit(1)
    if len(lcuContainer.lfname) == 0:
        log.error("_checkRootfsLoopFile(%s): lcuContainer has no CTLFNAME configured" % 
              lcuContainer.name)
        sys.exit(1)
    rootfsDevPath = "%s/%s.ext4" % (lcuContainer.lfdir, lcuContainer.lfname)
    if wantMissing:
        if os.path.isfile(rootfsDevPath):
            log.error("_checkRootfsLoopFile(%s): CTLFDIR/CTLFNAME %s already exists" % (
                lcuContainer.name, rootfsDevPath))
            sys.exit(1)
    elif not os.path.isfile(rootfsDevPath):
        log.error("_checkRootfsLoopFile(%s): CTLFDIR/CTLFNAME %s does not exist" % (
              lcuContainer.name, rootfsDevPath))
        sys.exit(1)
    log.debug("_checkRootfsLoopFile(%s): Found good rootfsDevPath %s" % (
            lcuContainer.name, rootfsDevPath))
    lcuContainer.checkedRootfsDevPath = rootfsDevPath
    return rootfsDevPath

def _checkRootfsIscsi(lcuContainer:dto.LCUContainer=None, wantMissing:bool=False, performLogin:bool=False): # lxcContainer:dto.LCUContainer=None
    '''
      computes the CTNASDATASET-based lcuContainer.rootfsDevPath 
      and checks if the dataset is present.
      on success with wantMissing==False lcuContainer.rootfsDevPath is set and returned
      on success with wantMissing==True lcuContainer.rootfsDevPath is UNDEFINED and "" is returned
      on failure sys.exit is called
    '''
    # on success set lcuContainer.ctIscsi*?
    if lcuContainer == None:
        raise Exception("ERROR: _checkRootfsIscsi() called with lcuContainer None")
    if len(lcuContainer.nasdataset) == 0:
        log.error("_checkRootfsIscsi(%s): lcuContainer has no CTNASDATASET configured" % 
              lcuContainer.name)
        sys.exit(1)
    if len(lcuContainer.nasname) == 0:
        log.error("_checkRootfsIscsi(%s): lcuContainer has no CTNASNAME configured, iSCSI support requires FreeNAS for now" % 
              lcuContainer.name)
        sys.exit(1)
    if not os.path.exists("/usr/lib/python3/dist-packages/iteg/freenasutils"):
        log.error("_checkRootfsIscsi(%s): current iSCSI support requires freenasutils installed and configured" % 
              lcuContainer.name)
        sys.exit(1)
    lcuConfiguration = dto.getLCUConfiguration()
    fnuConfiguration = lcuConfiguration.fnuConfiguration
    lcuContainer.fnuNasInfo = fnuConfiguration.nasInfos.get(lcuContainer.nasname)
    # evtl. cache in lcuContainer?
    if lcuContainer.fnuNasInfo == None:
        log.error("_checkRootfsIscsi(%s): CTNASNAME %s not known to freenasutils" % (
            lcuContainer.name, lcuContainer.nasname))
        sys.exit(1)
    fnuRestclient = FNURestClient(fnuConfiguration=fnuConfiguration)
    fnuRestclient.getAllISCSILists(nasInfo=lcuContainer.fnuNasInfo)
    # check if dataset exists
    if not lcuContainer.nasdataset in lcuContainer.fnuNasInfo.nas_datasets_flat.keys():
        if wantMissing:
            log.good("_checkRootfsIscsi(%s, wantMissing=True): No dataset with ID %s, not investigating further" % (
                lcuContainer.name, lcuContainer.nasdataset))
            return ""
        log.error("_checkRootfsIscsi(%s): No dataset with ID %s, not investigating further" % (
            lcuContainer.name, lcuContainer.nasdataset))
        sys.exit(1)
    lcuContainer.iscsiName = lcuContainer.nasdataset.split("/")[-1].lower()
    lcuContainer.fnuDataset = lcuContainer.fnuNasInfo.nas_datasets_flat[lcuContainer.nasdataset]
    lcuContainer.fnuExtent = None
    lcuContainer.fnuTarget = None
    lcuContainer.fnuTargetExtentAssignment = None
    for targetExtentAssignment in lcuContainer.fnuNasInfo.nas_iscsi_targetextentassignments.values():
        extent = lcuContainer.fnuNasInfo.nas_iscsi_extents[targetExtentAssignment.extentId]
        if extent.disk != "zvol/" + lcuContainer.nasdataset:
            continue
        target = lcuContainer.fnuNasInfo.nas_iscsi_targets[targetExtentAssignment.targetId]
        if target.name != lcuContainer.iscsiName:
            continue
        lcuContainer.fnuTargetExtentAssignment = targetExtentAssignment
        lcuContainer.fnuExtent = extent
        lcuContainer.fnuTarget = target
        break
    if lcuContainer.fnuTargetExtentAssignment == None:
        if wantMissing:
            log.warning("_checkRootfsIscsi(%s, wantMissing=True): No iSCSI target-extent-assigment for dataset %s, not investigating further" % (
                lcuContainer.name, lcuContainer.nasdataset))
            return ""
        log.error("_checkRootfsIscsi(%s): No iSCSI target-extent-assigment for dataset %s, not investigating further" % (
            lcuContainer.name, lcuContainer.nasdataset))
        sys.exit(1)
    # check if already logged in
    lcuContainer.fnuPortal = lcuContainer.fnuNasInfo.nas_iscsi_portals[lcuContainer.fnuTarget.portalId]
    lcuContainer.iscsiPortalListeners = []
    if len(lcuContainer.isportal) > 0:
        lcuContainer.iscsiPortalListeners = lcuContainer.isportal.split(",")
    nrPortalListeners = len(lcuContainer.iscsiPortalListeners)
    if nrPortalListeners == 0:
        if len(lcuContainer.fnuPortal.listeners) == 0:
            log.error("_checkRootfsIscsi(%s): Container has no CTISPORTAL configured and FNU-detected portal has no listeners, not investigating further" % (
                lcuContainer.name))
            sys.exit(1)
        lcuContainer.iscsiPortalListeners = lcuContainer.fnuPortal.listeners
        log.info("_checkRootfsIscsi(%s): Container has no CTISPORTAL configured, using FNU-detected portal listeners: '%s'" % (
            lcuContainer.name, "', '".join(lcuContainer.iscsiPortalListeners)))
        nrPortalListeners = len(lcuContainer.iscsiPortalListeners)
    lcuContainer.iscsiBasename = lcuConfiguration.iscsiBasename
    if len(lcuContainer.iscsiBasename) == 0:
        lcuContainer.iscsiBasename = lcuContainer.fnuNasInfo.nas_iscsi_global.basename
    # preliminary
    #if len(lcuContainer.iscsiPortals) > 1:
    rootfsDevPathCandidates = []
    for portalListener in lcuContainer.iscsiPortalListeners:
        # /dev/disk/by-path/ip-192.168.1.2:3260-iscsi-iqn.2011-05.org.clazzes:testvol-lun-0
        rootfsDevPath = "/dev/disk/by-path/ip-%s-iscsi-%s:%s-lun-%d" % (
            portalListener, lcuContainer.iscsiBasename, lcuContainer.iscsiName, lcuContainer.fnuTargetExtentAssignment.lunid)
        log.debug("_checkRootfsIscsi(%s): Computed potential rootfsDevPath to %s" % (
            lcuContainer.name, rootfsDevPath))
        if os.path.islink(rootfsDevPath):
            if wantMissing:
                log.error("_checkRootfsIscsi(%s, wantMissing=True): Found %s (not checking eventual further iscsi multipaths)" % (
                    lcuContainer.name, rootfsDevPath))
                sys.exit(1)
            else:
                log.debug("rootfsDevPath %s present already" % rootfsDevPath)
                if nrPortalListeners == 1:
                    log.good("_checkRootfsIscsi(%s): rootfsDevPath %s present and only 1 portal listener, done for good" % (
                        lcuContainer.name, rootfsDevPath))
                    lcuContainer.checkedRootfsDevPath = rootfsDevPath
                    return rootfsDevPath
            rootfsDevPathCandidates.append(rootfsDevPath)
            log.info("_checkRootfsIscsi(%s): rootfsDevPath %s present, but more than 1 portal listener to check" % (
                lcuContainer.name, rootfsDevPath))
        # check discovery
        if wantMissing or (not performLogin):
            discoveredLine = iscsihelpers.discoverIscsiTarget(basename=lcuContainer.iscsiBasename, portalIPPort=portalListener, targetName=lcuContainer.iscsiName)
            if len(discoveredLine) == 0:
                if wantMissing:
                    log.good("_checkRootfsIscsi(%s, wantMissing=True): Discovered NO target on portal %s for basename %s, targetname %s" % (
                        lcuContainer.name, portalListener, lcuContainer.iscsiBasename, lcuContainer.iscsiName))
                    return ""
                log.error("_checkRootfsIscsi(%s): Portal %s does not seem to have target for basename %s, targetname %s" % (
                    lcuContainer.name, portalListener, lcuContainer.iscsiBasename, lcuContainer.iscsiName))
                sys.exit(1)
            if wantMissing:
                log.error("_checkRootfsIscsi(%s, wantMissing=True): Discovered target on portal %s for basename %s, targetname %s" % (
                    lcuContainer.name, portalListener, lcuContainer.iscsiBasename, lcuContainer.iscsiName))
                sys.exit(1)
            if nrPortalListeners == 1:
                log.good("_checkRootfsIscsi(%s): Portal %s has target for basename %s, targetname %s, and no auto-login requested and only 1 portal listener" % (
                    lcuContainer.name, portalListener, lcuContainer.iscsiBasename, lcuContainer.iscsiName))
                lcuContainer.checkedRootfsDevPath = rootfsDevPath
                return rootfsDevPath
            log.info("_checkRootfsIscsi(%s): Portal %s has target for basename %s, targetname %s, and no auto-login requested, but multiple listeners to check" % (
                    lcuContainer.name, portalListener, lcuContainer.iscsiBasename, lcuContainer.iscsiName))
            if portalListener == lcuContainer.iscsiPortalListeners[-1]:
                log.good("_checkRootfsIscsi(%s): All portals have target for basename %s, targetname %s, and no auto-login requested, done for good but returning incorrect not-multipathed rootfsDevPath '%s'" % (
                    lcuContainer.name, lcuContainer.iscsiBasename, lcuContainer.iscsiName, rootfsDevPath))
                lcuContainer.checkedRootfsDevPath = rootfsDevPath
                return rootfsDevPath
            continue # avoid login attempt
        # try to login
        if not iscsihelpers.checkIscsiLogin(basename=lcuContainer.iscsiBasename, portalIPPort=portalListener, targetName=lcuContainer.iscsiName):
            log.error("_checkRootfsIscsi(%s): Failed to login to portal %s basename %s, targetname %s" % (
                lcuContainer.name, portalListener, lcuContainer.iscsiBasename, lcuContainer.iscsiName))
            sys.exit(1)
        log.notice("_checkRootfsIscsi(%s): Sucessfully logged in to portal %s, basename %s, targetname %s, re-checking path in 3 seconds" % (
            lcuContainer.name, portalListener, lcuContainer.iscsiBasename, lcuContainer.iscsiName))
        time.sleep(3)
        if os.path.islink(rootfsDevPath):
            if nrPortalListeners == 1:
                log.good("_checkRootfsIscsi(%s): rootfsDevPath %s now present and only 1 portal listener, done for good" % (
                    lcuContainer.name, rootfsDevPath))
                lcuContainer.checkedRootfsDevPath = rootfsDevPath
                return rootfsDevPath
            else:
                rootfsDevPathCandidates.append(rootfsDevPath)
                log.info("_checkRootfsIscsi(%s): rootfsDevPath %s now present, but more than 1 portal listener to check" % (
                    lcuContainer.name, rootfsDevPath))
        else:
            log.error("_checkRootfsIscsi(%s): Still no %s after good login to portal %s basename %s, targetname %s" % (
                lcuContainer.name, rootfsDevPath, portalListener, lcuContainer.iscsiBasename, lcuContainer.iscsiName))
            sys.exit(1)
    # if we got here we have multiple listeners and good logins and can and must detect the multipath device
    checkedMultipathLLHeadline = ""
    # give multipath time to find devices
    log.debug("_checkRootfsIscsi(%s): Giving multipathd time to pick up devices")
    time.sleep(3)
    for rootfsDevPathCandidate in rootfsDevPathCandidates:
        candidateDevPathResolved = os.path.realpath(rootfsDevPathCandidate)
        if not candidateDevPathResolved.startswith("/dev/sd"):
            log.error("_checkRootfsIscsi(%s): rootfsDevPathCandidate %s does not point to /dev/sd* but to '%s', aborting" % (
                lcuContainer.name, rootfsDevPathCandidate, candidateDevPathResolved))
            sys.exit(1)
        cmdParts = [ "/sbin/multipath", "-l", candidateDevPathResolved ]
        log.debug("_checkRootfsIscsi(%s): Calling %s" % (
            lcuContainer.name, " ".join(cmdParts)))
        cp = subprocess.run(cmdParts, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
        if cp.returncode != 0:
            log.error("_checkRootfsIscsi(%s): Failed to check rootfsDevPathCandidate, '%s' gave non-0 returncode %d, aborting" % (
                lcuContainer.name, " ".join(cmdParts), cp.returncode))
            sys.exit(1)
        #log.debug("_checkRootfsIscsi(%s): Output of %s was:\n%s\n" % (
        #          lcuContainer.name, " ".join(cmdParts), cp.stdout))
        multipathLLHeadline = cp.stdout.split("\n")[0]
        if len(multipathLLHeadline) == 0:
            log.error("_checkRootfsIscsi(%s): Failed to check rootfsDevPathCandidate, '%s' gave empty response, aborting" % (
                lcuContainer.name, " ".join(cmdParts)))
            sys.exit(1)
        if len(checkedMultipathLLHeadline) == 0:
            checkedMultipathLLHeadline = multipathLLHeadline
            continue
        if checkedMultipathLLHeadline != multipathLLHeadline:
            log.error("_checkRootfsIscsi(%s): %s gave response '%s' different from former response '%s', aborting" % (
                lcuContainer.name, " ".join(cmdParts), multipathLLHeadline, checkedMultipathLLHeadline))
            sys.exit(1)
    log.debug("_checkRootfsIscsi(%s): All calls to multipath -l gave same response: '%s'" % (
        lcuContainer.name, multipathLLHeadline))
    if len(checkedMultipathLLHeadline) == 0:
        log.error("_checkRootfsIscsi(%s): Got no multipath -l response from any of the rootfsDevPathCandidates %s, aborting" % (
            lcuContainer.name, ", ".join(rootfsDevPathCandidates)))
        sys.exit(1)
    '''
      parse
        mpatha (36589cfc00000054116295561807bb031) dm-1 FreeNAS,iSCSI Disk
      or
        36589cfc00000054116295561807bb031 dm-1 FreeNAS,iSCSI Disk
    '''
    multipathUuid = ""
    cmllhlParts = checkedMultipathLLHeadline.split(" ");
    if checkedMultipathLLHeadline.startswith("mpath") and (") dm-" in checkedMultipathLLHeadline) and (len(cmllhlParts) >= 4):
        multipathUuid = cmllhlParts[1][1:-1]
    elif (" dm-" in checkedMultipathLLHeadline) and (len(cmllhlParts) >= 3):
        multipathUuid = cmllhlParts[0]
    if len(multipathUuid) == 0:
        log.error("_checkRootfsIscsi(%s): Failed to find multipathUuid from checkedMultipathLLHeadline '%s', aborting" % (
            lcuContainer.name, checkedMultipathLLHeadline))
        sys.exit(1)
    log.debug("_checkRootfsIscsi(%s): Parsed multipath -l response to multipathUuid '%s'" % (
        lcuContainer.name, multipathUuid))
    mpDevPath = "/dev/disk/by-id/dm-uuid-mpath-%s" % multipathUuid
    if not os.path.islink(mpDevPath):
        log.error("_checkRootfsIscsi(%s): mpDevPath '%s' is not an existing link, aborting" % (
            lcuContainer.name, mpDevPath))
        sys.exit(1)
    rootfsDevPath = os.path.realpath(mpDevPath)
    if len(rootfsDevPath) == 0:
        log.error("_checkRootfsIscsi(%s): mpDevPath '%s' does not point to /dev/dm-* but to '%s', aborting" % (
            lcuContainer.name, mpDevPath, rootfsDevPath))
        sys.exit(1)
    log.good("_checkRootfsIscsi(%s): Successfully found rootfsDevPath for multipath situation, returning %s" % (
        lcuContainer.name, rootfsDevPath))
    lcuContainer.checkedRootfsDevPath = rootfsDevPath
    return rootfsDevPath

'''
  CRYPT helpers
'''

def _decryptRootfsDev(lcuContainer:dto.LCUContainer=None):
    '''
      decrypts the lcuContainer.checkedRootfsDevPath for lcuContainer
      on success the ready-to-mount decrypted device path is returned.
      on failure sys.exit is called.
    '''
    if lcuContainer.crypttype != "luks":
        log.error("_decryptRootfsDev(%s): Unsupported CTCRYPTTYPE '%s'" % (
            lcuContainer.name, lcuContainer.crypttype))
        sys.exit(1)
    decryptedDevBaseName = "%s_decrypted" % lcuContainer.name
    decryptedRootFsDevPath = "/dev/mapper/%s" % decryptedDevBaseName
    if os.path.islink(decryptedRootFsDevPath):
        log.notice("decryptedRootFsDevPath %s already exists, probably nothing luksOpen for container %s" % (
            decryptedRootFsDevPath, lcuContainer.name))
        return decryptedRootFsDevPath
    log.notice("Going to cryptsetup luksOpen %s as %s, please enter LUKS password when requested" % (
        lcuContainer.checkedRootfsDevPath, decryptedDevBaseName))
    cp = subprocess.run(["/sbin/cryptsetup", "-v", "--timeout", "30", "luksOpen", lcuContainer.checkedRootfsDevPath, decryptedDevBaseName])
    if cp.returncode != 0:
        log.error("Failed to cryptsetup luksOpen %s as %s for LCU/LXC container %s" %
             (lcuContainer.checkedRootfsDevPath, decryptedDevBaseName, lcuContainer.name))
        sys.exit(1)
    log.good("Sucessfully decrypted %s as %s resp. %s for LCU/LXC container %s" %
             (lcuContainer.checkedRootfsDevPath, decryptedDevBaseName, decryptedRootFsDevPath, lcuContainer.name))
    return decryptedRootFsDevPath

def _undecryptRootfsDev(lcuContainer:dto.LCUContainer=None):
    '''
      un-decrypts resp. luksCloses the rootfs-backing LUKS device for lcuContainer
      on success the un-decrypted device path is returned.
      on failure sys.exit is called.
    '''
    if lcuContainer.crypttype != "luks":
        log.error("_undecryptRootfsDev(%s): Unsupported CTCRYPTTYPE '%s'" % (
            lcuContainer.name, lcuContainer.crypttype))
        sys.exit(1)
    decryptedDevBaseName = "%s_decrypted" % lcuContainer.name
    decryptedRootFsDevPath = "/dev/mapper/%s" % decryptedDevBaseName
    if not os.path.islink(decryptedRootFsDevPath):
        log.notice("decryptedRootFsDevPath %s does not exist, probably nothing to luksClose for container %s" % (
            decryptedRootFsDevPath, lcuContainer.name))
        return decryptedRootFsDevPath
    log.notice("Going to luksClose %s for container %s" % (
        decryptedDevBaseName, lcuContainer.name))
    cp = subprocess.run(["/sbin/cryptsetup", "-v", "luksClose", decryptedDevBaseName])
    if cp.returncode != 0:
        log.error("Failed to luksClose %s for LCU/LXC container %s" % (
            decryptedDevBaseName, lcuContainer.name))
        sys.exit(1)
    log.good("Sucessfully luksClosed %s for LCU/LXC container %s" % (
        decryptedDevBaseName, lcuContainer.name))
    return decryptedRootFsDevPath

def _encryptRootfsDev(lcuContainer:dto.LCUContainer=None):
    '''
      encrypts the lcuContainer.checkedRootfsDevPath to later bootstrap lcuContainer
      on success the ready-to-mount decrypted device path is returned.
      on failure sys.exit is called.
    '''
    if lcuContainer.crypttype != "luks":
        log.error("_encryptRootfsDev(%s): Unsupported CTCRYPTTYPE '%s'" % (
            lcuContainer.name, lcuContainer.crypttype))
        sys.exit(1)
    decryptedDevBaseName = "%s_decrypted" % lcuContainer.name
    decryptedRootFsDevPath = "/dev/mapper/%s" % decryptedDevBaseName
    if os.path.islink(decryptedRootFsDevPath):
        log.error("decryptedRootFsDevPath %s already exists, not re-formatting for container %s" % (
            decryptedRootFsDevPath, lcuContainer.name))
        sys.exit(1)

    log.notice("Going to cryptsetup luksFormat %s, please enter a LUKS twice when requested" % (
        lcuContainer.checkedRootfsDevPath))
    
    cmdParts = ["/sbin/cryptsetup", "-v", "luksFormat", "-c","aes-xts-plain64", "-s","512", "-h","sha512", "--verify-passphrase", lcuContainer.checkedRootfsDevPath]
    cp = subprocess.run(cmdParts)
    if cp.returncode != 0:
        problem = "ERROR: Failed to cryptsetup luksFormat %s for LCU/LXC container %s" % (
             lcuContainer.checkedRootfsDevPath, lcuContainer.name)
        helpers.backcheckOnProblem(rc=cp.returncode, cmdParts=cmdParts, problem=problem)
    else:
        log.good("Sucessfully cryptsetup luksFormatted %s for LCU/LXC container %s" % (
             lcuContainer.checkedRootfsDevPath, lcuContainer.name))
    
    log.notice("Going to cryptsetup luksOpen %s as %s resp. %s please enter LUKS password when requested" % (
        lcuContainer.checkedRootfsDevPath, decryptedDevBaseName, decryptedRootFsDevPath))
    cmdParts = ["/sbin/cryptsetup", "-v", "--timeout", "60", "luksOpen", lcuContainer.checkedRootfsDevPath, decryptedDevBaseName]
    cp = subprocess.run(cmdParts)
    if cp.returncode != 0:
        problem = "ERROR: Failed to cryptsetup luksOpen %s as %s for LCU/LXC container %s" % (
             lcuContainer.checkedRootfsDevPath, decryptedDevBaseName, lcuContainer.name)
        helpers.backcheckOnProblem(rc=cp.returncode, cmdParts=cmdParts, problem=problem)
    else:
        log.good("Sucessfully cryptsetup luksOpenend %s as %s for LCU/LXC container %s" % (
             lcuContainer.checkedRootfsDevPath, decryptedDevBaseName, lcuContainer.name))

    return decryptedRootFsDevPath

def _resizeCryptedRootfsDev(lcuContainer:dto.LCUContainer=None):
    '''
      resizes the rootfs-backing LUKS device for lcuContainer
      on success the device path is returned.
      on failure sys.exit is called.
    '''
    if lcuContainer == None:
        raise Exception("ERROR: _resizeCryptedRootfsDev() called with lcuContainer None")
    if lcuContainer.crypttype != "luks":
        log.error("_undecryptRootfsDev(%s): Unsupported CTCRYPTTYPE '%s'" % (
            lcuContainer.name, lcuContainer.crypttype))
        sys.exit(1)
    decryptedDevBaseName = "%s_decrypted" % lcuContainer.name
    decryptedRootFsDevPath = "/dev/mapper/%s" % decryptedDevBaseName
    if not os.path.islink(decryptedRootFsDevPath):
        log.notice("decryptedRootFsDevPath %s does not exist for container %s, trying to luksOpen first" % (
            decryptedRootFsDevPath, lcuContainer.name))
        cmdParts = ["/sbin/cryptsetup", "-v", "--timeout", "60", "luksOpen", lcuContainer.checkedRootfsDevPath, decryptedDevBaseName]
        cp = subprocess.run(cmdParts)
        if cp.returncode != 0:
            problem = "ERROR: Failed to cryptsetup luksOpen %s as %s for LCU/LXC container %s" % (
                 lcuContainer.checkedRootfsDevPath, decryptedDevBaseName, lcuContainer.name)
            helpers.backcheckOnProblem(rc=cp.returncode, cmdParts=cmdParts, problem=problem)
        else:
            log.good("Sucessfully cryptsetup luksOpenend %s as %s for LCU/LXC container %s" % (
                 lcuContainer.checkedRootfsDevPath, decryptedDevBaseName, lcuContainer.name))
    if not os.path.islink(decryptedRootFsDevPath):
        log.error("decryptedRootFsDevPath %s still does not exist after luksOpen, hopefully nothing to resize for container %s" % (
            decryptedRootFsDevPath, lcuContainer.name))
        return decryptedRootFsDevPath
    log.notice("Going to resize %s for container %s" % (
        decryptedDevBaseName, lcuContainer.name))
    cp = subprocess.run(["/sbin/cryptsetup", "-v", "--timeout", "60", "resize", decryptedDevBaseName])
    if cp.returncode != 0:
        log.error("Failed to resize %s for LCU/LXC container %s" % (
            decryptedDevBaseName, lcuContainer.name))
        sys.exit(1)
    log.good("Sucessfully resize %s for LCU/LXC container %s" % (
        decryptedDevBaseName, lcuContainer.name))
    return decryptedRootFsDevPath


'''
  Creating rootfs
'''

def createRootfs(lcuContainer:dto.LCUContainer=None): #, lxcContainer:dto.LCUContainer=None):
    '''
        creates the rootfs filesystem the specified container up to and including
        the eventual mkfs.ext4 and mount
        lcuContainer MUST be checked, i.e. semantically valid in regard to mount options in .conf file!
        on failure sys.exit is called.
    ''' 
    if lcuContainer == None:
        raise Exception("ERROR: createRootfs() called with lcuContainer None")
    #if len(lcuContainer.name) == 0:
    #    raise Exception("ERROR: createRootfs() called with lcuContainer with empty name")
    
    # lxcFile is not ready yet, so do NOT DO THIS: rootfsMountDir = lxcContainer.lxcFileDict.get("lxc.rootfs.path", "")
    rootfsMountDir = "/var/lib/lxc/%s/rootfs" % lcuContainer.name

    log.notice("Creating rootfs mount directory %s for container %s" % (
        rootfsMountDir, lcuContainer.name))
    os.makedirs(rootfsMountDir, 0o755)

    lcuContainer.checkedRootfsDevPath = ""
    if len(lcuContainer.lvgroup) > 0:
        _createRootfsLvm(lcuContainer=lcuContainer)
    elif len(lcuContainer.lfdir) > 0:
        _createRootfsLoopFile(lcuContainer=lcuContainer)
    elif len(lcuContainer.nasdataset) > 0:
        _createRootfsIscsi(lcuContainer=lcuContainer)
    else:
        log.warning("createRootfs(%s): Neither CTLV* nor CTLF* nor CTNASNAME set, nothing to do, or is there?" %
              lcuContainer.name)
        return

    if len(lcuContainer.checkedRootfsDevPath) == 0:
        log.error("createRootfs(%s): Failed to detect resp. create rootfsDevPath" %
              lcuContainer.name)
        sys.exit(1)
    log.notice("Successfully created checkedRootfsDevPath %s" %
          lcuContainer.checkedRootfsDevPath)
    
    # crypt stuff
    finalRootfsMountDevPath = lcuContainer.checkedRootfsDevPath
    if len(lcuContainer.crypttype) > 0:
        finalRootfsMountDevPath = _encryptRootfsDev(lcuContainer)

    # continue lcu2-bootstrap:180f
    log.notice("Creating ext4 fs on %s for LCU/LXC container %s" %
             (finalRootfsMountDevPath, lcuContainer.name))
    cmdParts = [ "/sbin/mkfs.ext4", finalRootfsMountDevPath]
    cp = subprocess.run(cmdParts)
    if cp.returncode != 0:
        problem = "ERROR: Failed to mkfs.ext4 %s for LCU/LXC container %s" % (
             finalRootfsMountDevPath, lcuContainer.name)
        helpers.backcheckOnProblem(rc=cp.returncode, cmdParts=cmdParts, problem=problem)
    else:
        log.good("Sucessfully created ext4 fs on %s for LCU/LXC container %s" % (
             finalRootfsMountDevPath, lcuContainer.name))

    log.notice("Mounting %s on %s for LCU/LXC container %s" %
             (finalRootfsMountDevPath, rootfsMountDir, lcuContainer.name))
    cmdParts = ["/bin/mount", "-t","ext4", "-o","defaults,errors=remount-ro,acl", "-v", finalRootfsMountDevPath, rootfsMountDir]
    cp = subprocess.run(cmdParts)
    if cp.returncode != 0:
        problem = "ERROR: Failed to mount %s on %s for LCU/LXC container %s" % (
             finalRootfsMountDevPath, rootfsMountDir, lcuContainer.name)
        helpers.backcheckOnProblem(rc=cp.returncode, cmdParts=cmdParts, problem=problem)
    else:
        log.good("Sucessfully mounted %s on %s for LCU/LXC container %s" % (
             finalRootfsMountDevPath, rootfsMountDir, lcuContainer.name))

    log.notice("createRootfs(%s): Done creating and mounting %s" %
          (lcuContainer.name, rootfsMountDir))
    

def _createRootfsLvm(lcuContainer:dto.LCUContainer=None):
    '''
      creates the LVM for the rootfs of the specified lcuContainer 
      on success the returned device is ready to be mkfs'd and stored in lcuContainer.checkedRootfsDevPath
      on failure sys.exit is called
    '''
    rootfsDevPath = _checkRootfsLvm(lcuContainer=lcuContainer, wantMissing=True)

    # create LV
    log.notice("Going to create LV %s in VG %s with %d GB" % (
        lcuContainer.lvname, lcuContainer.lvgroup, lcuContainer.size))
    cmdParts = [ "/sbin/lvcreate", "--size", "%dG" % lcuContainer.size, "--name", lcuContainer.lvname, lcuContainer.lvgroup ]
    cp = subprocess.run(cmdParts)
    if cp.returncode != 0:
        problem = "Unable to create LV %s in VG %s" % (
            lcuContainer.lvname, lcuContainer.lvgroup)
        helpers.backcheckOnProblem(rc=cp.returncode, cmdParts=cmdParts, problem=problem)
    else:
        log.notice("Successfully created LV %s in VG %s with %d GB" % (
            lcuContainer.lvname, lcuContainer.lvgroup, lcuContainer.size))

    lcuContainer.checkedRootfsDevPath = rootfsDevPath
    return rootfsDevPath

def _createRootfsLoopFile(lcuContainer:dto.LCUContainer=None):
    '''
      creates the .ext4 file for the rootfs of the specified lcuContainer 
      on success the returned device is ready to be mkfs'd and stored in lcuContainer.checkedRootfsDevPath
      on failure sys.exit is called
    '''
    rootfsDevPath = _checkRootfsLoopFile(lcuContainer=lcuContainer, wantMissing=True)

    # create LV
    log.notice("Going to create CTLFile %s with size %d GB" % (
        rootfsDevPath, lcuContainer.size))
    wantedSizeBytes = lcuContainer.size*1073741824
    cmdParts = [ "/usr/bin/truncate", "--size=%d" % wantedSizeBytes, rootfsDevPath ]
    cp = subprocess.run(cmdParts)
    if cp.returncode != 0:
        problem = "Unable to create CTLFile %s" % (
            rootfsDevPath)
        helpers.backcheckOnProblem(rc=cp.returncode, cmdParts=cmdParts, problem=problem)
    else:
        log.notice("Successfully created CTLFile %s with %d GB" % (
            rootfsDevPath, lcuContainer.size))

    lcuContainer.checkedRootfsDevPath = rootfsDevPath
    return rootfsDevPath

def _createRootfsIscsi(lcuContainer:dto.LCUContainer=None):
    '''
      creates the zvol dataset and iscsi target-extent-association for the rootfs of the specified lcuContainer 
      on success the returned device is ready to be mkfs'd and is stored in lcuContainer.checkedRootfsDevPath
      on failure sys.exit is called
    '''
    _checkRootfsIscsi(lcuContainer=lcuContainer, wantMissing=True, performLogin=False)

    lcuConfiguration = dto.getLCUConfiguration()
    fnuConfiguration = lcuConfiguration.fnuConfiguration
    #ctNasInfo = fnuConfiguration.nasInfos.get(lcuContainer.nasname)
    # create zbol dataset
    log.notice("Going to create zvol dataset %s:%s with size %d GB" % (
        lcuContainer.fnuNasInfo.name, lcuContainer.nasdataset, lcuContainer.size))
    fnuRestclient = FNURestClient(fnuConfiguration=fnuConfiguration)
    wantedSizeBytes = lcuContainer.size*1073741824
    fnuRestclient.createDataset(
        nasInfo=lcuContainer.fnuNasInfo,
        datasetPath=lcuContainer.nasdataset,
        volSizeBytes=wantedSizeBytes,
        datasetComment="LXC-Container %s" % lcuContainer.name)
    # create iscsi stuff
    log.notice("Going to create ISCSI parts on %s for %s:%s" % (
        lcuContainer.fnuNasInfo.name, lcuContainer.nasdataset, lcuContainer.size))
    lcuContainer.fnuTarget = fnuRestclient.createISCSITarget(
        nasInfo=lcuContainer.fnuNasInfo, datasetPath=lcuContainer.nasdataset)
    if lcuContainer.fnuTarget == None:
        log.error("_createRootfsIscsi(%s): Failed to create iSCSI target for dataset %s." % (
              lcuContainer.name, lcuContainer.nasdataset))
        sys.exit(1)
    lcuContainer.fnuExtent = fnuRestclient.createISCSIExtent(
        nasInfo=lcuContainer.fnuNasInfo, datasetPath=lcuContainer.nasdataset, comment="LXC-Container %s"%lcuContainer.name)
    if lcuContainer.fnuExtent == None:
        log.error("_createRootfsIscsi(%s): Failed to create iSCSI extent for dataset %s." % (
              lcuContainer.name, lcuContainer.nasdataset))
        sys.exit(1)
    lcuContainer.fnuTargetExtentAssignment = fnuRestclient.createISCSITargetExtent(
        nasInfo=lcuContainer.fnuNasInfo, target=lcuContainer.fnuTarget, extent=lcuContainer.fnuExtent)
    if lcuContainer.fnuTargetExtentAssignment == None:
        log.error("_createRootfsIscsi(%s): Failed to create iSCSI target-extent assignment for dataset %s." % (
              lcuContainer.name, lcuContainer.nasdataset))
        sys.exit(1)
    # give portal time to pick the target up
    log.notice("Giving NAS time to make new ISCSI target available");
    time.sleep(5)
    # login
    log.notice("Going to connect to %s ISCSI target for %s:%s" % (
        lcuContainer.fnuNasInfo.name, lcuContainer.nasdataset, lcuContainer.size))
    rootfsDevPath = _checkRootfsIscsi(lcuContainer=lcuContainer, wantMissing=False, performLogin=True)
    # last check
    if os.path.exists(rootfsDevPath):
        log.good("_createRootfsIscsi(%s): Successfully created & prepared iscsi rootfsDevPath %s" % (
            lcuContainer.name, rootfsDevPath))
        lcuContainer.checkedRootfsDevPath = rootfsDevPath
        return rootfsDevPath
    # or?
    log.error("_createRootfsIscsi(%s): Despite good work unable to create iscsi rootfsDevPath %s." % (
        lcuContainer.name, rootfsDevPath))
    sys.exit(1)


'''
  Destroying rootfs
'''

def destroyRootfs(lcuContainer:dto.LCUContainer=None): #, lxcContainer:dto.LCUContainer=None):
    '''
        destroys the rootfs filesystem (if any) and directory of the specified container and 
        lcuContainer MUST be checked, i.e. semantically valid in regard to mount options in .conf file!
        lcuContainer MUST be STOPPED
        on failure sys.exit is called.
    ''' 
    if lcuContainer == None:
        raise Exception("ERROR: destroyRootfs() called with lcuContainer None")
    
    # lxcFile might be dead already, so do NOT DO THIS: rootfsMountDir = lxcContainer.lxcFileDict.get("lxc.rootfs.path", "")
    rootfsMountDir = "/var/lib/lxc/%s/rootfs" % lcuContainer.name

    #lcuContainer.checkedRootfsDevPath = ""
    if len(lcuContainer.lvgroup) > 0:
        _destroyRootfsLvm(lcuContainer)
    elif len(lcuContainer.lfdir) > 0:
        _destroyRootfsLoopFile(lcuContainer)
    elif len(lcuContainer.nasdataset) > 0:
        _destroyRootfsIscsi(lcuContainer)
    else:
        log.warning("destroyRootfs(%s): Neither CTLV* nor CTLF* nor CTNASNAME set, no device to destroy or is there?" %
              lcuContainer.name)
        #return
    
    if len(lcuContainer.checkedRootfsDevPath) > 0:
        log.good("Successfully destroyed checkedRootfsDevPath %s" %
              lcuContainer.checkedRootfsDevPath)

    if os.path.exists(rootfsMountDir):
        log.notice("Removing rootfsMountDir %s" %
                   rootfsMountDir)
        shutil.rmtree(rootfsMountDir)

    # FUTURE: Evtl. rm backuptarget too?
    
    log.good("Done removing %s" %
          rootfsMountDir)
    

def _destroyRootfsLvm(lcuContainer:dto.LCUContainer=None):
    '''
      destroys the LVM for the rootfs of the specified lcuContainer 
      on success the destroy device is stored to lcuContainer.checkedRootfsDevPath and returned 
      on failure sys.exit is called
    '''
    rootfsDevPath = _checkRootfsLvm(lcuContainer=lcuContainer)

    # destroy LV
    log.notice("Going to lvremove LV %s in VG %s" % (
        lcuContainer.lvname, lcuContainer.lvgroup))
    cmdParts = [ "/sbin/lvremove", rootfsDevPath ]
    cp = subprocess.run(cmdParts)
    if cp.returncode != 0:
        problem = "Unable to lvremove LV %s in VG %s" % (
            lcuContainer.lvname, lcuContainer.lvgroup)
        helpers.backcheckOnProblem(rc=cp.returncode, cmdParts=cmdParts, problem=problem)
    else:
        log.good("Successfully lvremoved LV %s in VG %s" % (
            lcuContainer.lvname, lcuContainer.lvgroup))

    lcuContainer.checkedRootfsDevPath = rootfsDevPath
    return rootfsDevPath

def _destroyRootfsLoopFile(lcuContainer:dto.LCUContainer=None):
    '''
      destroys the .ext4 file for the rootfs of the specified lcuContainer 
      on success the destroyed device is stored to lcuContainer.checkedRootfsDevPath and returned 
      on failure sys.exit is called
    '''
    rootfsDevPath = _checkRootfsLoopFile(lcuContainer=lcuContainer)

    # destroy LV
    log.notice("Going to rm CTLFile %s" %
               rootfsDevPath)
    # os.remove(rootfsDevPath) would work but let's stick to subprocess pattern
    cmdParts = [ "/bin/rm", "-v", rootfsDevPath ]
    cp = subprocess.run(cmdParts)
    if cp.returncode != 0:
        problem = "Unable to rm CTLFile %s" % rootfsDevPath
        helpers.backcheckOnProblem(rc=cp.returncode, cmdParts=cmdParts, problem=problem)
    else:
        log.good("Successfully rm CTLFile %s" %
                 rootfsDevPath)

    lcuContainer.checkedRootfsDevPath = rootfsDevPath
    return rootfsDevPath

def _destroyRootfsIscsi(lcuContainer:dto.LCUContainer=None):
    '''
      destroys the iscsi target-extent-assignment, target, extent, and zvol dataset 
      on success the destroyed device is stored to lcuContainer.checkedRootfsDevPath and is returned 
      on failure sys.exit is called
    '''
    rootfsDevPath = _checkRootfsIscsi(lcuContainer=lcuContainer)

    lcuConfiguration = dto.getLCUConfiguration()
    fnuConfiguration = lcuConfiguration.fnuConfiguration
    fnuRestclient = FNURestClient(fnuConfiguration=fnuConfiguration)

    # destroy LV
    log.notice("Going to delete ISCSI target-extent-assignment with id %s" %
               lcuContainer.fnuTargetExtentAssignment.id)
    fnuRestclient.deleteISCSITargetExtent(lcuContainer.fnuNasInfo, lcuContainer.fnuTargetExtentAssignment.id)
    fnuRestclient.deleteISCSIExtent(lcuContainer.fnuNasInfo, lcuContainer.fnuExtent.id)
    fnuRestclient.deleteISCSITarget(lcuContainer.fnuNasInfo, lcuContainer.fnuTarget.id)
    # dataset
    log.notice("Going to delete zvol dataset %s" %
               lcuContainer.nasdataset)
    fnuRestclient.deleteDataset(lcuContainer.fnuNasInfo, lcuContainer.nasdataset, forceDeletion=True)

    lcuContainer.checkedRootfsDevPath = rootfsDevPath
    return rootfsDevPath


'''
  Resizing rootfs device
'''

def resizeRootfs(lcuContainer:dto.LCUContainer=None, lxcContainer:dto.LXCContainer=None, newSizeGBs:int=-1):
    '''
        resizes the rootfs filesystem of the specified container to the specified size.
        so far only increasing the size is possible automatically.
        lcuContainer MUST be checked, i.e. semantically valid in regard to mount options in .conf file!
        on failure sys.exit is called.
    ''' 
    if lcuContainer == None:
        raise Exception("ERROR: resizeRootfs() called with lcuContainer None")
    if lxcContainer == None:
        raise Exception("ERROR: resizeRootfs() called with lxcContainer None")
    #if len(lcuContainer.name) == 0:
    #    raise Exception("ERROR: resizeRootfs() called with lcuContainer with empty name")
    if newSizeGBs <= 0:
        raise Exception("ERROR: resizeRootfs() called with unusable size %d" % newSizeGBs)
    
    if newSizeGBs <= lcuContainer.size:
        log.error("Specified new container size %d GB is smaller or equal to originally configured size %d GB, not resizing container %s" % (
            newSizeGBs, lcuContainer.size, lcuContainer.name))
        sys.exit(1)

    rootfsMountDir = lxcContainer.lxcFileDict.get("lxc.rootfs.path", "")
    if len(rootfsMountDir) == 0:
        rootfsMountDir = "/var/lib/lxc/%s/rootfs" % lxcContainer.name

    # check old size
    oldSize = detectRootfsSize(lcuContainer=lcuContainer)
    if oldSize <= 0:
        log.error("Unable to detect old size of eventual rootfsBlockDev of container %s, not checking old size" %
            lcuContainer.name)
        sys.exit(1)
    else:  
        log.notice("Detected old size of rootfsBlockDev of container %s to be %d bytes (%s)" % (
            lcuContainer.name, oldSize, helpers.formatFileSize(oldSize)))
        if newSizeGBs*1073741824 <= oldSize: 
            log.error("Container %s's old rootfsBlockDev %d bytes (%s) big, requested new size %s bytes (%s GB) is not larger, aborting resize" % (
                lcuContainer.name, 
                oldSize, helpers.formatFileSize(oldSize),
                newSizeGBs*1073741824, newSizeGBs))
            sys.exit(1)

    lcuContainer.checkedRootfsDevPath = ""
    if len(lcuContainer.lvgroup) > 0:
        _resizeRootfsLvm(lcuContainer=lcuContainer, newSizeGBs=newSizeGBs)
    elif len(lcuContainer.lfdir) > 0:
        _resizeRootfsLoopFile(lcuContainer=lcuContainer, newSizeGBs=newSizeGBs)
    elif len(lcuContainer.nasdataset) > 0:
        _resizeRootfsIscsi(lcuContainer=lcuContainer, newSizeGBs=newSizeGBs)
    else:
        log.warning("resizeRootfs(%s): Neither CTLV* nor CTLF* nor CTNASNAME set, nothing to do, or is there?" %
              lcuContainer.name)
        return

    if len(lcuContainer.checkedRootfsDevPath) == 0:
        log.error("resizeRootfs(%s): Failed to detect resp. resize rootfsDevPath" %
              lcuContainer.name)
        sys.exit(1)
    log.notice("Successfully resized checkedRootfsDevPath %s" %
          lcuContainer.checkedRootfsDevPath)
    
    # crypt stuff
    finalRootfsMountDevPath = lcuContainer.checkedRootfsDevPath
    if len(lcuContainer.crypttype) > 0:
        finalRootfsMountDevPath = _resizeCryptedRootfsDev(lcuContainer)

    if not os.path.ismount("/var/lib/lxc/%s/rootfs" % lcuContainer.name):
        log.warning("resizeRootfs(%s): Container is not running nor mounted, performing fsck.ext4 first" %
              lcuContainer.name)
        cmdParts = [ "/sbin/fsck.ext4", "-p", finalRootfsMountDevPath]
        cp = subprocess.run(cmdParts)
        if cp.returncode == 0:
            log.good("Sucessfully performed fsck.ext4 -p %s for LCU/LXC container %s, filesystem was clean" % (
                 finalRootfsMountDevPath, lcuContainer.name))
        elif cp.returncode == 1:
            log.warning("Sucessfully performed fsck.ext4 -p %s for LCU/LXC container %s, file system errors were corrected" % (
                 finalRootfsMountDevPath, lcuContainer.name))
        else:
            problem = "ERROR: Failed to fsck.ext4 -p %s for LCU/LXC container %s" % (
                 finalRootfsMountDevPath, lcuContainer.name)
            helpers.backcheckOnProblem(rc=cp.returncode, cmdParts=cmdParts, problem=problem)

    log.notice("Resizing ext4 fs on %s for LCU/LXC container %s" %
             (finalRootfsMountDevPath, lcuContainer.name))
    cmdParts = [ "/sbin/resize2fs", finalRootfsMountDevPath]
    cp = subprocess.run(cmdParts)
    if cp.returncode != 0:
        problem = "ERROR: Failed to resize2fs %s for LCU/LXC container %s" % (
             finalRootfsMountDevPath, lcuContainer.name)
        helpers.backcheckOnProblem(rc=cp.returncode, cmdParts=cmdParts, problem=problem)
    else:
        log.good("Sucessfully resized ext4 fs on %s for LCU/LXC container %s" % (
             finalRootfsMountDevPath, lcuContainer.name))

    log.notice("resizeRootfs(%s): Done resizing %s" %
          (lcuContainer.name, rootfsMountDir))
    

def _resizeRootfsLvm(lcuContainer:dto.LCUContainer=None, newSizeGBs:int=-1):
    '''
      resizes the LVM for the rootfs of the specified lcuContainer 
      on success the returned device is ready to be resize2fs'd and stored in lcuContainer.checkedRootfsDevPath
      on failure sys.exit is called
    '''
    if newSizeGBs < 2:
        raise Exception("ERROR: _resizeRootfsLvm() called with unusable size %d GB" % newSizeGBs)

    rootfsDevPath = _checkRootfsLvm(lcuContainer=lcuContainer)

    # resize LV
    log.notice("Going to resize LV %s in VG %s with %d GB" % (
        lcuContainer.lvname, lcuContainer.lvgroup, lcuContainer.size))
    cmdParts = [ "/sbin/lvresize", "--size", "%dG" % newSizeGBs, rootfsDevPath ]
    cp = subprocess.run(cmdParts)
    if cp.returncode != 0:
        problem = "Unable to resize LV %s in VG %s" % (
            lcuContainer.lvname, lcuContainer.lvgroup)
        helpers.backcheckOnProblem(rc=cp.returncode, cmdParts=cmdParts, problem=problem)
    else:
        log.notice("Successfully resized LV %s in VG %s to %d GB" % (
            lcuContainer.lvname, lcuContainer.lvgroup, lcuContainer.size))

    lcuContainer.checkedRootfsDevPath = rootfsDevPath
    return rootfsDevPath

def _resizeRootfsLoopFile(lcuContainer:dto.LCUContainer=None, newSizeGBs:int=-1):
    '''
      resizes the .ext4 file for the rootfs of the specified lcuContainer 
      on success the returned device is ready to be resize2fs'd and stored in lcuContainer.checkedRootfsDevPath
      on failure sys.exit is called
    '''
    if newSizeGBs < 2:
        raise Exception("ERROR: _resizeRootfsLoopFile() called with unusable size %d GB" % newSizeGBs)

    rootfsDevPath = _checkRootfsLoopFile(lcuContainer=lcuContainer)

    # resize file
    log.notice("Going to resize CTLFile %s with %d GB" % (
        rootfsDevPath, lcuContainer.size))
    
    rootfsDevPath = _resizeLoopFile(loopFile=rootfsDevPath, wantedSizeBytes=newSizeGBs*1073741824)
    lcuContainer.checkedRootfsDevPath = rootfsDevPath
    return rootfsDevPath

def _resizeLoopFile(loopFile:str="", wantedSizeBytes:int=-1):
    '''
      resizes the specfied file 
      on failure sys.exit is called
      
      if the resized file is loop-mounted the reg. dev path is returned, like /dev/loop0
      otherwise the loopfile itself is returned 
    '''
    if len(loopFile) == 0 or not loopFile.startswith("/"):
        raise Exception("ERROR: _resizeLoopFile() called with unusable loopFile %s" % loopFile)
    if wantedSizeBytes < 1073741824:
        raise Exception("ERROR: _resizeLoopFile() called with unusable small size %d bytes (%s)" % (
            wantedSizeBytes, helpers.formatFileSize(wantedSizeBytes)))

    cmdParts = [ "/usr/bin/truncate", "--size=%d" % wantedSizeBytes, loopFile ]
    cp = subprocess.run(cmdParts)
    if cp.returncode != 0:
        problem = "Unable to resize loopfile %s" % (
            loopFile)
        helpers.backcheckOnProblem(rc=cp.returncode, cmdParts=cmdParts, problem=problem)
    else:
        log.notice("Successfully resized loopfile %s to %d bytes (%s)" % (
            loopFile, wantedSizeBytes, helpers.formatFileSize(wantedSizeBytes)))

    # in this case we need to use the loop device, like /dev/loop0

    cmdParts = [ "/sbin/losetup", "-j", loopFile ]
    cp = subprocess.run(cmdParts, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
    if cp.stdout == None or len(cp.stdout) == 0:
        log.notice("Returning %s itself because it is not currently used as /dev/loop* (losetup returned nothing)" %
                   loopFile)
        return loopFile
    losetupInfoParts = cp.stdout.split(":")
    #log.debug("losetupInfoParts = '%s'" % "' _:_ '".join(losetupInfoParts))
    if not losetupInfoParts[0].startswith("/dev/loop"):
        log.error("_resizeLoopFile(%s): Dying becaue losetup -j output does not start with '/dev/loop.*:' but is\n%s\n" % (
            loopFile, cp.stdout))
        sys.exit(1)

    loopDevFile = losetupInfoParts[0]
    log.notice("%s is used by loopDevFile %s, resize2fs should use this then, but resizing the loopDev first" % (
                loopFile, loopDevFile))
    cmdParts = [ "/sbin/losetup", "-v", "-c", loopDevFile ]
    cp = subprocess.run(cmdParts)
    if cp.returncode != 0:
        problem = "_resizeLoopFile(%s): Unable to resize loop device %s" % (
            loopFile, loopDevFile)
        helpers.backcheckOnProblem(rc=cp.returncode, cmdParts=cmdParts, problem=problem)
    else:
        log.notice("Successfully resized loop device %s which is using the loop file %s" % (
            loopDevFile, loopFile))
    return loopDevFile

def _resizeRootfsIscsi(lcuContainer:dto.LCUContainer=None, newSizeGBs:int=-1):
    '''
      resizes the zvol dataset and iscsi target file for the rootfs of the specified lcuContainer 
      on success the returned device is ready to be resize2fs'd and stored in lcuContainer.checkedRootfsDevPath
      on failure sys.exit is called
    '''
    if newSizeGBs < 2:
        raise Exception("ERROR: _resizeRootfsIscsi() called with unusable size %d GB" % newSizeGBs)

    rootfsDevPath = _checkRootfsIscsi(lcuContainer=lcuContainer, performLogin=True)

    # resize dataset
    lcuConfiguration = dto.getLCUConfiguration()
    fnuConfiguration = lcuConfiguration.fnuConfiguration
    #ctNasInfo = fnuConfiguration.nasInfos.get(lcuContainer.nasname)
    fnuRestclient = FNURestClient(fnuConfiguration=fnuConfiguration)
    wantedSizeBytes = newSizeGBs*1073741824
    fnuRestclient.updateDataset(
        nasInfo=lcuContainer.fnuNasInfo,
        datasetPath=lcuContainer.nasdataset,
        volSizeBytes=wantedSizeBytes)
        # do not touch: datasetComment=""
    # rescan target
    iscsihelpers.rescanIscsiTargets(basename=lcuContainer.iscsiBasename, portalListeners=lcuContainer.iscsiPortalListeners, targetName=lcuContainer.iscsiName)
    if len(lcuContainer.iscsiPortalListeners) > 1:
        cmdParts = [ "/sbin/multipathd", "resize", "multipath", rootfsDevPath ]
        log.debug("_resizeRootfsIscsi(%s): Calling %s" % (
            lcuContainer.name, " ".join(cmdParts)))
        cp = subprocess.run(cmdParts, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
        if cp.returncode != 0:
            log.error("_resizeRootfsIscsi(%s): Failed to adapt multipathd size, '%s' gave non-0 returncode %d, aborting" % (
                lcuContainer.name, " ".join(cmdParts), cp.returncode))
            sys.exit(1)
    # done!
    lcuContainer.checkedRootfsDevPath = rootfsDevPath
    return rootfsDevPath


'''
    detect existing rootfs sizes
'''

def detectRootfsSize(lcuContainer:dto.LCUContainer=None):
    '''
        detects size of eventual block device backing the specified container's rootfs.
        if there is such a block device the size in bytes is returned
        for containers without a block device rootfs -1 is returned
    '''
    if lcuContainer == None:
        raise Exception("ERROR: detectRootfsSize() called with lcuContainer None")
    
    lcuContainer.checkedRootfsDevPath = ""
    if len(lcuContainer.lvgroup) > 0:
        return _detectRootfsLvmSize(lcuContainer=lcuContainer)
    elif len(lcuContainer.lfdir) > 0:
        return _detectRootfsLoopFileSize(lcuContainer=lcuContainer)
    elif len(lcuContainer.nasdataset) > 0:
        return _detectRootfsIscsiSize(lcuContainer=lcuContainer)
    else:
        log.warning("detectRootfsSize(%s): Neither CTLV* nor CTLF* nor CTNASNAME set, unable to detect meaningful rootfs size" %
              lcuContainer.name)
        return -1


def _detectRootfsLvmSize(lcuContainer:dto.LCUContainer=None):
    '''
        detects size of LVM LV backing the specified container's rootfs.
    ''' 
    if lcuContainer == None:
        raise Exception("ERROR: _detectRootfsLvmSize() called with lcuContainer None")
    if len(lcuContainer.lvgroup) == 0 or len(lcuContainer.lvname) == 0:
        raise Exception("ERROR: _detectRootfsLvmSize(%s) called with lcuContainer without CTLV* set" % lcuContainer.name)
    size = -1
    try:
        cp = subprocess.run(["/sbin/lvs", "--units","b", "%s/%s" % (lcuContainer.lvgroup, lcuContainer.lvname)],
                            stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
        #log.debug("_detectRootfsLvmSize(%s): cp.stdout=\n%s\n" % (lcuContainer.name, cp.stdout))
        lines = cp.stdout.split("\n")
        #log.debug("_detectRootfsLvmSize(%s): len(lines) = %d" % (lcuContainer.name, len(lines)))
        for line in lines:
            #log.debug("_detectRootfsLvmSize(%s): line=\n%s\n" % (lcuContainer.name, line))
            strippedLine = line.strip()
            if strippedLine.startswith(lcuContainer.lvname+" "):
                #log.debug("_detectRootfsLvmSize(%s): matching strippedLine=\n%s\n" % (lcuContainer.name, strippedLine))
                lineParts = strippedLine.split(" ")
                #log.debug("_detectRootfsLvmSize(%s): len(lineParts) = %d" % (lcuContainer.name, len(lineParts)))
                #log.debug("_detectRootfsLvmSize(%s): lineParts[-1] = %s" % (lcuContainer.name, lineParts[-1]))
                if len(lineParts) >= 4:
                    size = int(lineParts[-1][:-1]) # the [:-1] cuts away the unit char 'B'
                    break
    except Exception as e:
        log.warning("_detectRootfsLvmSize(%s): Swallowing following exception:\n%s\n" % (
                    lcuContainer.name, e))
        size = -2
    log.debug("_detectRootfsLvmSize(%s): Detected CTLV rootfsDev size tb be %d (%s)" % (
        lcuContainer.name, size, helpers.formatFileSize(size)))
    return size


def _detectRootfsLoopFileSize(lcuContainer:dto.LCUContainer=None):
    '''
        detects size of ext4 loop file backing the specified container's rootfs.
    ''' 
    # check old size
    rootfsDevPath = _checkRootfsLoopFile(lcuContainer=lcuContainer)
    size = -1
    try:
        statResult = os.stat(rootfsDevPath)
        if statResult != None:
            size = int(statResult.st_size)
    except Exception as e:
        log.warning("_detectRootfsLoopFileSize(%s): Swallowing following exception:\n%s\n" % (
                    lcuContainer.name, e))
        size = -2
    log.debug("_detectRootfsLoopFileSize(%s): Detected .ext4 loopfile rootfsDev size tb be %d (%s)" % (
        lcuContainer.name, size, helpers.formatFileSize(size)))
    return size

def _detectRootfsIscsiSize(lcuContainer:dto.LCUContainer=None):
    '''
        detects size of dataset backing the specified container's rootfs.
    ''' 
    # check old size
    #rootfsDevPath = 
    _checkRootfsIscsi(lcuContainer=lcuContainer, performLogin=False)
    size = -1
    try:
        #lcuConfiguration = dto.getLCUConfiguration()
        #fnuConfiguration = lcuConfiguration.fnuConfiguration
        #ctNasInfo = fnuConfiguration.nasInfos.get(lcuContainer.nasname)
        #fnuRestclient = FNURestClient(fnuConfiguration=fnuConfiguration)
        #fnuRestclient.getDatasets(lcuContainer.fnuNasInfo)
        #ctDataset = lcuContainer.fnuNasInfo.nas_datasets_flat[lcuContainer.nasdataset]
        size = lcuContainer.fnuDataset.volsize
    except Exception as e:
        log.warning("_detectRootfsIscsiSize(%s): Swallowing following exception:\n%s" % (
                    lcuContainer.name, e))
        size = -2
    log.debug("_detectRootfsIscsiSize(%s): Detected zvol dataset size tb be %d (%s)" % (
        lcuContainer.name, size, helpers.formatFileSize(size)))
    return size


'''
    other mount related stuff
'''

def parseProcMounts():
    '''
        parses /proc/mounts
        returns a dict of lists, each containing
          mountsDict[mountpoint] = [ device, mountpoint, fstype, fsoptions, checkHint1:int, checkHint2:int ]
    '''
    mountsDict = {}
    for mountsLine in fileinput.input(files=("/proc/mounts")):
        if len(mountsLine) == 0:
            continue
        mountLineParts = mountsLine.split(" ")
        if len(mountLineParts) < 6:
            continue
        mountpoint = mountLineParts[1]
        mountsDict[mountpoint] = mountLineParts
        #log.debug("Detected /proc/mounts mount: %s mounted on %s as %s with options %s" % (
        #    mountLineParts[0], mountpoint, mountLineParts[2], mountLineParts[3]))
    return mountsDict


'''
    locking stuff
'''

def lockContainer(lcuContainer:dto.LCUContainer=None):
    '''
        locks container by creating /etc/lxcutils/conf/%s.lock containing the VZ's short hostname
    '''
    if lcuContainer == None:
        raise Exception("ERROR: lockContainer() called with lcuContainer None")
    lockfileName = "/etc/lxcutils/conf/%s.lock" % lcuContainer.name
    if os.path.exists(lockfileName):
        raise Exception("ERROR: lockContainer(): Container already locked, %s exists" % lockfileName)
    hostname = socket.gethostname().split('.')[0]
    if hostname == None or len(hostname) == 0:
        raise Exception("ERROR: lockContainer(): Unable to detect my hostname")
    with open(lockfileName, mode='w') as lockfileHandle:
        print(hostname, file=lockfileHandle, flush=True)
    log.debug("lockContainer(): Successfully written %s to %s" % (hostname, lockfileName))

def haveContainerLock(lcuContainer:dto.LCUContainer=None):
    '''
        detects if we have the lock for the container, by checking if
        /etc/lxcutils/conf/%s.lock exists and contains the VZ's short hostname.
        does not raise exception but returns True or False
    '''
    if lcuContainer == None:
        raise Exception("ERROR: haveContainerLock() called with lcuContainer None")
    hostname = socket.gethostname().split('.')[0]
    if hostname == None or len(hostname) == 0:
        raise Exception("ERROR: lockContainer(): Unable to detect my hostname")
    lockfileName = "/etc/lxcutils/conf/%s.lock" % lcuContainer.name
    if not os.path.exists(lockfileName):
        log.debug("haveContainerLock(): Container not locked, no file %s" % lockfileName)
        return False
    with open(lockfileName, mode='r') as lockfileHandle:
        lockfileContent = lockfileHandle.read()
    if lockfileContent.split("\n")[0] == hostname:
        log.debug("haveContainerLock(): File %s contains %s, we have the lock" % (lockfileName, hostname))
        return True
    log.debug("haveContainerLock(): File %s contains not '%s' but '%s', somenoe else has the lock" % (
        lockfileName, hostname, lockfileContent))
    return False

def readContainerLock(lcuContainer:dto.LCUContainer=None):
    '''
        reads an eventual /etc/lxcutils/conf/%s.lock
        if the file exists the content is returned
        otherwise "" is returned
    '''
    if lcuContainer == None:
        raise Exception("ERROR: readContainerLock() called with lcuContainer None")
    hostname = socket.gethostname().split('.')[0]
    if hostname == None or len(hostname) == 0:
        raise Exception("ERROR: lockContainer(): Unable to detect my hostname")
    lockfileName = "/etc/lxcutils/conf/%s.lock" % lcuContainer.name
    if not os.path.exists(lockfileName):
        log.debug("readContainerLock(): Container not locked, no file %s" % lockfileName)
        return ""
    with open(lockfileName, mode='r') as lockfileHandle:
        lockfileContent = lockfileHandle.read()
    log.debug("readContainerLock(): Returning content of %s, '%s'" % (
        lockfileName, lockfileContent))
    return lockfileContent

def unlockContainer(lcuContainer:dto.LCUContainer=None):
    '''
        unlocks container by removing /etc/lxcutils/conf/%s.lock if and only if it contains the VZ's short hostname
    '''
    if lcuContainer == None:
        raise Exception("ERROR: unlockContainer() called with lcuContainer None")
    if not haveContainerLock(lcuContainer):
        log.warning("unlockContainer() Unable to unlock container %s that we don't have the lock on. May be normal after LCU updates or with lcu-umount in CT's post-stop script." % lcuContainer.name)
        return
    lockfileName = "/etc/lxcutils/conf/%s.lock" % lcuContainer.name
    os.remove(lockfileName)
    log.debug("unlockContainer(): Successfully removed %s" % lockfileName)
