'''
  $Id$
  
  Copyleft ITEG IT-Engineers GmbH 2021
'''

#import fileinput
import ipaddress
#import iptc
#import json
#import os
#import re
import socket
import subprocess

from iteg.lxcutils import dto
from iteg.lxcutils import helpers
from iteg.lxcutils import lculogging


log = lculogging.getLogger()


'''
  ping wrappers
'''

successfullyPingedHostnames = []
successfullyPingedIPv4 = []
successfullyPingedIPv6 = []

def hostReachable(hostname:str=""):
    '''
        ping hostname specified as string
    '''
    if len(hostname) == 0:
        raise Exception("hostReachable() called with empty hostname")
    if hostname in successfullyPingedHostnames:
        log.debug("Host %s in list of successfully pinged hostnames, not re-pinging" % hostname)
        return True
    cp = subprocess.run(["/usr/bin/fping", "-q", hostname])
    if cp.returncode == 0:
        log.debug("Successfully pinged hostname %s" % hostname)
        successfullyPingedHostnames.append(hostname)
        return True
    log.debug("Host %s unreachable (fping failed), might be the desired result." % hostname)
    return False

def ip4Reachable(ip:ipaddress.IPv4Address=None):
    '''
        pings ip 
    '''
    if ip == None:
        raise Exception("hostReachable() called with ip==None")
    if ip.exploded in successfullyPingedIPv4:
        log.debug("IPv4 %s in list of successfully pinged IPv4s, not re-pinging" % ip.exploded)
        return True
    cp = subprocess.run(["/usr/bin/fping", "-q", "-4", ip.exploded])
    if cp.returncode == 0:
        log.debug("Successfully pinged IPv4 %s" % ip.exploded)
        successfullyPingedIPv4.append(ip.exploded)
        return True
    log.debug("IPv4 %s unreachable (fping failed), might be the desired result." % ip.exploded)
    return False

def ip6Reachable(ip:ipaddress.IPv6Address=None):
    '''
        pings ip 
    '''
    if ip == None:
        raise Exception("hostReachable() called with ip==None")
    if ip.compressed in successfullyPingedIPv6:
        log.debug("IPv6 %s in list of successfully pinged IPv6s" % ip.compressed)
        return True
    cp = subprocess.run(["/usr/bin/fping", "-q", "-6", ip.compressed])
    if cp.returncode == 0:
        log.debug("Successfully pinged IPv6 %s" % ip.compressed)
        successfullyPingedIPv6.append(ip.compressed)
        return True
    log.debug("IPv6 %s unreachable (fping failed), might be the desired result." % ip.compressed)
    return False


def checkLCUEnvironmentIPsReachable(lcuConfiguration:dto.LCUConfiguration=None):
    '''
        checks if the conf-checked environment Gateway IPs are reachable.
        If some are unreachble according messages are added to lcuConfiguration.checkErrors
    '''
    if len(lcuConfiguration.defaultDnsList) > 0:
        for hostname in lcuConfiguration.defaultDnsList.split(" "):
            if hostReachable(hostname):
                log.debug("Successfully pinged default DNS %s" % hostname)
            else:
                log.debug("Failed to ping DNS %s" % hostname)
                lcuConfiguration.checkWarnings.append("WARNINGS: Default DNS %s not reachable" % hostname)
    if lcuConfiguration.checkedDmzGateway != None:
        if ip4Reachable(lcuConfiguration.checkedDmzGateway):
            log.debug("Successfully pinged DMZ Gateway %s" % lcuConfiguration.checkedDmzGateway)
        else:
            log.debug("Failed to ping DMZ Gateway %s" % lcuConfiguration.checkedDmzGateway)
            lcuConfiguration.checkErrors.append("ERROR: DMZ Gateway %s not reachable" % lcuConfiguration.checkedDmzGateway)
    if lcuConfiguration.checkedWanGateway != None:
        if ip4Reachable(lcuConfiguration.checkedWanGateway):
            log.debug("Successfully pinged WAN Gateway %s" % lcuConfiguration.checkedWanGateway)
        else:
            log.debug("Failed to ping WAN Gateway %s" % lcuConfiguration.checkedWanGateway)
            lcuConfiguration.checkErrors.append("ERROR: WAN IPv4 Gateway %s not reachable" % lcuConfiguration.checkedWanGateway)
    if lcuConfiguration.checkedWan6Gateway != None:
        if ip6Reachable(lcuConfiguration.checkedWan6Gateway):
            log.debug("Successfully pinged WAN IPv6 Gateway %s" % lcuConfiguration.checkedWan6Gateway)
        else:
            log.debug("Failed to ping WAN Gateway %s" % lcuConfiguration.checkedWan6Gateway)
            lcuConfiguration.checkErrors.append("ERROR: WAN IPv6 Gateway %s not reachable" % lcuConfiguration.checkedWan6Gateway)


def checkLCUContainerEnvironmentIPsReachable(lcuContainer:dto.LCUContainer=None):
    '''
        checks if the conf-checked environment Gateway IPs are reachable.
        If some are unreachble according messages are added to lcuConfiguration.checkErrors
    '''
    if len(lcuContainer.dnsList) > 0:
        for hostname in lcuContainer.dnsList.split(" "):
            if hostReachable(hostname):
                log.debug("Successfully pinged container DNS %s" % hostname)
            else:
                log.debug("Failed to ping DNS %s" % hostname)
                lcuContainer.checkWarnings.append("WARNINGS: Container DNS %s not reachable" % hostname)
    else:
        log.warning("Container %s has empty dnsList: %s" % (lcuContainer.name, lcuContainer.dnsList))


def checkLCUContainerNameResolvable(lcuContainer:dto.LCUContainer=None):
    '''
        checks if the conf-checked lcuContainer's hostname & domain resolve to an IPv4
        If not an according message is added to lcuContainer.checkWarnings
    '''
    if lcuContainer == None:
        raise Exception("checkLCUContainerNameResolvablelcuContainer() called with lcuContainer==None")
    if (len(lcuContainer.hostname) + len(lcuContainer.domain)) > 0:
        try:
            mainip = socket.gethostbyname("%s.%s" % (lcuContainer.hostname, lcuContainer.domain))
            log.debug("GOOD: %s.%s resolves to %s" % (
                 lcuContainer.hostname, lcuContainer.domain, mainip))
        except socket.error:
            log.debug("WARNING: %s.%s does NOT resolve" % (
                 lcuContainer.hostname, lcuContainer.domain))
            lcuContainer.checkWarnings.append("CTNAME.CTDOMAIN %s.%s does NOT resolve to an IPv4 address" % (
                 lcuContainer.hostname, lcuContainer.domain))


def checkLCUContainerIPsAsFree(lcuContainer:dto.LCUContainer=None, lxcContainer:dto.LXCContainer=None):
    '''
        checks if the conf-checked IPs for the specified LCU container are free,
        i.e. if they are unreachable as expected.
        If IPs are unexpectedly reachable messages are added to lcuContainer.checkErrors
        Returns the number of pingable IPs.
    '''
    if lcuContainer == None:
        raise Exception("countLCUContainerIPsReachable() called with lcuContainer==None")
    log.notice("Going to fping container's IPs hoping to fail, this will take a few seconds")
    
    reasonsToCheckByIP = {}
    
    if lcuContainer.checkedIp0 != None:
        reasonsToCheckByIP[lcuContainer.checkedIp0.exploded] = "CTIP0 %s" % lcuContainer.checkedIp0.exploded
    if lcuContainer.checkedIp1 != None:
        reasonsToCheckByIP[lcuContainer.checkedIp1.exploded] = "CTIP1 %s" % lcuContainer.checkedIp1.exploded
    if lcuContainer.checkedIp2 != None:
        reasonsToCheckByIP[lcuContainer.checkedIp2.exploded] = "CTIP2 %s" % lcuContainer.checkedIp2.exploded
    if lcuContainer.checkedIpSix0 != None:
        reasonsToCheckByIP[lcuContainer.checkedIpSix0.compressed] = "CTIPSIX0 %s" % lcuContainer.checkedIpSix0.compressed
    if lcuContainer.checkedIpSix1 != None:
        reasonsToCheckByIP[lcuContainer.checkedIpSix1.compressed] = "CTIPSIX1 %s" % lcuContainer.checkedIpSix1.compressed
        
    if lxcContainer != None:
        if lxcContainer.lxcFileDict != None:
            for nr in range(0, 4):
                ip4wnm = lxcContainer.lxcFileDict.get("lxc.net.%d.ipv4.address" % nr, "")
                if len(ip4wnm) > 0:
                    ip4 = ipaddress.IPv4Interface(ip4wnm)
                    reasonsToCheckByIP[ip4] = "lxc.net.%d.ipv4.address %s" % (nr, ip4)
                ip6wnm = lxcContainer.lxcFileDict.get("lxc.net.%d.ipv6.address" % nr, "")
                if len(ip6wnm) > 0:
                    ip6 = ipaddress.IPv6Interface(ip6wnm)
                    reasonsToCheckByIP[ip6] = "lxc.net.%d.ipv6.address %s" % (nr, ip6)
                    
    nrReachableIPs = 0
    for ip in reasonsToCheckByIP.keys():
        log.debug("INFO: fping-ing %s" % reasonsToCheckByIP[ip])
        if hostReachable(ip):
            lcuContainer.checkErrors.append("ERROR: %s reachable" % reasonsToCheckByIP[ip])
            nrReachableIPs += 1
    
    return nrReachableIPs


'''
    NAT helpers
'''

def natContainer(lcuContainer:dto.LCUContainer=None, lxcContainer:dto.LCUContainer=None, lcuConfiguration:dto.LCUConfiguration=None):
    '''
        activates eventual automatic NAT rules to make a container accessible from outside.
        based on lcu2-nat
    ''' 
    if lcuContainer == None:
        raise Exception("ERROR: natContainer() called with lcuContainer None")
    if lxcContainer == None:
        raise Exception("ERROR: natContainer() called with lxcContainer None")
    if lcuConfiguration == None:
        raise Exception("ERROR: natContainer() called with lcuConfiguration None")
    if lcuContainer.checkedNatIP == None and lcuContainer.natip != "auto":
        log.info("Nothing to NAT for container %s." % lcuContainer.name)
        return True
    
    # detect DMZ IP, i.e. internal IP of container
    ctLocalIp = detectLXCContainerFirstIPv4(lcuContainer=lcuContainer, lxcContainer=lxcContainer)
    if ctLocalIp == None:
        log.error("Unable to detect usable local IP of container %s, aborting NAT work." % lcuContainer.name)
        return False
    
    # detect wan IF
    wanIf = checkOrDetectHostsFirstWANInterface(lcuConfiguration=lcuConfiguration)
    if len(wanIf) == 0:
        log.error("Unable to auto-detect WAN interface (no usable output from ip route show), aborting NAT work for container %s." %
                  lcuContainer.name)
        return False

    # checkedNatIP
    if lcuContainer.natip == "auto":
        lcuContainer.checkedNatIP = checkOrDetectHostsFirstWANishIP(lcuConfiguration=lcuConfiguration)
        if lcuContainer.checkedNatIP == None:
            log.info("Unable to auto-detect non-private WANIP as automatic CTNATIP, aborting NAT work for container %s." %
                lcuContainer.name)
            return False
        if not ip4Reachable(lcuContainer.checkedNatIP):
            log.info("Unable to reach WANIP %s to use as automatic CTNATIP, aborting NAT work for container %s" % (
                lcuContainer.checkedNatIP.exploded, lcuContainer.name))
            return False
    else:
        if ip4Reachable(lcuContainer.checkedNatIP):
            log.info("Something is using configured CTNATIP %s, refusing to use this IP, aborting NAT work for container %s" % (
                lcuContainer.checkedNatIP.exploded, lcuContainer.name))
            return False
        log.info("Adding CTNATIP %s/%d to WAN interface %s for container %s" % (
            lcuContainer.checkedNatIP.exploded, lcuConfiguration.wanMaskBits, wanIf, lcuContainer.name))
        rc = helpers.performDebugLogged(["/sbin/ip", "addr", "add",
                                         "%s/%d" % (lcuContainer.checkedNatIP.exploded, lcuConfiguration.wanMaskBits),
                                         "dev", wanIf])
        if rc != 0:
            log.error("Unable to bring up configured CTNATIP, aborting NAT work for container %s." %
                     lcuContainer.name)
            return False
    
    log.notice("Actually creating NAT rules for public CTNATIP %s to local container IP %s for container %s" % (
        lcuContainer.checkedNatIP.exploded, ctLocalIp.exploded, lcuContainer.name))

    ### First attempt was based on python3-iptables see /usr/share/doc/python-iptables-doc/html/index.html from python-iptables-doc.
    ### BUT this creates 'legacy' iptables chains and was complicated and hence was dropped
    
    helpers.performDebugLogged(["/usr/sbin/iptables", "-t","nat", "-N","%s_post"%lcuContainer.natName])
    helpers.performDebugLogged(["/usr/sbin/iptables", "-t","nat", "-A","%s_post"%lcuContainer.natName,
                         "-s",ctLocalIp.exploded, "-o",wanIf, "-j","SNAT", "--to-source",lcuContainer.checkedNatIP.exploded])
    helpers.performDebugLogged(["/usr/sbin/iptables", "-t","nat", "-A","POSTROUTING", "-j","%s_post"%lcuContainer.natName])

    if len(lcuContainer.natportstcp+lcuContainer.natportsudp) > 0:
        helpers.performDebugLogged(["/usr/sbin/iptables", "-t","nat", "-N","%s_pre"%lcuContainer.natName])
    if len(lcuContainer.natportstcp) > 0:
        for port in lcuContainer.natportstcp.split(" "):
            helpers.performDebugLogged(cmdParts=[
                "/usr/sbin/iptables", "-t","nat", "-A","%s_pre"%lcuContainer.natName,
                "-d",lcuContainer.checkedNatIP.exploded, "-m","state", "--state","NEW",
                "-p","tcp", "--destination-port",port,
                "-j","DNAT", "--to-destination",ctLocalIp.exploded])
    if len(lcuContainer.natportsudp) > 0:
        for port in lcuContainer.natportsudp.split(" "):
            helpers.performDebugLogged(cmdParts=[
                "/usr/sbin/iptables", "-t","nat", "-A","%s_pre"%lcuContainer.natName,
                "-d",lcuContainer.checkedNatIP.exploded, "-m","state", "--state","NEW",
                "-p","udp", "--destination-port",port,
                "-j","DNAT", "--to-destination",ctLocalIp.exploded])
    if len(lcuContainer.natportstcp+lcuContainer.natportsudp) > 0:
        helpers.performDebugLogged(["/usr/sbin/iptables", "-t","nat", "-A","PREROUTING", "-j","%s_pre"%lcuContainer.natName])

    log.notice("Done creating NAT rules for public CTNATIP %s to local container IP %s for container %s" % (
        lcuContainer.checkedNatIP.exploded, ctLocalIp.exploded, lcuContainer.name))
    
    return True
    
    
def unnatContainer(lcuContainer:dto.LCUContainer=None, lxcContainer:dto.LCUContainer=None, lcuConfiguration:dto.LCUConfiguration=None):
    '''
        deactivates eventual automatic NAT rules that made a container accessible from outside.
        based on lcu2-unnat
    ''' 
    if lcuContainer == None:
        raise Exception("ERROR: unnatContainer() called with lcuContainer None")
    if lxcContainer == None:
        raise Exception("ERROR: unnatContainer() called with lxcContainer None")
    if lcuConfiguration == None:
        raise Exception("ERROR: unnatContainer() called with lcuConfiguration None")
    if lcuContainer.checkedNatIP == None and lcuContainer.natip != "auto":
        log.info("Nothing to NAT for container %s." % lcuContainer.name)
        return True

    # detect DMZ IP, i.e. internal IP of container
    ctLocalIp = detectLXCContainerFirstIPv4(lcuContainer=lcuContainer, lxcContainer=lxcContainer)
    if ctLocalIp == None:
        log.error("Unable to detect usable local IP of container %s, continuing un-NAT work anyway." % lcuContainer.name)
    
    # detect wan IF
    wanIf = checkOrDetectHostsFirstWANInterface(lcuConfiguration=lcuConfiguration)
    if len(wanIf) == 0:
        log.error("Unable to auto-detect WAN interface (no usable output from ip route show), continuing un-NAT work for container %s anyway." %
                  lcuContainer.name)

    # checkedNatIP 
    if lcuContainer.natip == "auto":
        lcuContainer.checkedNatIP = checkOrDetectHostsFirstWANishIP(lcuConfiguration=lcuConfiguration)
        if lcuContainer.checkedNatIP == None:
            log.info("Unable to auto-detect non-private WANIP as automatic CTNATIP, continuing un-NAT work for container %s anyway." %
                      lcuContainer.name)
        if not ip4Reachable(lcuContainer.checkedNatIP):
            log.info("Unable to reach WANIP %s to use as automatic CTNATIP, continuing un-NAT work for container %s anyway." %
                      lcuContainer.name)
    else:
        log.info("Deleting CTNATIP %s/%d from WAN interface %s for container %s" % (
            lcuContainer.checkedNatIP.exploded, lcuConfiguration.wanMaskBits, wanIf, lcuContainer.name))
        rc = helpers.performDebugLogged(["/sbin/ip", "addr", "del",
                                        "%s/%d" % (lcuContainer.checkedNatIP.exploded, lcuConfiguration.wanMaskBits),
                                        "dev", wanIf])
        if rc != 0:
            log.error("Unable to bring up configured CTNATIP, continuing un-NAT work for container %s anyway." %
                      lcuContainer.name)

    log.notice("Actually deleting NAT rules for container %s" %
               lcuContainer.name)
    
    helpers.performDebugLogged(["/usr/sbin/iptables", "-t","nat", "-D","POSTROUTING", "-j","%s_post"%lcuContainer.natName])
    helpers.performDebugLogged(["/usr/sbin/iptables", "-t","nat", "-F","%s_post"%lcuContainer.natName])
    helpers.performDebugLogged(["/usr/sbin/iptables", "-t","nat", "-X","%s_post"%lcuContainer.natName])

    if len(lcuContainer.natportstcp+lcuContainer.natportsudp) > 0:
        helpers.performDebugLogged(["/usr/sbin/iptables", "-t","nat", "-D","PREROUTING", "-j","%s_pre"%lcuContainer.natName])
        helpers.performDebugLogged(["/usr/sbin/iptables", "-t","nat", "-F","%s_pre"%lcuContainer.natName])
        helpers.performDebugLogged(["/usr/sbin/iptables", "-t","nat", "-X","%s_pre"%lcuContainer.natName])

    log.notice("Done deleting NAT rules for container %s" %
               lcuContainer.name)
    
    return True

    
def detectLXCContainerFirstIPv4(lcuContainer:dto.LCUContainer=None, lxcContainer:dto.LCUContainer=None):
    '''
        detects a container's first IPv4 and returns it as ipaddress.IPv4Address (or None in case of failure)
        used for automatic NATing
    '''
    if lcuContainer == None:
        raise Exception("ERROR: detectLXCContainerFirstIPv4() called with lcuContainer None")
    if lxcContainer == None:
        raise Exception("ERROR: detectLXCContainerFirstIPv4() called with lxcContainer None")
       
    ctLocalIp = None
    if lxcContainer != None:
        if lxcContainer.lxcFileDict != None:
            for nr in range(0, 4):
                ip4wnm = lxcContainer.lxcFileDict.get("lxc.net.%d.ipv4.address" % nr, "")
                if len(ip4wnm) > 0:
                    try:
                        ctLocalIp = ipaddress.IPv4Interface(ip4wnm)
                    except:
                        log.warning("LXC container has weird IPv4 address configured as lxc.net.%d.ipv4.address: %s" % (
                            nr, ip4wnm))
                        continue
                    break
    if ctLocalIp == None:
        log.error("Unable to detect local IP of container %s, aborting NAT work." % lcuContainer.name)
        return None
    ctLocalIpOk = False
    #log.debug("ctLocalIp.exploded = %s" % ctLocalIp.ip.exploded)
    if lcuContainer.checkedIp0 != None:
        #log.debug("checkedIp0.exploded = %s" % lcuContainer.checkedIp0.exploded)
        if ctLocalIp.ip.exploded == lcuContainer.checkedIp0.exploded:
            ctLocalIpOk = True
    if lcuContainer.checkedIp2 != None:
        #log.debug("checkedIp2.exploded = %s" % lcuContainer.checkedIp2.exploded)
        if ctLocalIp.ip.exploded == lcuContainer.checkedIp2.exploded:
            ctLocalIpOk = True
    if not ctLocalIpOk:
        log.error("First local IP of container %s, %s, differs from CTIP0 and CTIP2, unusable for NAT work." % (
            lcuContainer.name, ctLocalIp.ip.exploded))
        return None
    return ctLocalIp.ip


#def detectLXCContainerNATIP(lcuContainer:dto.LCUContainer=None, lxcContainer:dto.LCUContainer=None):    '''
#        detects a container's NAT IP, i.e. the public IPv4 it shall get by automatic NAT work,
#        if and only if lcuContainer.natip == "auto" 
#        and sets lcuContainer.natip to checked IP or None
#        used for automatic NATing
#    '''
    


def checkOrDetectHostsFirstWANInterface(lcuConfiguration:dto.LCUConfiguration=None):
    '''
        if lcuConfiguration.wanInterface is not empty, from environment WANIF, it's returned, unchecked for now.
        otherwise auto-detection is attempted, grepping interface of first default route.
        a string is returned, non-empty values should be usable.
    '''
    if lcuConfiguration == None:
        raise Exception("ERROR: checkOrDetectHostsFirstWANInterface() called with lcuConfiguration None")
    wanIf = lcuConfiguration.wanInterface
    if len(wanIf) > 0:
        # FIXME future check
        return wanIf
    log.debug("Trying to auto-detect WAN interface")
    try:
        cp = subprocess.run(["/sbin/ip","route","show"],
                        stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True)
    except:
        log.error("Unable to auto-detect WAN interface (failed to perform ip route show), probably aborting NAT work.")
        return False
    for ipRouteLine in cp.stdout.split("\n"):
        if ipRouteLine.startswith("default via "):
            wanIf = ipRouteLine.split(" ")[4]
            log.info("Auto-detected WAN interface as %s" % wanIf)
            break
    log.debug("checkOrDetectHostsFirstWANInterface(): Returning wanIf=%s" % wanIf)
    return wanIf


def checkOrDetectHostsFirstWANishIP(lcuConfiguration:dto.LCUConfiguration=None):
    '''
        if 
        tries to auto-detect the host's first public IPv4, to use as automatic CTNATIP
        if no public IPv4 is found, but a non-local private IPv4 is found that one is used 
        is nothing is found None is returned
    '''
    if lcuConfiguration == None:
        raise Exception("ERROR: checkOrDetectHostsFirstWANishIP() called with lcuConfiguration None")
    if lcuConfiguration.checkedWanIP != None:
        log.info("Using environment's WANIP as automatic CTNATIP: %s" % lcuConfiguration.checkedWanIP.exploded)
        return lcuConfiguration.checkedWanIP

    log.info("Trying to autodetect WANIP as automatic CTNATIP")
    try:
        cp = subprocess.run(["/sbin/ip","addr","list"],
                            stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True)
    except:
        log.error("Unable to auto-detect WANIP (failed to perform ip addr list), probably aborting NAT work for container.")
        return None
    hostFirstPrivateIp = None
    usableWanIP = None
    for ipAddrLine in cp.stdout.split("\n"):
        strippedIpAddrLine = ipAddrLine.strip()
        if strippedIpAddrLine.startswith("inet "):
            ip = strippedIpAddrLine.split(" ")[1]
            try:
                checkedIp = ipaddress.IPv4Interface(ip).ip
            except:
                log.error("Unable to auto-detect WANIP (ip addr list showed invalid IP %s) as automatic CTNATIP, probably aborting NAT work for container." % ip)
                return None
            #log.debug("checkedIp from %s = %s" % (ip, checkedIp.exploded))
            if checkedIp.is_loopback:
                log.debug("Ignoring loopback IP %s as WAN IP candidate" % checkedIp.exploded)
                continue
            if checkedIp.is_private:
                if hostFirstPrivateIp == None:
                    log.debug("Ignoring first private IP %s as WAN IP canditate for now, but remembering in case there's no public WAN IP candidate" % checkedIp.exploded)
                    hostFirstPrivateIp = checkedIp
                else:
                    log.debug("Ignoring further private IP %s as WAN IP canditate" % checkedIp.exploded)
                continue
            usableWanIP = checkedIp
            log.info("Auto-detected first public WANIP as automatic CTNATIP for NATing to be %s" % (
                usableWanIP.exploded))
            break
    if usableWanIP == None:
        if hostFirstPrivateIp == None:
            log.info("Unable to auto-detect non-private WANIP as automatic CTNATIP, probably aborting NAT work for container.")
        else:
            log.info("Unable to auto-detect non-private WANIP as automatic CTNATIP but found private WAN-ish IP %s, using this." % (
                hostFirstPrivateIp.exploded))
            usableWanIP = hostFirstPrivateIp
    log.debug("detectHostsFirstWANishIP(): Returning %s" % usableWanIP)
    return usableWanIP















